# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from typing import Final

HEROKU_APP_ID: Final = "heroku.app.id"
"""
Unique identifier for the application.
"""

HEROKU_RELEASE_COMMIT: Final = "heroku.release.commit"
"""
Commit hash for the current release.
"""

HEROKU_RELEASE_CREATION_TIMESTAMP: Final = "heroku.release.creation_timestamp"
"""
Time and date the release was created.
"""
