# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from typing import Final

ELASTICSEARCH_NODE_NAME: Final = "elasticsearch.node.name"
"""
Represents the human-readable identifier of the node/instance to which a request was routed.
"""
