"""
The matcher subsystem needs a function called "confirm_top", which
takes the data passed to a top file environment and determines if that
data matches this minion.
"""

import logging

import salt.loader

log = logging.getLogger(__file__)


def confirm_top(match, data, nodegroups=None):
    """
    Takes the data passed to a top file environment and determines if the
    data matches this minion
    """
    matcher = "compound"
    for item in data:
        if isinstance(item, dict):
            if "match" in item:
                matcher = item["match"]

    if "matchers" in __context__:
        matchers = __context__["matchers"]
    else:
        # Matchers need pillar data if available
        pillar = __pillar__ if "__pillar__" in globals() else None
        if hasattr(pillar, "value"):
            pillar = pillar.value()
        matchers = salt.loader.matchers(__opts__, context=__context__, pillar=pillar)
        __context__["matchers"] = matchers
    funcname = matcher + "_match.match"
    if matcher == "nodegroup":
        return matchers[funcname](match, nodegroups)
    else:
        m = matchers[funcname]
        return m(match)
    # except TypeError, KeyError:
    #     log.error("Attempting to match with unknown matcher: %s", matcher)
