{% if networking %}NETWORKING={{networking}}
{%endif%}{% if hostname %}HOSTNAME={{hostname}}
{%endif-%}{% if domainname %}DOMAIN={{domainname}}
{%endif-%}{% if searchdomain %}SEARCH={{searchdomain}}
{%endif-%}
