# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from typing import Final

POOL_NAME: Final = "pool.name"
"""
Deprecated: Replaced by `db.client.connection.pool.name`.
"""
