# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from typing import Final

MAINFRAME_LPAR_NAME: Final = "mainframe.lpar.name"
"""
Name of the logical partition that hosts a systems with a mainframe operating system.
"""
