# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from typing import Final

OPENSHIFT_CLUSTERQUOTA_NAME: Final = "openshift.clusterquota.name"
"""
The name of the cluster quota.
"""

OPENSHIFT_CLUSTERQUOTA_UID: Final = "openshift.clusterquota.uid"
"""
The UID of the cluster quota.
"""
