# File generated by contrib/scrape-ec2-sizes.py script - DO NOT EDIT manually
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

INSTANCE_TYPES = {
    "a1.2xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "No",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "a1.2xlarge",
            "instancesku": "A6PYBM8K8FAMKCGJ",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "a1.2xlarge",
        "name": "a1.2xlarge",
        "ram": 16384,
    },
    "a1.4xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "No",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "a1.4xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AWS Graviton Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "a1.4xlarge",
        "name": "a1.4xlarge",
        "ram": 32768,
    },
    "a1.large": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "No",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "a1.large",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "a1.large",
        "name": "a1.large",
        "ram": 4096,
    },
    "a1.medium": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "No",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "a1.medium",
            "instancesku": "6H6ECBH6BV59BRSN",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "2 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AWS Graviton Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "1",
            "vpcnetworkingsupport": "true",
        },
        "id": "a1.medium",
        "name": "a1.medium",
        "ram": 2048,
    },
    "a1.xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "No",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "a1.xlarge",
            "instancesku": "MPNJUDSDMWJQYKGQ",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "a1.xlarge",
        "name": "a1.xlarge",
        "ram": 8192,
    },
    "c1.medium": {
        "bandwidth": None,
        "disk": 350,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "currentGeneration": "No",
            "ecu": "5",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c1.medium",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "1.7 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "32-bit or 64-bit",
            "regionCode": "us-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 350 SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "c1.medium",
        "name": "c1.medium",
        "ram": 1740,
    },
    "c1.xlarge": {
        "bandwidth": None,
        "disk": 1680,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "currentGeneration": "No",
            "ecu": "20",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c1.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "7 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 420 SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "c1.xlarge",
        "name": "c1.xlarge",
        "ram": 7168,
    },
    "c3.2xlarge": {
        "bandwidth": None,
        "disk": 160,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "No",
            "ecu": "28",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c3.2xlarge",
            "instancesku": "YCNS7SECEMFTZFYH",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "15 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon E5-2680 v2 (Ivy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 80 SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "c3.2xlarge",
        "name": "c3.2xlarge",
        "ram": 15360,
    },
    "c3.4xlarge": {
        "bandwidth": None,
        "disk": 320,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "No",
            "ecu": "55",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c3.4xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "30 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon E5-2680 v2 (Ivy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 160 SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "c3.4xlarge",
        "name": "c3.4xlarge",
        "ram": 30720,
    },
    "c3.8xlarge": {
        "bandwidth": 10,
        "disk": 640,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "No",
            "ecu": "108",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c3.8xlarge",
            "instancesku": "FKFNCVGF8F4VBXQ4",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "60 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon E5-2680 v2 (Ivy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "regionCode": "sa-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 320 SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "c3.8xlarge",
        "name": "c3.8xlarge",
        "ram": 61440,
    },
    "c3.large": {
        "bandwidth": None,
        "disk": 32,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "No",
            "ecu": "7",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c3.large",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "3.75 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon E5-2680 v2 (Ivy Bridge)",
            "processorArchitecture": "32-bit or 64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 16 SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "c3.large",
        "name": "c3.large",
        "ram": 3840,
    },
    "c3.xlarge": {
        "bandwidth": None,
        "disk": 80,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "No",
            "ecu": "14",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c3.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "7.5 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon E5-2680 v2 (Ivy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 40 SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "c3.xlarge",
        "name": "c3.xlarge",
        "ram": 7680,
    },
    "c4.2xlarge": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.9 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "1000 Mbps",
            "ecu": "31",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c4.2xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "15 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon E5-2666 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "us-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "c4.2xlarge",
        "name": "c4.2xlarge",
        "ram": 15360,
    },
    "c4.4xlarge": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.9 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "2000 Mbps",
            "ecu": "62",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c4.4xlarge",
            "instancesku": "RGU9HRNUAS2TX83W",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "30 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon E5-2666 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "c4.4xlarge",
        "name": "c4.4xlarge",
        "ram": 30720,
    },
    "c4.8xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.9 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4000 Mbps",
            "ecu": "132",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c4.8xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "60 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon E5-2666 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "36",
            "vpcnetworkingsupport": "true",
        },
        "id": "c4.8xlarge",
        "name": "c4.8xlarge",
        "ram": 61440,
    },
    "c4.large": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.9 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "500 Mbps",
            "ecu": "8",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c4.large",
            "instancesku": "TX5YYKC8NVXPN43D",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "3.75 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon E5-2666 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "c4.large",
        "name": "c4.large",
        "ram": 3840,
    },
    "c4.xlarge": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.9 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "750 Mbps",
            "ecu": "16",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c4.xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "7.5 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon E5-2666 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "c4.xlarge",
        "name": "c4.xlarge",
        "ram": 7680,
    },
    "c5.12xlarge": {
        "bandwidth": 12,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "188",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5.12xlarge",
            "instancesku": "B2D6JZJWM7HPMFFP",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "12 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Platinum 8275L",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Deep Learning Boost; Intel Turbo",
            "regionCode": "eu-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5.12xlarge",
        "name": "c5.12xlarge",
        "ram": 98304,
    },
    "c5.18xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "9000 Mbps",
            "ecu": "281",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5.18xlarge",
            "instancesku": "H2XX5A4Y823VVMXZ",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "144 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "144",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "72",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5.18xlarge",
        "name": "c5.18xlarge",
        "ram": 147456,
    },
    "c5.24xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14000 Mbps",
            "ecu": "375",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5.24xlarge",
            "instancesku": "GB2CVQJXDU9FR8GT",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8275L",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Deep Learning Boost; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5.24xlarge",
        "name": "c5.24xlarge",
        "ram": 196608,
    },
    "c5.2xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2250 Mbps",
            "ecu": "39",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5.2xlarge",
            "instancesku": "A5V9YQVUZMY5XA83",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5.2xlarge",
        "name": "c5.2xlarge",
        "ram": 16384,
    },
    "c5.4xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2250 Mbps",
            "ecu": "73",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5.4xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-3",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5.4xlarge",
        "name": "c5.4xlarge",
        "ram": 32768,
    },
    "c5.9xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4500 Mbps",
            "ecu": "139",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5.9xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "72 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "72",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "36",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5.9xlarge",
        "name": "c5.9xlarge",
        "ram": 73728,
    },
    "c5.large": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2250 Mbps",
            "ecu": "10",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5.large",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5.large",
        "name": "c5.large",
        "ram": 4096,
    },
    "c5.xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2250 Mbps",
            "ecu": "20",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5.xlarge",
            "instancesku": "K6WNSK6Q2CWEYGHX",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5.xlarge",
        "name": "c5.xlarge",
        "ram": 8192,
    },
    "c5a.12xlarge": {
        "bandwidth": 12,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4750 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5a.12xlarge",
            "instancesku": "EGSVE9S2NJJWJ3SC",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "12 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "32-bit or 64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "eu-west-3",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5a.12xlarge",
        "name": "c5a.12xlarge",
        "ram": 98304,
    },
    "c5a.16xlarge": {
        "bandwidth": 20,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "6300 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5a.16xlarge",
            "instancesku": "3YMAUTFQ8JZCV6VS",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "32-bit or 64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5a.16xlarge",
        "name": "c5a.16xlarge",
        "ram": 131072,
    },
    "c5a.24xlarge": {
        "bandwidth": 20,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "9500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5a.24xlarge",
            "instancesku": "DPHJ9PS8GF7ZRFEF",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "32-bit or 64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "eu-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5a.24xlarge",
        "name": "c5a.24xlarge",
        "ram": 196608,
    },
    "c5a.2xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3170 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5a.2xlarge",
            "instancesku": "F4FFZANT2KEGQPZU",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "32-bit or 64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5a.2xlarge",
        "name": "c5a.2xlarge",
        "ram": 16384,
    },
    "c5a.4xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3170 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5a.4xlarge",
            "instancesku": "FXZCY68C879B2BXD",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "32-bit or 64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5a.4xlarge",
        "name": "c5a.4xlarge",
        "ram": 32768,
    },
    "c5a.8xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3170 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5a.8xlarge",
            "instancesku": "2MVSRPHH92E3542M",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "32-bit or 64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5a.8xlarge",
        "name": "c5a.8xlarge",
        "ram": 65536,
    },
    "c5a.large": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3170 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5a.large",
            "instancesku": "TPQTDU7RNDMJK6KP",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "32-bit or 64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5a.large",
        "name": "c5a.large",
        "ram": 4096,
    },
    "c5a.xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3170 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5a.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "32-bit or 64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5a.xlarge",
        "name": "c5a.xlarge",
        "ram": 8192,
    },
    "c5ad.12xlarge": {
        "bandwidth": 12,
        "disk": 1800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4750 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5ad.12xlarge",
            "instancesku": "FXDEW36Z8YRSGM2S",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "12 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "eu-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 900 NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5ad.12xlarge",
        "name": "c5ad.12xlarge",
        "ram": 98304,
    },
    "c5ad.16xlarge": {
        "bandwidth": 20,
        "disk": 2400,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "6300 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5ad.16xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1200 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5ad.16xlarge",
        "name": "c5ad.16xlarge",
        "ram": 131072,
    },
    "c5ad.24xlarge": {
        "bandwidth": 20,
        "disk": 3800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "9500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5ad.24xlarge",
            "instancesku": "NKHYKUFYUFS54VHC",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1900 NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5ad.24xlarge",
        "name": "c5ad.24xlarge",
        "ram": 196608,
    },
    "c5ad.2xlarge": {
        "bandwidth": 10,
        "disk": 300,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3170 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5ad.2xlarge",
            "instancesku": "34Z2EYXRD347K3SA",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 300 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5ad.2xlarge",
        "name": "c5ad.2xlarge",
        "ram": 16384,
    },
    "c5ad.4xlarge": {
        "bandwidth": 10,
        "disk": 600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3170 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5ad.4xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "sa-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 300 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5ad.4xlarge",
        "name": "c5ad.4xlarge",
        "ram": 32768,
    },
    "c5ad.8xlarge": {
        "bandwidth": 10,
        "disk": 1200,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3170 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5ad.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 600 NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5ad.8xlarge",
        "name": "c5ad.8xlarge",
        "ram": 65536,
    },
    "c5ad.large": {
        "bandwidth": 10,
        "disk": 75,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3170 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5ad.large",
            "instancesku": "HT8MSHND4S9PP8YT",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 75 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5ad.large",
        "name": "c5ad.large",
        "ram": 4096,
    },
    "c5ad.xlarge": {
        "bandwidth": 10,
        "disk": 150,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3170 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5ad.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 150 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5ad.xlarge",
        "name": "c5ad.xlarge",
        "ram": 8192,
    },
    "c5d.12xlarge": {
        "bandwidth": 12,
        "disk": 1800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "188",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5d.12xlarge",
            "instancesku": "SYZAEXP42ASXK43Y",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "12 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Platinum 8275L",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Deep Learning Boost; Intel Turbo",
            "regionCode": "eu-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 900 NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5d.12xlarge",
        "name": "c5d.12xlarge",
        "ram": 98304,
    },
    "c5d.18xlarge": {
        "bandwidth": 25,
        "disk": 1800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "9000 Mbps",
            "ecu": "281",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5d.18xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "144 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "144",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 900 NVMe SSD",
            "vcpu": "72",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5d.18xlarge",
        "name": "c5d.18xlarge",
        "ram": 147456,
    },
    "c5d.24xlarge": {
        "bandwidth": 25,
        "disk": 3600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14000 Mbps",
            "ecu": "375",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5d.24xlarge",
            "instancesku": "CBCNASX9DSMBPMQW",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8275L",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Deep Learning Boost; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 900 NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5d.24xlarge",
        "name": "c5d.24xlarge",
        "ram": 196608,
    },
    "c5d.2xlarge": {
        "bandwidth": 10,
        "disk": 200,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2250 Mbps",
            "ecu": "39",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5d.2xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "sa-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 200 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5d.2xlarge",
        "name": "c5d.2xlarge",
        "ram": 16384,
    },
    "c5d.4xlarge": {
        "bandwidth": 10,
        "disk": 400,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2250 Mbps",
            "ecu": "73",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5d.4xlarge",
            "instancesku": "FUGNF8J88PBAQDYT",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 400 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5d.4xlarge",
        "name": "c5d.4xlarge",
        "ram": 32768,
    },
    "c5d.9xlarge": {
        "bandwidth": 10,
        "disk": 900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4500 Mbps",
            "ecu": "139",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5d.9xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "72 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "72",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 900 NVMe SSD",
            "vcpu": "36",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5d.9xlarge",
        "name": "c5d.9xlarge",
        "ram": 73728,
    },
    "c5d.large": {
        "bandwidth": 10,
        "disk": 50,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2250 Mbps",
            "ecu": "10",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5d.large",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "sa-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 50 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5d.large",
        "name": "c5d.large",
        "ram": 4096,
    },
    "c5d.xlarge": {
        "bandwidth": 10,
        "disk": 100,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2250 Mbps",
            "ecu": "20",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5d.xlarge",
            "instancesku": "BEKRK57J4EHTR2VJ",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 100 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5d.xlarge",
        "name": "c5d.xlarge",
        "ram": 8192,
    },
    "c5n.18xlarge": {
        "bandwidth": 100,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14000 Mbps",
            "ecu": "281",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5n.18xlarge",
            "instancesku": "4VUXUHRQ7ZTHKVDG",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "144",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "72",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5n.18xlarge",
        "name": "c5n.18xlarge",
        "ram": 196608,
    },
    "c5n.2xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "39",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5n.2xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "21 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5n.2xlarge",
        "name": "c5n.2xlarge",
        "ram": 21504,
    },
    "c5n.4xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "73",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5n.4xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "42 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5n.4xlarge",
        "name": "c5n.4xlarge",
        "ram": 43008,
    },
    "c5n.9xlarge": {
        "bandwidth": 50,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "139",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5n.9xlarge",
            "instancesku": "R6T5FWW48CMRQ55E",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "72",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "36",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5n.9xlarge",
        "name": "c5n.9xlarge",
        "ram": 98304,
    },
    "c5n.large": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "10",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5n.large",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "5.25 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "sa-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5n.large",
        "name": "c5n.large",
        "ram": 5376,
    },
    "c5n.xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "20",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c5n.xlarge",
            "instancesku": "9S2WRMC8XXXDEGVU",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "10.5 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8124M",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "c5n.xlarge",
        "name": "c5n.xlarge",
        "ram": 10752,
    },
    "c6a.12xlarge": {
        "bandwidth": 18750,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6a.12xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "18750 Megabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6a.12xlarge",
        "name": "c6a.12xlarge",
        "ram": 98304,
    },
    "c6a.16xlarge": {
        "bandwidth": 25000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "13333 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6a.16xlarge",
            "instancesku": "RS7S28QKZJKJSZNJ",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "25000 Megabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6a.16xlarge",
        "name": "c6a.16xlarge",
        "ram": 131072,
    },
    "c6a.24xlarge": {
        "bandwidth": 37500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6a.24xlarge",
            "instancesku": "8ZKSQWSAQGRJ5U9F",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "37500 Megabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6a.24xlarge",
        "name": "c6a.24xlarge",
        "ram": 196608,
    },
    "c6a.2xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 6667 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6a.2xlarge",
            "instancesku": "GFUW4QDP4XTQR6QV",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "us-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6a.2xlarge",
        "name": "c6a.2xlarge",
        "ram": 16384,
    },
    "c6a.32xlarge": {
        "bandwidth": 50000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "26667 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6a.32xlarge",
            "instancesku": "5N3P5DVUANCGQA54",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "50000 Megabit",
            "normalizationSizeFactor": "256",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "sa-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "128",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6a.32xlarge",
        "name": "c6a.32xlarge",
        "ram": 262144,
    },
    "c6a.48xlarge": {
        "bandwidth": 50000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "40000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6a.48xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "50000 Megabit",
            "normalizationSizeFactor": "384",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "192",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6a.48xlarge",
        "name": "c6a.48xlarge",
        "ram": 393216,
    },
    "c6a.4xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 6667 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6a.4xlarge",
            "instancesku": "Z8JWH2QQ3K4VX624",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "ap-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6a.4xlarge",
        "name": "c6a.4xlarge",
        "ram": 32768,
    },
    "c6a.8xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "6667 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6a.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "12500 Megabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "ap-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6a.8xlarge",
        "name": "c6a.8xlarge",
        "ram": 65536,
    },
    "c6a.large": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 6667 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6a.large",
            "instancesku": "RR9ZPGTHWWC4U5NQ",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6a.large",
        "name": "c6a.large",
        "ram": 4096,
    },
    "c6a.xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 6667 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6a.xlarge",
            "instancesku": "A32CS95FPD6HAU69",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6a.xlarge",
        "name": "c6a.xlarge",
        "ram": 8192,
    },
    "c6g.12xlarge": {
        "bandwidth": 20,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "13500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6g.12xlarge",
            "instancesku": "AY5FPVXQ2UDEUWN3",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6g.12xlarge",
        "name": "c6g.12xlarge",
        "ram": 98304,
    },
    "c6g.16xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "19000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6g.16xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6g.16xlarge",
        "name": "c6g.16xlarge",
        "ram": 131072,
    },
    "c6g.2xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6g.2xlarge",
            "instancesku": "64UAZ5JQ6EFAVPMB",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6g.2xlarge",
        "name": "c6g.2xlarge",
        "ram": 16384,
    },
    "c6g.4xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6g.4xlarge",
            "instancesku": "MFVUP7HCAZXZGZ4E",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-3",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6g.4xlarge",
        "name": "c6g.4xlarge",
        "ram": 32768,
    },
    "c6g.8xlarge": {
        "bandwidth": 12,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "9000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6g.8xlarge",
            "instancesku": "PB5P6ZJYNWR2SRGE",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "12 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6g.8xlarge",
        "name": "c6g.8xlarge",
        "ram": 65536,
    },
    "c6g.large": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6g.large",
            "instancesku": "6PQ4U29URPRQ2MYZ",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6g.large",
        "name": "c6g.large",
        "ram": 4096,
    },
    "c6g.medium": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6g.medium",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "2 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "1",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6g.medium",
        "name": "c6g.medium",
        "ram": 2048,
    },
    "c6g.xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6g.xlarge",
            "instancesku": "URHNUKBP5VQJSGKX",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6g.xlarge",
        "name": "c6g.xlarge",
        "ram": 8192,
    },
    "c6gd.12xlarge": {
        "bandwidth": 20,
        "disk": 2850,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "13500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6gd.12xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1425 NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6gd.12xlarge",
        "name": "c6gd.12xlarge",
        "ram": 98304,
    },
    "c6gd.16xlarge": {
        "bandwidth": 25,
        "disk": 3800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "18000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6gd.16xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1900 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6gd.16xlarge",
        "name": "c6gd.16xlarge",
        "ram": 131072,
    },
    "c6gd.2xlarge": {
        "bandwidth": 10,
        "disk": 475,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6gd.2xlarge",
            "instancesku": "TDQ7AUJRCKNXMPDK",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 475 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6gd.2xlarge",
        "name": "c6gd.2xlarge",
        "ram": 16384,
    },
    "c6gd.4xlarge": {
        "bandwidth": 10,
        "disk": 950,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6gd.4xlarge",
            "instancesku": "5BHRV5TW4H89DRKT",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 950 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6gd.4xlarge",
        "name": "c6gd.4xlarge",
        "ram": 32768,
    },
    "c6gd.8xlarge": {
        "bandwidth": 12,
        "disk": 1900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "9000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6gd.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "12 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1900 NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6gd.8xlarge",
        "name": "c6gd.8xlarge",
        "ram": 65536,
    },
    "c6gd.large": {
        "bandwidth": 10,
        "disk": 118,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6gd.large",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 118 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6gd.large",
        "name": "c6gd.large",
        "ram": 4096,
    },
    "c6gd.medium": {
        "bandwidth": 10,
        "disk": 59,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6gd.medium",
            "instancesku": "YE6EEQZMKGT4KERM",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "2 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 59 NVMe SSD",
            "vcpu": "1",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6gd.medium",
        "name": "c6gd.medium",
        "ram": 2048,
    },
    "c6gd.xlarge": {
        "bandwidth": 10,
        "disk": 237,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6gd.xlarge",
            "instancesku": "GGZTRMP3FBJ7M5QD",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 237 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6gd.xlarge",
        "name": "c6gd.xlarge",
        "ram": 8192,
    },
    "c6gn.12xlarge": {
        "bandwidth": 75,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "28500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6gn.12xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "75 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-southeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6gn.12xlarge",
        "name": "c6gn.12xlarge",
        "ram": 98304,
    },
    "c6gn.16xlarge": {
        "bandwidth": 100,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "38000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6gn.16xlarge",
            "instancesku": "Y9DP6NNEHBSYS443",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "sa-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6gn.16xlarge",
        "name": "c6gn.16xlarge",
        "ram": 131072,
    },
    "c6gn.2xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 9500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6gn.2xlarge",
            "instancesku": "M6MVR7VW6XTWPHQ5",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6gn.2xlarge",
        "name": "c6gn.2xlarge",
        "ram": 16384,
    },
    "c6gn.4xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "9500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6gn.4xlarge",
            "instancesku": "5472CKGEUNZGUDSP",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-3",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6gn.4xlarge",
        "name": "c6gn.4xlarge",
        "ram": 32768,
    },
    "c6gn.8xlarge": {
        "bandwidth": 50,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "19000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6gn.8xlarge",
            "instancesku": "NX8CJEX2JY5HXHQR",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6gn.8xlarge",
        "name": "c6gn.8xlarge",
        "ram": 65536,
    },
    "c6gn.large": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 9500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6gn.large",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6gn.large",
        "name": "c6gn.large",
        "ram": 4096,
    },
    "c6gn.medium": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 9500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6gn.medium",
            "instancesku": "KETXG4986RNYCQJ8",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "2 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "1",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6gn.medium",
        "name": "c6gn.medium",
        "ram": 2048,
    },
    "c6gn.xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 9500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6gn.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-3",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6gn.xlarge",
        "name": "c6gn.xlarge",
        "ram": 8192,
    },
    "c6i.12xlarge": {
        "bandwidth": 18750,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "15000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6i.12xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "18750 Megabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6i.12xlarge",
        "name": "c6i.12xlarge",
        "ram": 98304,
    },
    "c6i.16xlarge": {
        "bandwidth": 25000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6i.16xlarge",
            "instancesku": "U2QZS64THC8FN76P",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "25000 Megabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6i.16xlarge",
        "name": "c6i.16xlarge",
        "ram": 131072,
    },
    "c6i.24xlarge": {
        "bandwidth": 37500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "30000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6i.24xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "37500 Megabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6i.24xlarge",
        "name": "c6i.24xlarge",
        "ram": 196608,
    },
    "c6i.2xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6i.2xlarge",
            "instancesku": "BQT6TD2YB6MSN4G4",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6i.2xlarge",
        "name": "c6i.2xlarge",
        "ram": 16384,
    },
    "c6i.32xlarge": {
        "bandwidth": 50000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "40000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6i.32xlarge",
            "instancesku": "QNUQSM3BUHRHUWTU",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "50000 Megabit",
            "normalizationSizeFactor": "256",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "128",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6i.32xlarge",
        "name": "c6i.32xlarge",
        "ram": 262144,
    },
    "c6i.4xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6i.4xlarge",
            "instancesku": "KBFT9CRJ8V4C6W5F",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6i.4xlarge",
        "name": "c6i.4xlarge",
        "ram": 32768,
    },
    "c6i.8xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6i.8xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "12500 Megabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6i.8xlarge",
        "name": "c6i.8xlarge",
        "ram": 65536,
    },
    "c6i.large": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6i.large",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6i.large",
        "name": "c6i.large",
        "ram": 4096,
    },
    "c6i.xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6i.xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-3",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6i.xlarge",
        "name": "c6i.xlarge",
        "ram": 8192,
    },
    "c6id.12xlarge": {
        "bandwidth": 18750,
        "disk": 2850,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "15000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6id.12xlarge",
            "instancesku": "FQE85V2GX4ZFSD2S",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "18750 Megabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-southeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1425 NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6id.12xlarge",
        "name": "c6id.12xlarge",
        "ram": 98304,
    },
    "c6id.16xlarge": {
        "bandwidth": 25000,
        "disk": 3800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6id.16xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "25000 Megabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1900 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6id.16xlarge",
        "name": "c6id.16xlarge",
        "ram": 131072,
    },
    "c6id.24xlarge": {
        "bandwidth": 37500,
        "disk": 5700,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "30000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6id.24xlarge",
            "instancesku": "Y4ZVY5VNPN8STK52",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "37500 Megabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 1425 NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6id.24xlarge",
        "name": "c6id.24xlarge",
        "ram": 196608,
    },
    "c6id.2xlarge": {
        "bandwidth": 12500,
        "disk": 474,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6id.2xlarge",
            "instancesku": "X9N8F8866QWAPQNP",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 474 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6id.2xlarge",
        "name": "c6id.2xlarge",
        "ram": 16384,
    },
    "c6id.32xlarge": {
        "bandwidth": 50000,
        "disk": 7600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "40000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6id.32xlarge",
            "instancesku": "MG4MW9ZM927UDCXJ",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "50000 Megabit",
            "normalizationSizeFactor": "256",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "sa-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 1900 NVMe SSD",
            "vcpu": "128",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6id.32xlarge",
        "name": "c6id.32xlarge",
        "ram": 262144,
    },
    "c6id.4xlarge": {
        "bandwidth": 12500,
        "disk": 950,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6id.4xlarge",
            "instancesku": "ZZSD3B794W6EQUBC",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 950 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6id.4xlarge",
        "name": "c6id.4xlarge",
        "ram": 32768,
    },
    "c6id.8xlarge": {
        "bandwidth": 12500,
        "disk": 1900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6id.8xlarge",
            "instancesku": "6NA33DRTEUAMJA2N",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "12500 Megabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1900 NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6id.8xlarge",
        "name": "c6id.8xlarge",
        "ram": 65536,
    },
    "c6id.large": {
        "bandwidth": 12500,
        "disk": 118,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6id.large",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 118 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6id.large",
        "name": "c6id.large",
        "ram": 4096,
    },
    "c6id.xlarge": {
        "bandwidth": 12500,
        "disk": 237,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6id.xlarge",
            "instancesku": "6KDKHB5SQW6BX8N2",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 237 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6id.xlarge",
        "name": "c6id.xlarge",
        "ram": 8192,
    },
    "c6in.12xlarge": {
        "bandwidth": 75000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "37500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6in.12xlarge",
            "instancesku": "F8D9XRE64QEESJRE",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "75000 Megabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6in.12xlarge",
        "name": "c6in.12xlarge",
        "ram": 98304,
    },
    "c6in.16xlarge": {
        "bandwidth": 100000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "50000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6in.16xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "100000 Megabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6in.16xlarge",
        "name": "c6in.16xlarge",
        "ram": 131072,
    },
    "c6in.24xlarge": {
        "bandwidth": 150000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "75000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6in.24xlarge",
            "instancesku": "A9XBZ7SQVVKSY9AF",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "150000 Megabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6in.24xlarge",
        "name": "c6in.24xlarge",
        "ram": 196608,
    },
    "c6in.2xlarge": {
        "bandwidth": 40000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 25000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6in.2xlarge",
            "instancesku": "4MHZEWP3BBH6PYGB",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 40000 Megabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6in.2xlarge",
        "name": "c6in.2xlarge",
        "ram": 16384,
    },
    "c6in.32xlarge": {
        "bandwidth": 200000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "100000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6in.32xlarge",
            "instancesku": "C9DWEPQPKMXGXS4T",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "200000 Megabit",
            "normalizationSizeFactor": "256",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "128",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6in.32xlarge",
        "name": "c6in.32xlarge",
        "ram": 262144,
    },
    "c6in.4xlarge": {
        "bandwidth": 50000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 25000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6in.4xlarge",
            "instancesku": "H38WNUZQAFK88DC3",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 50000 Megabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6in.4xlarge",
        "name": "c6in.4xlarge",
        "ram": 32768,
    },
    "c6in.8xlarge": {
        "bandwidth": 50000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "25000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6in.8xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "50000 Megabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6in.8xlarge",
        "name": "c6in.8xlarge",
        "ram": 65536,
    },
    "c6in.large": {
        "bandwidth": 25000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 25000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6in.large",
            "instancesku": "GSXW7HA3ADXKUQCG",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 25000 Megabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6in.large",
        "name": "c6in.large",
        "ram": 4096,
    },
    "c6in.xlarge": {
        "bandwidth": 30000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 25000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c6in.xlarge",
            "instancesku": "EHH289PRYCPUP9HQ",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 30000 Megabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "c6in.xlarge",
        "name": "c6in.xlarge",
        "ram": 8192,
    },
    "c7a.12xlarge": {
        "bandwidth": 18750,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "15000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7a.12xlarge",
            "instancesku": "BVWGQB6MV968URSH",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "18750 Megabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7a.12xlarge",
        "name": "c7a.12xlarge",
        "ram": 98304,
    },
    "c7a.16xlarge": {
        "bandwidth": 25000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7a.16xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "25000 Megabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7a.16xlarge",
        "name": "c7a.16xlarge",
        "ram": 131072,
    },
    "c7a.24xlarge": {
        "bandwidth": 37500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "30000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7a.24xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "37500 Megabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7a.24xlarge",
        "name": "c7a.24xlarge",
        "ram": 196608,
    },
    "c7a.2xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7a.2xlarge",
            "instancesku": "463W6TZN959Y4KDR",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7a.2xlarge",
        "name": "c7a.2xlarge",
        "ram": 16384,
    },
    "c7a.32xlarge": {
        "bandwidth": 50000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "40000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7a.32xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "50000 Megabit",
            "normalizationSizeFactor": "256",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "128",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7a.32xlarge",
        "name": "c7a.32xlarge",
        "ram": 262144,
    },
    "c7a.48xlarge": {
        "bandwidth": 50000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "40000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7a.48xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "50000 Megabit",
            "normalizationSizeFactor": "384",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "192",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7a.48xlarge",
        "name": "c7a.48xlarge",
        "ram": 393216,
    },
    "c7a.4xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7a.4xlarge",
            "instancesku": "KA3N6AD5GYB9SBDF",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7a.4xlarge",
        "name": "c7a.4xlarge",
        "ram": 32768,
    },
    "c7a.8xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7a.8xlarge",
            "instancesku": "3VCTVAWBTY8WBYEZ",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "12500 Megabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7a.8xlarge",
        "name": "c7a.8xlarge",
        "ram": 65536,
    },
    "c7a.large": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7a.large",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7a.large",
        "name": "c7a.large",
        "ram": 4096,
    },
    "c7a.medium": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7a.medium",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "2 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "1",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7a.medium",
        "name": "c7a.medium",
        "ram": 2048,
    },
    "c7a.xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7a.xlarge",
            "instancesku": "HPPW7E3GCHGU3MUQ",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7a.xlarge",
        "name": "c7a.xlarge",
        "ram": 8192,
    },
    "c7g.12xlarge": {
        "bandwidth": 22500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "15000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7g.12xlarge",
            "instancesku": "TN8APU5VNTFQQR2R",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "22500 Megabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7g.12xlarge",
        "name": "c7g.12xlarge",
        "ram": 98304,
    },
    "c7g.16xlarge": {
        "bandwidth": 30,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7g.16xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "30 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7g.16xlarge",
        "name": "c7g.16xlarge",
        "ram": 131072,
    },
    "c7g.2xlarge": {
        "bandwidth": 15,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7g.2xlarge",
            "instancesku": "VVFTAGMDE2YV8AT3",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 15 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7g.2xlarge",
        "name": "c7g.2xlarge",
        "ram": 16384,
    },
    "c7g.4xlarge": {
        "bandwidth": 15,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7g.4xlarge",
            "instancesku": "GMSXGX25V8JBVJJU",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 15 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7g.4xlarge",
        "name": "c7g.4xlarge",
        "ram": 32768,
    },
    "c7g.8xlarge": {
        "bandwidth": 15,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7g.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "15 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7g.8xlarge",
        "name": "c7g.8xlarge",
        "ram": 65536,
    },
    "c7g.large": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7g.large",
            "instancesku": "XCUDUSUEF98YBBJ8",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7g.large",
        "name": "c7g.large",
        "ram": 4096,
    },
    "c7g.medium": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7g.medium",
            "instancesku": "7SPG6TBVVQVTJ6ZS",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "2 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "1",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7g.medium",
        "name": "c7g.medium",
        "ram": 2048,
    },
    "c7g.xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7g.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7g.xlarge",
        "name": "c7g.xlarge",
        "ram": 8192,
    },
    "c7gd.12xlarge": {
        "bandwidth": 22500,
        "disk": 2850,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "15000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7gd.12xlarge",
            "instancesku": "Q9QQTXJNDFQ6E9NY",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "22500 Megabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1425 NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7gd.12xlarge",
        "name": "c7gd.12xlarge",
        "ram": 98304,
    },
    "c7gd.16xlarge": {
        "bandwidth": 30,
        "disk": 3800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7gd.16xlarge",
            "instancesku": "SYVPBBHHCKHGVNNG",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "30 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1900 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7gd.16xlarge",
        "name": "c7gd.16xlarge",
        "ram": 131072,
    },
    "c7gd.2xlarge": {
        "bandwidth": 15,
        "disk": 475,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7gd.2xlarge",
            "instancesku": "XMAWGNWQEVXWGXFW",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 15 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 475 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7gd.2xlarge",
        "name": "c7gd.2xlarge",
        "ram": 16384,
    },
    "c7gd.4xlarge": {
        "bandwidth": 15,
        "disk": 950,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7gd.4xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 15 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 950 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7gd.4xlarge",
        "name": "c7gd.4xlarge",
        "ram": 32768,
    },
    "c7gd.8xlarge": {
        "bandwidth": 15,
        "disk": 1900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7gd.8xlarge",
            "instancesku": "5VF8FN2TYA4Y2QN3",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "15 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1900 NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7gd.8xlarge",
        "name": "c7gd.8xlarge",
        "ram": 65536,
    },
    "c7gd.large": {
        "bandwidth": 12500,
        "disk": 118,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7gd.large",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 118 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7gd.large",
        "name": "c7gd.large",
        "ram": 4096,
    },
    "c7gd.medium": {
        "bandwidth": 12500,
        "disk": 59,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7gd.medium",
            "instancesku": "Q99HFCWCDYUBU7AG",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "2 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 59 NVMe SSD",
            "vcpu": "1",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7gd.medium",
        "name": "c7gd.medium",
        "ram": 2048,
    },
    "c7gd.xlarge": {
        "bandwidth": 12500,
        "disk": 237,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7gd.xlarge",
            "instancesku": "NTATYNX8P99WK9HF",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-southeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 237 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7gd.xlarge",
        "name": "c7gd.xlarge",
        "ram": 8192,
    },
    "c7gn.12xlarge": {
        "bandwidth": 150,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "30000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7gn.12xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "150 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7gn.12xlarge",
        "name": "c7gn.12xlarge",
        "ram": 98304,
    },
    "c7gn.16xlarge": {
        "bandwidth": 200,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "40000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7gn.16xlarge",
            "instancesku": "BCDJJ8FFU9DWK4VB",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "200 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7gn.16xlarge",
        "name": "c7gn.16xlarge",
        "ram": 131072,
    },
    "c7gn.2xlarge": {
        "bandwidth": 50,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7gn.2xlarge",
            "instancesku": "N33ZDXWC2PTAH9YP",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 50 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7gn.2xlarge",
        "name": "c7gn.2xlarge",
        "ram": 16384,
    },
    "c7gn.4xlarge": {
        "bandwidth": 50,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7gn.4xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7gn.4xlarge",
        "name": "c7gn.4xlarge",
        "ram": 32768,
    },
    "c7gn.8xlarge": {
        "bandwidth": 100,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7gn.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7gn.8xlarge",
        "name": "c7gn.8xlarge",
        "ram": 65536,
    },
    "c7gn.large": {
        "bandwidth": 30,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7gn.large",
            "instancesku": "DFN9T3JRTY8JSEEH",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 30 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7gn.large",
        "name": "c7gn.large",
        "ram": 4096,
    },
    "c7gn.medium": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7gn.medium",
            "instancesku": "JMMFJ8Y2RWTYNHYY",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "2 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "1",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7gn.medium",
        "name": "c7gn.medium",
        "ram": 2048,
    },
    "c7gn.xlarge": {
        "bandwidth": 40,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7gn.xlarge",
            "instancesku": "UV36UAT5A9U55K8X",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 40 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7gn.xlarge",
        "name": "c7gn.xlarge",
        "ram": 8192,
    },
    "c7i-flex.12xlarge": {
        "bandwidth": 18750,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 15000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7i-flex.12xlarge",
            "instancesku": "GSZ3E5T2PQ9RTMN8",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "Up to 18750 Megabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AMX; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7i-flex.12xlarge",
        "name": "c7i-flex.12xlarge",
        "ram": 98304,
    },
    "c7i-flex.16xlarge": {
        "bandwidth": 25000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7i-flex.16xlarge",
            "instancesku": "47ZF4F474QNTMXH3",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 25000 Megabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AMX; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7i-flex.16xlarge",
        "name": "c7i-flex.16xlarge",
        "ram": 131072,
    },
    "c7i-flex.2xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7i-flex.2xlarge",
            "instancesku": "UPFYFUG8AKV7JU8V",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AMX; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7i-flex.2xlarge",
        "name": "c7i-flex.2xlarge",
        "ram": 16384,
    },
    "c7i-flex.4xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7i-flex.4xlarge",
            "instancesku": "644ZZ3N8ZX2T6Q64",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AMX; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7i-flex.4xlarge",
        "name": "c7i-flex.4xlarge",
        "ram": 32768,
    },
    "c7i-flex.8xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7i-flex.8xlarge",
            "instancesku": "5PYKYM2ECW7NZ3XH",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AMX; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7i-flex.8xlarge",
        "name": "c7i-flex.8xlarge",
        "ram": 65536,
    },
    "c7i-flex.large": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7i-flex.large",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AMX; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "sa-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7i-flex.large",
        "name": "c7i-flex.large",
        "ram": 4096,
    },
    "c7i-flex.xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7i-flex.xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AMX; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7i-flex.xlarge",
        "name": "c7i-flex.xlarge",
        "ram": 8192,
    },
    "c7i.12xlarge": {
        "bandwidth": 18750,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "15000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7i.12xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "18750 Megabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "ap-southeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7i.12xlarge",
        "name": "c7i.12xlarge",
        "ram": 98304,
    },
    "c7i.16xlarge": {
        "bandwidth": 25000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7i.16xlarge",
            "instancesku": "GZ7MV6V72WV899QY",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "25000 Megabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "ap-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7i.16xlarge",
        "name": "c7i.16xlarge",
        "ram": 131072,
    },
    "c7i.24xlarge": {
        "bandwidth": 37500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "30000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7i.24xlarge",
            "instancesku": "WA98FAHPNWDDK979",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "37500 Megabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "eu-west-3",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7i.24xlarge",
        "name": "c7i.24xlarge",
        "ram": 196608,
    },
    "c7i.2xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7i.2xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7i.2xlarge",
        "name": "c7i.2xlarge",
        "ram": 16384,
    },
    "c7i.48xlarge": {
        "bandwidth": 50000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "40000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7i.48xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "50000 Megabit",
            "normalizationSizeFactor": "384",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "192",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7i.48xlarge",
        "name": "c7i.48xlarge",
        "ram": 393216,
    },
    "c7i.4xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7i.4xlarge",
            "instancesku": "WHKNSAQ832PRZBXT",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "us-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7i.4xlarge",
        "name": "c7i.4xlarge",
        "ram": 32768,
    },
    "c7i.8xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7i.8xlarge",
            "instancesku": "AWQZUQWQZPJN8RVW",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "12500 Megabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7i.8xlarge",
        "name": "c7i.8xlarge",
        "ram": 65536,
    },
    "c7i.large": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7i.large",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7i.large",
        "name": "c7i.large",
        "ram": 4096,
    },
    "c7i.xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c7i.xlarge",
            "instancesku": "QS2EVUNVY9DVPKDQ",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "ap-southeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "c7i.xlarge",
        "name": "c7i.xlarge",
        "ram": 8192,
    },
    "c8g.12xlarge": {
        "bandwidth": 22,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "15000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c8g.12xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "22.5 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "c8g.12xlarge",
        "name": "c8g.12xlarge",
        "ram": 98304,
    },
    "c8g.16xlarge": {
        "bandwidth": 30,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c8g.16xlarge",
            "instancesku": "MWCSPTWX23NYB7VQ",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "30 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "c8g.16xlarge",
        "name": "c8g.16xlarge",
        "ram": 131072,
    },
    "c8g.24xlarge": {
        "bandwidth": 40,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "30000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c8g.24xlarge",
            "instancesku": "36F49S5R44BPRKQ6",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "40 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "c8g.24xlarge",
        "name": "c8g.24xlarge",
        "ram": 196608,
    },
    "c8g.2xlarge": {
        "bandwidth": 15,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c8g.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 15 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "c8g.2xlarge",
        "name": "c8g.2xlarge",
        "ram": 16384,
    },
    "c8g.48xlarge": {
        "bandwidth": 50,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "40000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c8g.48xlarge",
            "instancesku": "PWHPJCNBRZJ5K579",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "384",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "192",
            "vpcnetworkingsupport": "true",
        },
        "id": "c8g.48xlarge",
        "name": "c8g.48xlarge",
        "ram": 393216,
    },
    "c8g.4xlarge": {
        "bandwidth": 15,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c8g.4xlarge",
            "instancesku": "5CJ8R29XEESN6RMA",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 15 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "c8g.4xlarge",
        "name": "c8g.4xlarge",
        "ram": 32768,
    },
    "c8g.8xlarge": {
        "bandwidth": 15,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c8g.8xlarge",
            "instancesku": "CV8NXT6RRMWZNU7Y",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "15 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "c8g.8xlarge",
        "name": "c8g.8xlarge",
        "ram": 65536,
    },
    "c8g.large": {
        "bandwidth": 12,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c8g.large",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 12.5 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "c8g.large",
        "name": "c8g.large",
        "ram": 4096,
    },
    "c8g.medium": {
        "bandwidth": 12,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c8g.medium",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "2 GiB",
            "networkPerformance": "Up to 12.5 Gigabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "1",
            "vpcnetworkingsupport": "true",
        },
        "id": "c8g.medium",
        "name": "c8g.medium",
        "ram": 2048,
    },
    "c8g.xlarge": {
        "bandwidth": 12,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "c8g.xlarge",
            "instancesku": "N36XWBCR85XKHG87",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 12.5 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "c8g.xlarge",
        "name": "c8g.xlarge",
        "ram": 8192,
    },
    "cc2.8xlarge": {
        "bandwidth": 10,
        "disk": 3360,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "No",
            "ecu": "88",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "cc2.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "60.5 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon E5-2670",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 840 SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "cc2.8xlarge",
        "name": "cc2.8xlarge",
        "ram": 61952,
    },
    "cr1.8xlarge": {
        "bandwidth": 10,
        "disk": 240,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "currentGeneration": "No",
            "dedicatedEbsThroughput": "400 Mbps",
            "ecu": "88",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "cr1.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "244 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon E5-2670",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 120 SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "cr1.8xlarge",
        "name": "cr1.8xlarge",
        "ram": 249856,
    },
    "d2.2xlarge": {
        "bandwidth": None,
        "disk": 12000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.4 GHz",
            "currentGeneration": "No",
            "dedicatedEbsThroughput": "1000 Mbps",
            "ecu": "28",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "d2.2xlarge",
            "instancesku": "MMEKDXY58RD4JWKV",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "61 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon E5-2676 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "ap-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "6 x 2000 HDD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "d2.2xlarge",
        "name": "d2.2xlarge",
        "ram": 62464,
    },
    "d2.4xlarge": {
        "bandwidth": None,
        "disk": 24000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.4 GHz",
            "currentGeneration": "No",
            "dedicatedEbsThroughput": "2000 Mbps",
            "ecu": "56",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "d2.4xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "122 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon E5-2676 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "eu-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "12 x 2000 HDD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "d2.4xlarge",
        "name": "d2.4xlarge",
        "ram": 124928,
    },
    "d2.8xlarge": {
        "bandwidth": 10,
        "disk": 48000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.4 GHz",
            "currentGeneration": "No",
            "dedicatedEbsThroughput": "4000 Mbps",
            "ecu": "116",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "d2.8xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "244 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon E5-2676 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "24 x 2000 HDD",
            "vcpu": "36",
            "vpcnetworkingsupport": "true",
        },
        "id": "d2.8xlarge",
        "name": "d2.8xlarge",
        "ram": 249856,
    },
    "d2.xlarge": {
        "bandwidth": None,
        "disk": 6000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.4 GHz",
            "currentGeneration": "No",
            "dedicatedEbsThroughput": "750 Mbps",
            "ecu": "14",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "d2.xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "30.5 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon E5-2676 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "3 x 2000 HDD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "d2.xlarge",
        "name": "d2.xlarge",
        "ram": 31232,
    },
    "d3.2xlarge": {
        "bandwidth": 15,
        "disk": 12000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2800 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "d3.2xlarge",
            "instancesku": "5TGGFNPG4M6GTUKU",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 15 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "6 x 2000 HDD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "d3.2xlarge",
        "name": "d3.2xlarge",
        "ram": 65536,
    },
    "d3.4xlarge": {
        "bandwidth": 15,
        "disk": 24000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "2800 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "d3.4xlarge",
            "instancesku": "4DVAYKJPKD2C77C6",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 15 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "12 x 2000 HDD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "d3.4xlarge",
        "name": "d3.4xlarge",
        "ram": 131072,
    },
    "d3.8xlarge": {
        "bandwidth": 25,
        "disk": 48000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "5000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "d3.8xlarge",
            "instancesku": "4KEDQSATUFFFFRGR",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "24 x 2000 HDD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "d3.8xlarge",
        "name": "d3.8xlarge",
        "ram": 262144,
    },
    "d3.xlarge": {
        "bandwidth": 15,
        "disk": 6000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2800 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "d3.xlarge",
            "instancesku": "E6KXGC3UNB34ACFA",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 15 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "3 x 2000 HDD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "d3.xlarge",
        "name": "d3.xlarge",
        "ram": 32768,
    },
    "d3en.12xlarge": {
        "bandwidth": 75,
        "disk": 336000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "d3en.12xlarge",
            "instancesku": "GANWKN2KAMKGJ986",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "75 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-southeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "24 x 14000 HDD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "d3en.12xlarge",
        "name": "d3en.12xlarge",
        "ram": 196608,
    },
    "d3en.2xlarge": {
        "bandwidth": 25,
        "disk": 56000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2800 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "d3en.2xlarge",
            "instancesku": "AW2HWMVHERHTXRQB",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-southeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 14000 HDD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "d3en.2xlarge",
        "name": "d3en.2xlarge",
        "ram": 32768,
    },
    "d3en.4xlarge": {
        "bandwidth": 25,
        "disk": 112000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "2800 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "d3en.4xlarge",
            "instancesku": "24U62KWPMYB8EJ5M",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "8 x 14000 HDD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "d3en.4xlarge",
        "name": "d3en.4xlarge",
        "ram": 65536,
    },
    "d3en.6xlarge": {
        "bandwidth": 40,
        "disk": 168000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "d3en.6xlarge",
            "instancesku": "VRB5UKVQRHGS88H2",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "40 Gigabit",
            "normalizationSizeFactor": "48",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "12 x 14000 HDD",
            "vcpu": "24",
            "vpcnetworkingsupport": "true",
        },
        "id": "d3en.6xlarge",
        "name": "d3en.6xlarge",
        "ram": 98304,
    },
    "d3en.8xlarge": {
        "bandwidth": 50,
        "disk": 224000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "5000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "d3en.8xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "16 x 14000 HDD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "d3en.8xlarge",
        "name": "d3en.8xlarge",
        "ram": 131072,
    },
    "d3en.xlarge": {
        "bandwidth": 25,
        "disk": 28000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2800 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "d3en.xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-southeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 14000 HDD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "d3en.xlarge",
        "name": "d3en.xlarge",
        "ram": 16384,
    },
    "dl1.24xlarge": {
        "bandwidth": 400,
        "disk": 4000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "19000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Machine Learning ASIC Instances",
            "instanceType": "dl1.24xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "400 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8275L",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 1000 GB NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "dl1.24xlarge",
        "name": "dl1.24xlarge",
        "ram": 786432,
    },
    "dl2q.24xlarge": {
        "bandwidth": 100,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "19000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "128 GiB",
            "instanceFamily": "Machine Learning ASIC Instances",
            "instanceType": "dl2q.24xlarge",
            "instancesku": "EQGU7KRQR4RA2DPN",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "dl2q.24xlarge",
        "name": "dl2q.24xlarge",
        "ram": 786432,
    },
    "f1.16xlarge": {
        "bandwidth": 25,
        "disk": 3760,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14000 Mbps",
            "ecu": "201",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "FPGA Instances",
            "instanceType": "f1.16xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "976 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 940 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "f1.16xlarge",
        "name": "f1.16xlarge",
        "ram": 999424,
    },
    "f1.2xlarge": {
        "bandwidth": 10,
        "disk": 470,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "1700 Mbps",
            "ecu": "31",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "FPGA Instances",
            "instanceType": "f1.2xlarge",
            "instancesku": "MXDPYTEVHNYM6QTA",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "122 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "eu-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 470 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "f1.2xlarge",
        "name": "f1.2xlarge",
        "ram": 124928,
    },
    "f1.4xlarge": {
        "bandwidth": 10,
        "disk": 940,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "58",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "FPGA Instances",
            "instanceType": "f1.4xlarge",
            "instancesku": "6YUJAEPDDF47NZUQ",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "244 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 940 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "f1.4xlarge",
        "name": "f1.4xlarge",
        "ram": 249856,
    },
    "f2.12xlarge": {
        "bandwidth": 25,
        "disk": 1900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.95 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "15000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "FPGA Instances",
            "instanceType": "f2.12xlarge",
            "instancesku": "J9RXNF7DMTERX3WW",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 950 NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "f2.12xlarge",
        "name": "f2.12xlarge",
        "ram": 524288,
    },
    "f2.48xlarge": {
        "bandwidth": 100,
        "disk": 7600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.95 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "60000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "FPGA Instances",
            "instanceType": "f2.48xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "2048 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "384",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "8 x 950 NVMe SSD",
            "vcpu": "192",
            "vpcnetworkingsupport": "true",
        },
        "id": "f2.48xlarge",
        "name": "f2.48xlarge",
        "ram": 2097152,
    },
    "f2.6xlarge": {
        "bandwidth": 12,
        "disk": 950,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.95 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "FPGA Instances",
            "instanceType": "f2.6xlarge",
            "instancesku": "Y3UUMR6W3G63ZRUW",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "12.5 Gigabit",
            "normalizationSizeFactor": "48",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 950 NVMe SSD",
            "vcpu": "24",
            "vpcnetworkingsupport": "true",
        },
        "id": "f2.6xlarge",
        "name": "f2.6xlarge",
        "ram": 262144,
    },
    "g2.2xlarge": {
        "bandwidth": None,
        "disk": 60,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "No",
            "ecu": "26",
            "enhancedNetworkingSupported": "No",
            "gpu": "1",
            "gpuMemory": "NA",
            "instanceFamily": "GPU instance",
            "instanceType": "g2.2xlarge",
            "instancesku": "GYTDXWQD9F7S5WN2",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "15 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon E5-2670 (Sandy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 60 SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "g2.2xlarge",
        "name": "g2.2xlarge",
        "ram": 15360,
    },
    "g2.8xlarge": {
        "bandwidth": None,
        "disk": 240,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "No",
            "ecu": "104",
            "enhancedNetworkingSupported": "No",
            "gpu": "4",
            "gpuMemory": "NA",
            "instanceFamily": "GPU instance",
            "instanceType": "g2.8xlarge",
            "instancesku": "GP54RNQ37B9N43MJ",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "60 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon E5-2670 (Sandy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 120 SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "g2.8xlarge",
        "name": "g2.8xlarge",
        "ram": 61440,
    },
    "g3.16xlarge": {
        "bandwidth": 20,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14000 Mbps",
            "ecu": "201",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "4",
            "gpuMemory": "32 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g3.16xlarge",
            "instancesku": "TFGPCCSSAM9DNFGT",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "488 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "g3.16xlarge",
        "name": "g3.16xlarge",
        "ram": 499712,
    },
    "g3.4xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "58",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "gpuMemory": "8 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g3.4xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "122 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "g3.4xlarge",
        "name": "g3.4xlarge",
        "ram": 124928,
    },
    "g3.8xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "97",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "2",
            "gpuMemory": "16 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g3.8xlarge",
            "instancesku": "5BWHVPBAG2VXEG2N",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "244 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "ap-southeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "g3.8xlarge",
        "name": "g3.8xlarge",
        "ram": 249856,
    },
    "g3s.xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "100 Mbps",
            "ecu": "13",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "gpuMemory": "8 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g3s.xlarge",
            "instancesku": "H33F2YFNF92GTBPB",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "30.5 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "g3s.xlarge",
        "name": "g3s.xlarge",
        "ram": 31232,
    },
    "g4ad.16xlarge": {
        "bandwidth": 25,
        "disk": 2400,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "6000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "4",
            "gpuMemory": "32 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g4ad.16xlarge",
            "instancesku": "WKWR9U6CWSY34FJZ",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2400 GB NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "g4ad.16xlarge",
        "name": "g4ad.16xlarge",
        "ram": 262144,
    },
    "g4ad.2xlarge": {
        "bandwidth": 10,
        "disk": 300,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "gpuMemory": "8 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g4ad.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "19.97394",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "300 GB NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "g4ad.2xlarge",
        "name": "g4ad.2xlarge",
        "ram": 32768,
    },
    "g4ad.4xlarge": {
        "bandwidth": 10,
        "disk": 600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "gpuMemory": "8 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g4ad.4xlarge",
            "instancesku": "9VMBVH3KEV82T8ZW",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "600 GB NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "g4ad.4xlarge",
        "name": "g4ad.4xlarge",
        "ram": 65536,
    },
    "g4ad.8xlarge": {
        "bandwidth": 15,
        "disk": 1200,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "2",
            "gpuMemory": "16 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g4ad.8xlarge",
            "instancesku": "7FPHMYTC7W39FTM5",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "15 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1200 GB NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "g4ad.8xlarge",
        "name": "g4ad.8xlarge",
        "ram": 131072,
    },
    "g4ad.xlarge": {
        "bandwidth": 10,
        "disk": 150,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "gpuMemory": "8 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g4ad.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "13.97113",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "eu-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "150 GB NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "g4ad.xlarge",
        "name": "g4ad.xlarge",
        "ram": 16384,
    },
    "g4dn.12xlarge": {
        "bandwidth": 50,
        "disk": 900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "4",
            "gpuMemory": "64 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g4dn.12xlarge",
            "instancesku": "TPYSQ64R6HTY3C2M",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "7.437262357",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "900 GB NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "g4dn.12xlarge",
        "name": "g4dn.12xlarge",
        "ram": 196608,
    },
    "g4dn.16xlarge": {
        "bandwidth": 50,
        "disk": 900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "gpuMemory": "16 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g4dn.16xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "8.273764259",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "900 GB NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "g4dn.16xlarge",
        "name": "g4dn.16xlarge",
        "ram": 262144,
    },
    "g4dn.2xlarge": {
        "bandwidth": 25,
        "disk": 225,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "gpuMemory": "16 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g4dn.2xlarge",
            "instancesku": "K24CAS4K6W4HJXN8",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "1.429657795",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "225 GB NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "g4dn.2xlarge",
        "name": "g4dn.2xlarge",
        "ram": 32768,
    },
    "g4dn.4xlarge": {
        "bandwidth": 25,
        "disk": 225,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "gpuMemory": "16 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g4dn.4xlarge",
            "instancesku": "YMZGCYKDK8626BJT",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "2.288973384",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "225 GB NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "g4dn.4xlarge",
        "name": "g4dn.4xlarge",
        "ram": 65536,
    },
    "g4dn.8xlarge": {
        "bandwidth": 50,
        "disk": 900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "gpuMemory": "16 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g4dn.8xlarge",
            "instancesku": "AE5GUKCAC8GSBSBD",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "4.136882129",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "900 GB NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "g4dn.8xlarge",
        "name": "g4dn.8xlarge",
        "ram": 131072,
    },
    "g4dn.xlarge": {
        "bandwidth": 25,
        "disk": 125,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "gpuMemory": "16 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g4dn.xlarge",
            "instancesku": "N5MXSK5M9ZYYBS9Y",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "1",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "sa-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "125 GB NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "g4dn.xlarge",
        "name": "g4dn.xlarge",
        "ram": 16384,
    },
    "g5.12xlarge": {
        "bandwidth": 40,
        "disk": 3800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "16000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "4",
            "gpuMemory": "96 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g5.12xlarge",
            "instancesku": "Z22N58R7N3GKA4C3",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "40 Gigabit",
            "normalizationSizeFactor": "5.638170974",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 3800 GB NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "g5.12xlarge",
        "name": "g5.12xlarge",
        "ram": 196608,
    },
    "g5.16xlarge": {
        "bandwidth": 25,
        "disk": 1900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "16000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "gpuMemory": "24 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g5.16xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "4.071570577",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1900 GB NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "g5.16xlarge",
        "name": "g5.16xlarge",
        "ram": 262144,
    },
    "g5.24xlarge": {
        "bandwidth": 50,
        "disk": 3800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "19000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "4",
            "gpuMemory": "96 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g5.24xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "8.095427435",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 3800 GB NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "g5.24xlarge",
        "name": "g5.24xlarge",
        "ram": 393216,
    },
    "g5.2xlarge": {
        "bandwidth": 10,
        "disk": 450,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "gpuMemory": "24 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g5.2xlarge",
            "instancesku": "3V2YRDDASNUNHK8Q",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "1.204771372",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "eu-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 450 GB NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "g5.2xlarge",
        "name": "g5.2xlarge",
        "ram": 32768,
    },
    "g5.48xlarge": {
        "bandwidth": 100,
        "disk": 7600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "19000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "8",
            "gpuMemory": "192 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g5.48xlarge",
            "instancesku": "DPNNR9TE888NYC83",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "16.190854871",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 3800 GB NVMe SSD",
            "vcpu": "192",
            "vpcnetworkingsupport": "true",
        },
        "id": "g5.48xlarge",
        "name": "g5.48xlarge",
        "ram": 786432,
    },
    "g5.4xlarge": {
        "bandwidth": 25,
        "disk": 600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "8000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "gpuMemory": "24 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g5.4xlarge",
            "instancesku": "VKEJ46YCBF7V8WVN",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "1.614314115",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "ap-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 600 GB NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "g5.4xlarge",
        "name": "g5.4xlarge",
        "ram": 65536,
    },
    "g5.8xlarge": {
        "bandwidth": 25,
        "disk": 900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "16000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "gpuMemory": "24 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g5.8xlarge",
            "instancesku": "G77E9FGSFGQ5XBQR",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "2.433399602",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 900 GB NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "g5.8xlarge",
        "name": "g5.8xlarge",
        "ram": 131072,
    },
    "g5.xlarge": {
        "bandwidth": 10,
        "disk": 250,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.8 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "gpuMemory": "24 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g5.xlarge",
            "instancesku": "RGX2TRWU27FJXUD4",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "1",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 250 GB NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "g5.xlarge",
        "name": "g5.xlarge",
        "ram": 16384,
    },
    "g5g.16xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "19000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "2",
            "gpuMemory": "32 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g5g.16xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "6.533333333",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "g5g.16xlarge",
        "name": "g5g.16xlarge",
        "ram": 131072,
    },
    "g5g.2xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "gpuMemory": "16 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g5g.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "1.323809524",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "g5g.2xlarge",
        "name": "g5g.2xlarge",
        "ram": 16384,
    },
    "g5g.4xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4750 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "gpuMemory": "16 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g5g.4xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "1.971428571",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "g5g.4xlarge",
        "name": "g5g.4xlarge",
        "ram": 32768,
    },
    "g5g.8xlarge": {
        "bandwidth": 12,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "9500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "gpuMemory": "16 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g5g.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 12 Gigabit",
            "normalizationSizeFactor": "3.266666667",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "g5g.8xlarge",
        "name": "g5g.8xlarge",
        "ram": 65536,
    },
    "g5g.xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "gpuMemory": "16 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g5g.xlarge",
            "instancesku": "GWKJ7SBX9EJ97KHP",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "1",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "g5g.xlarge",
        "name": "g5g.xlarge",
        "ram": 8192,
    },
    "g6.12xlarge": {
        "bandwidth": 40,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "4",
            "gpuMemory": "96 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g6.12xlarge",
            "instancesku": "DQTH69PA6TQYJEPB",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "40 Gigabit",
            "normalizationSizeFactor": "5.717693837",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 X 940 GB NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "g6.12xlarge",
        "name": "g6.12xlarge",
        "ram": 196608,
    },
    "g6.16xlarge": {
        "bandwidth": 25,
        "disk": 1880,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "gpuMemory": "24 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g6.16xlarge",
            "instancesku": "4V4FU9YKANXV8G2A",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "4.220675944",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 940 GB NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "g6.16xlarge",
        "name": "g6.16xlarge",
        "ram": 262144,
    },
    "g6.24xlarge": {
        "bandwidth": 50,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "30000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "4",
            "gpuMemory": "96 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g6.24xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "8.294234592",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 X 940 GB NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "g6.24xlarge",
        "name": "g6.24xlarge",
        "ram": 393216,
    },
    "g6.2xlarge": {
        "bandwidth": 10,
        "disk": 450,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 5000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "gpuMemory": "24 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g6.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "1.21471173",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 450 GB NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "g6.2xlarge",
        "name": "g6.2xlarge",
        "ram": 32768,
    },
    "g6.48xlarge": {
        "bandwidth": 100,
        "disk": 7520,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "60000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "8",
            "gpuMemory": "192 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g6.48xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "16.58846918",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "ap-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "8 x 940 NVMe SSD",
            "vcpu": "192",
            "vpcnetworkingsupport": "true",
        },
        "id": "g6.48xlarge",
        "name": "g6.48xlarge",
        "ram": 786432,
    },
    "g6.4xlarge": {
        "bandwidth": 25,
        "disk": 600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "8000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "gpuMemory": "24 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g6.4xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "1.644135189",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 600 GB NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "g6.4xlarge",
        "name": "g6.4xlarge",
        "ram": 65536,
    },
    "g6.8xlarge": {
        "bandwidth": 25,
        "disk": 900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "16000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "gpuMemory": "24 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g6.8xlarge",
            "instancesku": "U9KYRQSBZJZ7BEY5",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "2.502982107",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 450 GB NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "g6.8xlarge",
        "name": "g6.8xlarge",
        "ram": 131072,
    },
    "g6.xlarge": {
        "bandwidth": 10,
        "disk": 250,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 5000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "gpuMemory": "24 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g6.xlarge",
            "instancesku": "8YJMSK4Z67XAC3JS",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "1",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "eu-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 250 GB NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "g6.xlarge",
        "name": "g6.xlarge",
        "ram": 16384,
    },
    "g6e.12xlarge": {
        "bandwidth": 100,
        "disk": 3800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "4",
            "gpuMemory": "192 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g6e.12xlarge",
            "instancesku": "J5MZWTDZVU6CFEG9",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "5.638170974",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1900 GB NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "g6e.12xlarge",
        "name": "g6e.12xlarge",
        "ram": 393216,
    },
    "g6e.16xlarge": {
        "bandwidth": 35,
        "disk": 1900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "gpuMemory": "48 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g6e.16xlarge",
            "instancesku": "8RMS4UXRV9DKB4VU",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "35 Gigabit",
            "normalizationSizeFactor": "4.071570577",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1900 GB NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "g6e.16xlarge",
        "name": "g6e.16xlarge",
        "ram": 524288,
    },
    "g6e.24xlarge": {
        "bandwidth": 200,
        "disk": 3800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "30000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "4",
            "gpuMemory": "192 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g6e.24xlarge",
            "instancesku": "65MPARB6D8YCD2G7",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "200 Gigabit",
            "normalizationSizeFactor": "8.095427435",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1900 GB NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "g6e.24xlarge",
        "name": "g6e.24xlarge",
        "ram": 786432,
    },
    "g6e.2xlarge": {
        "bandwidth": 20,
        "disk": 450,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 5000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "gpuMemory": "48 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g6e.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 20 Gigabit",
            "normalizationSizeFactor": "1.204771372",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 450 GB NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "g6e.2xlarge",
        "name": "g6e.2xlarge",
        "ram": 65536,
    },
    "g6e.48xlarge": {
        "bandwidth": 400,
        "disk": 7600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "60000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "8",
            "gpuMemory": "384 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g6e.48xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "1536 GiB",
            "networkPerformance": "400 Gigabit",
            "normalizationSizeFactor": "16.19085487",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 1900 GB NVMe SSD",
            "vcpu": "192",
            "vpcnetworkingsupport": "true",
        },
        "id": "g6e.48xlarge",
        "name": "g6e.48xlarge",
        "ram": 1572864,
    },
    "g6e.4xlarge": {
        "bandwidth": 20,
        "disk": 600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "8000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "gpuMemory": "48 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g6e.4xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "1.614314115",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 600 GB NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "g6e.4xlarge",
        "name": "g6e.4xlarge",
        "ram": 131072,
    },
    "g6e.8xlarge": {
        "bandwidth": 25,
        "disk": 900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "16000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "gpuMemory": "48 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g6e.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "2.433399602",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 900 GB NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "g6e.8xlarge",
        "name": "g6e.8xlarge",
        "ram": 262144,
    },
    "g6e.xlarge": {
        "bandwidth": 20,
        "disk": 250,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 5000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "gpuMemory": "48 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "g6e.xlarge",
            "instancesku": "F6VQKF62RGG5EV6F",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 20 Gigabit",
            "normalizationSizeFactor": "1",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 250 GB NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "g6e.xlarge",
        "name": "g6e.xlarge",
        "ram": 32768,
    },
    "gr6.4xlarge": {
        "bandwidth": 25,
        "disk": 600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "8000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "gpuMemory": "24 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "gr6.4xlarge",
            "instancesku": "PG8KYKPFQ8JZ6KAD",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "1",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "eu-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 600 GB NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "gr6.4xlarge",
        "name": "gr6.4xlarge",
        "ram": 131072,
    },
    "gr6.8xlarge": {
        "bandwidth": 25,
        "disk": 900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "16000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "gpuMemory": "24 GB",
            "instanceFamily": "GPU instance",
            "instanceType": "gr6.8xlarge",
            "instancesku": "ARXY4W4A3M6DUEFT",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "1.589397089",
            "physicalProcessor": "AMD EPYC 7R32",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 450 GB NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "gr6.8xlarge",
        "name": "gr6.8xlarge",
        "ram": 262144,
    },
    "h1.16xlarge": {
        "bandwidth": 25,
        "disk": 16000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14000 Mbps",
            "ecu": "201",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "h1.16xlarge",
            "instancesku": "QZQ8KAA8X77EH39R",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "8 x 2000 HDD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "h1.16xlarge",
        "name": "h1.16xlarge",
        "ram": 262144,
    },
    "h1.2xlarge": {
        "bandwidth": 10,
        "disk": 2000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "1750 Mbps",
            "ecu": "31",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "h1.2xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 2000 HDD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "h1.2xlarge",
        "name": "h1.2xlarge",
        "ram": 32768,
    },
    "h1.4xlarge": {
        "bandwidth": 10,
        "disk": 4000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "58",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "h1.4xlarge",
            "instancesku": "DA8WABW4PXZEM9CX",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 2000 HDD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "h1.4xlarge",
        "name": "h1.4xlarge",
        "ram": 65536,
    },
    "h1.8xlarge": {
        "bandwidth": 10,
        "disk": 8000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "97",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "h1.8xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 2000 HDD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "h1.8xlarge",
        "name": "h1.8xlarge",
        "ram": 131072,
    },
    "hpc6a.48xlarge": {
        "bandwidth": 100,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.95 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2085 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "hpc6a.48xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "384",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "hpc6a.48xlarge",
        "name": "hpc6a.48xlarge",
        "ram": 393216,
    },
    "hpc6id.32xlarge": {
        "bandwidth": 200,
        "disk": 15200,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2085 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "hpc6id.32xlarge",
            "instancesku": "BFH769F89NTNBEEA",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "1024 GiB",
            "networkPerformance": "200 Gigabit",
            "normalizationSizeFactor": "256",
            "physicalProcessor": "Intel Xeon Scalable (Icelake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-3",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 3800 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "hpc6id.32xlarge",
        "name": "hpc6id.32xlarge",
        "ram": 1048576,
    },
    "hpc7a.12xlarge": {
        "bandwidth": 300,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "87 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "hpc7a.12xlarge",
            "instancesku": "DXRKH4UU6G9GDZ8Z",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "300 Gigabit",
            "normalizationSizeFactor": "384",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "24",
            "vpcnetworkingsupport": "true",
        },
        "id": "hpc7a.12xlarge",
        "name": "hpc7a.12xlarge",
        "ram": 786432,
    },
    "hpc7a.24xlarge": {
        "bandwidth": 300,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "87 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "hpc7a.24xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "300 Gigabit",
            "normalizationSizeFactor": "384",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "eu-west-3",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "hpc7a.24xlarge",
        "name": "hpc7a.24xlarge",
        "ram": 786432,
    },
    "hpc7a.48xlarge": {
        "bandwidth": 300,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "87 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "hpc7a.48xlarge",
            "instancesku": "HYBMTYWPBDAV2BTJ",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "300 Gigabit",
            "normalizationSizeFactor": "384",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "eu-west-3",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "hpc7a.48xlarge",
        "name": "hpc7a.48xlarge",
        "ram": 786432,
    },
    "hpc7a.96xlarge": {
        "bandwidth": 300,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "87 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "hpc7a.96xlarge",
            "instancesku": "DP7DTFB5J8SYZRPC",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "300 Gigabit",
            "normalizationSizeFactor": "384",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "192",
            "vpcnetworkingsupport": "true",
        },
        "id": "hpc7a.96xlarge",
        "name": "hpc7a.96xlarge",
        "ram": 786432,
    },
    "hpc7g.16xlarge": {
        "bandwidth": 200,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "87 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "hpc7g.16xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "200 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "hpc7g.16xlarge",
        "name": "hpc7g.16xlarge",
        "ram": 131072,
    },
    "hpc7g.4xlarge": {
        "bandwidth": 200,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "87 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "hpc7g.4xlarge",
            "instancesku": "4BRATKTQX2BNMYX8",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "200 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "hpc7g.4xlarge",
        "name": "hpc7g.4xlarge",
        "ram": 131072,
    },
    "hpc7g.8xlarge": {
        "bandwidth": 200,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "87 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Compute optimized",
            "instanceType": "hpc7g.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "200 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "hpc7g.8xlarge",
        "name": "hpc7g.8xlarge",
        "ram": 131072,
    },
    "hs1.8xlarge": {
        "bandwidth": 10,
        "disk": 48000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2 GHz",
            "currentGeneration": "No",
            "ecu": "35",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "hs1.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "117 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon E5-2650",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "24 x 2000 HDD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "hs1.8xlarge",
        "name": "hs1.8xlarge",
        "ram": 119808,
    },
    "i2.2xlarge": {
        "bandwidth": None,
        "disk": 1600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "No",
            "dedicatedEbsThroughput": "1500 Mbps",
            "ecu": "27",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i2.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "61 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon E5-2670 v2 (Ivy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "regionCode": "us-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 800 SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "i2.2xlarge",
        "name": "i2.2xlarge",
        "ram": 62464,
    },
    "i2.4xlarge": {
        "bandwidth": None,
        "disk": 3200,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "No",
            "dedicatedEbsThroughput": "1500 Mbps",
            "ecu": "53",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i2.4xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "122 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon E5-2670 v2 (Ivy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 800 SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "i2.4xlarge",
        "name": "i2.4xlarge",
        "ram": 124928,
    },
    "i2.8xlarge": {
        "bandwidth": 10,
        "disk": 6400,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "No",
            "dedicatedEbsThroughput": "1500 Mbps",
            "ecu": "104",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i2.8xlarge",
            "instancesku": "2UP3NPVY4NG62NY2",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "244 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon E5-2670 v2 (Ivy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "8 x 800 SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "i2.8xlarge",
        "name": "i2.8xlarge",
        "ram": 249856,
    },
    "i2.large": {
        "bandwidth": None,
        "disk": 800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "No",
            "ecu": "7",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i2.large",
            "instancesku": "WURYYYF99XGKXXRH",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "15 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon E5-2670 v2 (Ivy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "regionCode": "eu-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 800 SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "i2.large",
        "name": "i2.large",
        "ram": 15360,
    },
    "i2.xlarge": {
        "bandwidth": None,
        "disk": 800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "No",
            "dedicatedEbsThroughput": "750 Mbps",
            "ecu": "14",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i2.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "30.5 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon E5-2670 v2 (Ivy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "regionCode": "ap-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 800 SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "i2.xlarge",
        "name": "i2.xlarge",
        "ram": 31232,
    },
    "i3.16xlarge": {
        "bandwidth": 20,
        "disk": 15200,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14000 Mbps",
            "ecu": "201",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i3.16xlarge",
            "instancesku": "QVUP8HT625VGM28B",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "488 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "8 x 1900 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "i3.16xlarge",
        "name": "i3.16xlarge",
        "ram": 499712,
    },
    "i3.2xlarge": {
        "bandwidth": 10,
        "disk": 1900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "1750 Mbps",
            "ecu": "31",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i3.2xlarge",
            "instancesku": "344DF3TUKYKVUUYB",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "61 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1900 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "i3.2xlarge",
        "name": "i3.2xlarge",
        "ram": 62464,
    },
    "i3.4xlarge": {
        "bandwidth": 10,
        "disk": 3800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "58",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i3.4xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "122 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1900 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "i3.4xlarge",
        "name": "i3.4xlarge",
        "ram": 124928,
    },
    "i3.8xlarge": {
        "bandwidth": 10,
        "disk": 7600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "97",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i3.8xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "244 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 1900 NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "i3.8xlarge",
        "name": "i3.8xlarge",
        "ram": 249856,
    },
    "i3.large": {
        "bandwidth": 10,
        "disk": 475,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "425 Mbps",
            "ecu": "8",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i3.large",
            "instancesku": "T2R7DYXSCP5QZZGU",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "15.25 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 475 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "i3.large",
        "name": "i3.large",
        "ram": 15616,
    },
    "i3.xlarge": {
        "bandwidth": 10,
        "disk": 950,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "850 Mbps",
            "ecu": "16",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i3.xlarge",
            "instancesku": "DQBUA3WF2ZKB94XE",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "30.5 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 950 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "i3.xlarge",
        "name": "i3.xlarge",
        "ram": 31232,
    },
    "i3en.12xlarge": {
        "bandwidth": 50,
        "disk": 30000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "168",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i3en.12xlarge",
            "instancesku": "8RE9VN3MCA27FAPN",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 7500 NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "i3en.12xlarge",
        "name": "i3en.12xlarge",
        "ram": 393216,
    },
    "i3en.24xlarge": {
        "bandwidth": 100,
        "disk": 60000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14000 Mbps",
            "ecu": "337",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i3en.24xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "8 x 7500 NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "i3en.24xlarge",
        "name": "i3en.24xlarge",
        "ram": 786432,
    },
    "i3en.2xlarge": {
        "bandwidth": 25,
        "disk": 5000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "37",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i3en.2xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-3",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 2500 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "i3en.2xlarge",
        "name": "i3en.2xlarge",
        "ram": 65536,
    },
    "i3en.3xlarge": {
        "bandwidth": 25,
        "disk": 7500,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i3en.3xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "24",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 7500 NVMe SSD",
            "vcpu": "12",
            "vpcnetworkingsupport": "true",
        },
        "id": "i3en.3xlarge",
        "name": "i3en.3xlarge",
        "ram": 98304,
    },
    "i3en.6xlarge": {
        "bandwidth": 25,
        "disk": 15000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i3en.6xlarge",
            "instancesku": "JFUSUE6PEJ7BEWFW",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "48",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 7500 NVMe SSD",
            "vcpu": "24",
            "vpcnetworkingsupport": "true",
        },
        "id": "i3en.6xlarge",
        "name": "i3en.6xlarge",
        "ram": 196608,
    },
    "i3en.large": {
        "bandwidth": 25,
        "disk": 1250,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "10",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i3en.large",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1250 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "i3en.large",
        "name": "i3en.large",
        "ram": 16384,
    },
    "i3en.xlarge": {
        "bandwidth": 25,
        "disk": 2500,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i3en.xlarge",
            "instancesku": "KFX9MBMD2FPV9EAG",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "sa-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 2500 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "i3en.xlarge",
        "name": "i3en.xlarge",
        "ram": 32768,
    },
    "i4g.16xlarge": {
        "bandwidth": 37500,
        "disk": 15000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i4g.16xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "37500 Megabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 3750 SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "i4g.16xlarge",
        "name": "i4g.16xlarge",
        "ram": 524288,
    },
    "i4g.2xlarge": {
        "bandwidth": 12,
        "disk": 1875,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i4g.2xlarge",
            "instancesku": "6V9FAVBEUJNF7Y5N",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 12 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-southeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1875 SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "i4g.2xlarge",
        "name": "i4g.2xlarge",
        "ram": 65536,
    },
    "i4g.4xlarge": {
        "bandwidth": 25,
        "disk": 3750,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i4g.4xlarge",
            "instancesku": "QMEDNTDHFBJ5TBVK",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-southeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 3750 SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "i4g.4xlarge",
        "name": "i4g.4xlarge",
        "ram": 131072,
    },
    "i4g.8xlarge": {
        "bandwidth": 18750,
        "disk": 7500,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i4g.8xlarge",
            "instancesku": "88CKZ2RR64QV2SEC",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "18750 Megabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 3750 SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "i4g.8xlarge",
        "name": "i4g.8xlarge",
        "ram": 262144,
    },
    "i4g.large": {
        "bandwidth": 10,
        "disk": 468,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i4g.large",
            "instancesku": "ECPX7S99F7RP3PWS",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 468 SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "i4g.large",
        "name": "i4g.large",
        "ram": 16384,
    },
    "i4g.xlarge": {
        "bandwidth": 10,
        "disk": 937,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i4g.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 937 SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "i4g.xlarge",
        "name": "i4g.xlarge",
        "ram": 32768,
    },
    "i4i.12xlarge": {
        "bandwidth": 28125,
        "disk": 11250,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "15000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i4i.12xlarge",
            "instancesku": "5DP3MFBEZAWKNNU4",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "28125 Megabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-3",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "3 x 3750 NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "i4i.12xlarge",
        "name": "i4i.12xlarge",
        "ram": 393216,
    },
    "i4i.16xlarge": {
        "bandwidth": 35000,
        "disk": 15000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i4i.16xlarge",
            "instancesku": "5PHSTUESTSJBSFAK",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "35000 Megabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 3750 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "i4i.16xlarge",
        "name": "i4i.16xlarge",
        "ram": 524288,
    },
    "i4i.24xlarge": {
        "bandwidth": 56250,
        "disk": 22500,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "30000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i4i.24xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "56250 Megabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "6 x 3750 NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "i4i.24xlarge",
        "name": "i4i.24xlarge",
        "ram": 786432,
    },
    "i4i.2xlarge": {
        "bandwidth": 12,
        "disk": 1875,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i4i.2xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 12 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1875 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "i4i.2xlarge",
        "name": "i4i.2xlarge",
        "ram": 65536,
    },
    "i4i.32xlarge": {
        "bandwidth": 75000,
        "disk": 30000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "40000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i4i.32xlarge",
            "instancesku": "44CRDAV8AJ9ZMF4W",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "1024 GiB",
            "networkPerformance": "75000 Megabit",
            "normalizationSizeFactor": "256",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "8 x 3750 SSD",
            "vcpu": "128",
            "vpcnetworkingsupport": "true",
        },
        "id": "i4i.32xlarge",
        "name": "i4i.32xlarge",
        "ram": 1048576,
    },
    "i4i.4xlarge": {
        "bandwidth": 25,
        "disk": 3750,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i4i.4xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 3750 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "i4i.4xlarge",
        "name": "i4i.4xlarge",
        "ram": 131072,
    },
    "i4i.8xlarge": {
        "bandwidth": 18750,
        "disk": 7500,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i4i.8xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "18750 Megabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 3750 NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "i4i.8xlarge",
        "name": "i4i.8xlarge",
        "ram": 262144,
    },
    "i4i.large": {
        "bandwidth": 10,
        "disk": 468,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i4i.large",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 468 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "i4i.large",
        "name": "i4i.large",
        "ram": 16384,
    },
    "i4i.xlarge": {
        "bandwidth": 10,
        "disk": 937,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i4i.xlarge",
            "instancesku": "DU36AKRSXVPX42VR",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 937 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "i4i.xlarge",
        "name": "i4i.xlarge",
        "ram": 32768,
    },
    "i7ie.12xlarge": {
        "bandwidth": 50,
        "disk": 30000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "15000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i7ie.12xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "Up to 50 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Scalable (Emerald Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AMX; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 7500 NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "i7ie.12xlarge",
        "name": "i7ie.12xlarge",
        "ram": 393216,
    },
    "i7ie.18xlarge": {
        "bandwidth": 75,
        "disk": 45000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "22500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i7ie.18xlarge",
            "instancesku": "DUZ3BPZ8C9RT7UHP",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "576 GiB",
            "networkPerformance": "Up to 75 Gigabit",
            "normalizationSizeFactor": "144",
            "physicalProcessor": "Intel Xeon Scalable (Emerald Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AMX; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "6 x 7500 NVMe SSD",
            "vcpu": "72",
            "vpcnetworkingsupport": "true",
        },
        "id": "i7ie.18xlarge",
        "name": "i7ie.18xlarge",
        "ram": 589824,
    },
    "i7ie.24xlarge": {
        "bandwidth": 100,
        "disk": 60000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "30000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i7ie.24xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "Up to 100 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Scalable (Emerald Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AMX; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "8 x 7500 NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "i7ie.24xlarge",
        "name": "i7ie.24xlarge",
        "ram": 786432,
    },
    "i7ie.2xlarge": {
        "bandwidth": 25,
        "disk": 5000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i7ie.2xlarge",
            "instancesku": "6G4AS3J58SHH7Q39",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Scalable (Emerald Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AMX; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 2500 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "i7ie.2xlarge",
        "name": "i7ie.2xlarge",
        "ram": 65536,
    },
    "i7ie.3xlarge": {
        "bandwidth": 25,
        "disk": 7500,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i7ie.3xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "24",
            "physicalProcessor": "Intel Xeon Scalable (Emerald Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AMX; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 7500 NVMe SSD",
            "vcpu": "12",
            "vpcnetworkingsupport": "true",
        },
        "id": "i7ie.3xlarge",
        "name": "i7ie.3xlarge",
        "ram": 98304,
    },
    "i7ie.48xlarge": {
        "bandwidth": 100,
        "disk": 120000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "60000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i7ie.48xlarge",
            "instancesku": "K2Z4WKVRUR836V25",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "1536 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "384",
            "physicalProcessor": "Intel Xeon Scalable (Emerald Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AMX; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "16 x 7500 NVMe SSD",
            "vcpu": "192",
            "vpcnetworkingsupport": "true",
        },
        "id": "i7ie.48xlarge",
        "name": "i7ie.48xlarge",
        "ram": 1572864,
    },
    "i7ie.6xlarge": {
        "bandwidth": 25,
        "disk": 15000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i7ie.6xlarge",
            "instancesku": "BGVYBGUGSTSR3XZJ",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "48",
            "physicalProcessor": "Intel Xeon Scalable (Emerald Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AMX; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 7500 NVMe SSD",
            "vcpu": "24",
            "vpcnetworkingsupport": "true",
        },
        "id": "i7ie.6xlarge",
        "name": "i7ie.6xlarge",
        "ram": 196608,
    },
    "i7ie.large": {
        "bandwidth": 25,
        "disk": 1250,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i7ie.large",
            "instancesku": "VTAGUBST7RFQV7SE",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Scalable (Emerald Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AMX; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1250 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "i7ie.large",
        "name": "i7ie.large",
        "ram": 16384,
    },
    "i7ie.xlarge": {
        "bandwidth": 25,
        "disk": 2500,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i7ie.xlarge",
            "instancesku": "GSC68K245DZH7G3P",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Scalable (Emerald Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AMX; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 2500 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "i7ie.xlarge",
        "name": "i7ie.xlarge",
        "ram": 32768,
    },
    "i8g.12xlarge": {
        "bandwidth": 28,
        "disk": 11250,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "15000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i8g.12xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "Up to 28.125 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "3 x 3750GB",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "i8g.12xlarge",
        "name": "i8g.12xlarge",
        "ram": 393216,
    },
    "i8g.16xlarge": {
        "bandwidth": 37,
        "disk": 15000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i8g.16xlarge",
            "instancesku": "MPZ5GEWS9NFBFXYX",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "Up to 37.5 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 3750GB",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "i8g.16xlarge",
        "name": "i8g.16xlarge",
        "ram": 524288,
    },
    "i8g.24xlarge": {
        "bandwidth": 56,
        "disk": 22500,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "30000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i8g.24xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "Up to 56.25 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "6 x 3750GB",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "i8g.24xlarge",
        "name": "i8g.24xlarge",
        "ram": 786432,
    },
    "i8g.2xlarge": {
        "bandwidth": 12,
        "disk": 1875,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i8g.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 12 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1875GB",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "i8g.2xlarge",
        "name": "i8g.2xlarge",
        "ram": 65536,
    },
    "i8g.4xlarge": {
        "bandwidth": 25,
        "disk": 3750,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i8g.4xlarge",
            "instancesku": "TJX3WPXPT52JFXS6",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 3750GB",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "i8g.4xlarge",
        "name": "i8g.4xlarge",
        "ram": 131072,
    },
    "i8g.8xlarge": {
        "bandwidth": 25,
        "disk": 7500,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i8g.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 3750GB",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "i8g.8xlarge",
        "name": "i8g.8xlarge",
        "ram": 262144,
    },
    "i8g.large": {
        "bandwidth": 10,
        "disk": 468,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i8g.large",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 468GB",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "i8g.large",
        "name": "i8g.large",
        "ram": 16384,
    },
    "i8g.xlarge": {
        "bandwidth": 10,
        "disk": 937,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "i8g.xlarge",
            "instancesku": "9WTCK83TR9JKZQF3",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 937GB",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "i8g.xlarge",
        "name": "i8g.xlarge",
        "ram": 32768,
    },
    "im4gn.16xlarge": {
        "bandwidth": 100,
        "disk": 30000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "38000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "im4gn.16xlarge",
            "instancesku": "78JF8766XR25JQ4E",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 7500 SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "im4gn.16xlarge",
        "name": "im4gn.16xlarge",
        "ram": 262144,
    },
    "im4gn.2xlarge": {
        "bandwidth": 25,
        "disk": 3750,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 9500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "im4gn.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 3750 SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "im4gn.2xlarge",
        "name": "im4gn.2xlarge",
        "ram": 32768,
    },
    "im4gn.4xlarge": {
        "bandwidth": 25,
        "disk": 7500,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "9500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "im4gn.4xlarge",
            "instancesku": "GJU6C8CQMRJKD374",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 7500 SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "im4gn.4xlarge",
        "name": "im4gn.4xlarge",
        "ram": 65536,
    },
    "im4gn.8xlarge": {
        "bandwidth": 50,
        "disk": 15000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "19000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "im4gn.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 7500 SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "im4gn.8xlarge",
        "name": "im4gn.8xlarge",
        "ram": 131072,
    },
    "im4gn.large": {
        "bandwidth": 25,
        "disk": 937,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 9500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "im4gn.large",
            "instancesku": "PQ6Y8UKZTPWYA3CM",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-3",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 937 SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "im4gn.large",
        "name": "im4gn.large",
        "ram": 8192,
    },
    "im4gn.xlarge": {
        "bandwidth": 25,
        "disk": 1875,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 9500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "im4gn.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1875 SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "im4gn.xlarge",
        "name": "im4gn.xlarge",
        "ram": 16384,
    },
    "inf1.24xlarge": {
        "bandwidth": 100,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Machine Learning ASIC Instances",
            "instanceType": "inf1.24xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "NA",
            "physicalProcessor": "Intel Xeon Platinum 8275CL (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "inf1.24xlarge",
        "name": "inf1.24xlarge",
        "ram": 196608,
    },
    "inf1.2xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "875 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Machine Learning ASIC Instances",
            "instanceType": "inf1.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "NA",
            "physicalProcessor": "Intel Xeon Platinum 8275CL (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "inf1.2xlarge",
        "name": "inf1.2xlarge",
        "ram": 16384,
    },
    "inf1.6xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Machine Learning ASIC Instances",
            "instanceType": "inf1.6xlarge",
            "instancesku": "JE9QJJXKVRAHGWSA",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "48 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "NA",
            "physicalProcessor": "Intel Xeon Platinum 8275CL (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "24",
            "vpcnetworkingsupport": "true",
        },
        "id": "inf1.6xlarge",
        "name": "inf1.6xlarge",
        "ram": 49152,
    },
    "inf1.xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "875 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Machine Learning ASIC Instances",
            "instanceType": "inf1.xlarge",
            "instancesku": "26TMKJU38MB7PCPQ",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "NA",
            "physicalProcessor": "Intel Xeon Platinum 8275CL (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "inf1.xlarge",
        "name": "inf1.xlarge",
        "ram": 8192,
    },
    "inf2.24xlarge": {
        "bandwidth": 50,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.95 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "30 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "192 GB",
            "instanceFamily": "Machine Learning ASIC Instances",
            "instanceType": "inf2.24xlarge",
            "instancesku": "KDEF377GZKPPYTN8",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "8.560582329",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "sa-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "inf2.24xlarge",
        "name": "inf2.24xlarge",
        "ram": 393216,
    },
    "inf2.48xlarge": {
        "bandwidth": 100,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.95 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "60 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "384 GB",
            "instanceFamily": "Machine Learning ASIC Instances",
            "instanceType": "inf2.48xlarge",
            "instancesku": "XXQA8QTRGVP45XRD",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "17.12116466",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "192",
            "vpcnetworkingsupport": "true",
        },
        "id": "inf2.48xlarge",
        "name": "inf2.48xlarge",
        "ram": 786432,
    },
    "inf2.8xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.95 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "32 GB",
            "instanceFamily": "Machine Learning ASIC Instances",
            "instanceType": "inf2.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "2.595436335",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "inf2.8xlarge",
        "name": "inf2.8xlarge",
        "ram": 131072,
    },
    "inf2.xlarge": {
        "bandwidth": 15,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.95 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "32 GB",
            "instanceFamily": "Machine Learning ASIC Instances",
            "instanceType": "inf2.xlarge",
            "instancesku": "V4BCUZVAZY7AJT2E",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 15 Gigabit",
            "normalizationSizeFactor": "1",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "inf2.xlarge",
        "name": "inf2.xlarge",
        "ram": 16384,
    },
    "is4gen.2xlarge": {
        "bandwidth": 25,
        "disk": 7500,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 9500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "is4gen.2xlarge",
            "instancesku": "NY5C5H7P4B6UW9HK",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "48 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-3",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 7500 SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "is4gen.2xlarge",
        "name": "is4gen.2xlarge",
        "ram": 49152,
    },
    "is4gen.4xlarge": {
        "bandwidth": 25,
        "disk": 15000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "9500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "is4gen.4xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 7500 SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "is4gen.4xlarge",
        "name": "is4gen.4xlarge",
        "ram": 98304,
    },
    "is4gen.8xlarge": {
        "bandwidth": 50,
        "disk": 30000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "19000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "is4gen.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 7500 SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "is4gen.8xlarge",
        "name": "is4gen.8xlarge",
        "ram": 196608,
    },
    "is4gen.large": {
        "bandwidth": 25,
        "disk": 1875,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 9500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "is4gen.large",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "12 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1875 SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "is4gen.large",
        "name": "is4gen.large",
        "ram": 12288,
    },
    "is4gen.medium": {
        "bandwidth": 25,
        "disk": 937,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 9500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "is4gen.medium",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "6 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 937 SSD",
            "vcpu": "1",
            "vpcnetworkingsupport": "true",
        },
        "id": "is4gen.medium",
        "name": "is4gen.medium",
        "ram": 6144,
    },
    "is4gen.xlarge": {
        "bandwidth": 25,
        "disk": 3750,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 9500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Storage optimized",
            "instanceType": "is4gen.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "24 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 3750 SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "is4gen.xlarge",
        "name": "is4gen.xlarge",
        "ram": 24576,
    },
    "m1.large": {
        "bandwidth": None,
        "disk": 840,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "currentGeneration": "No",
            "ecu": "4",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m1.large",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "7.5 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-southeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 420 SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "m1.large",
        "name": "m1.large",
        "ram": 7680,
    },
    "m1.medium": {
        "bandwidth": None,
        "disk": 410,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "currentGeneration": "No",
            "ecu": "2",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m1.medium",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "3.75 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "32-bit or 64-bit",
            "regionCode": "sa-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 410 SSD",
            "vcpu": "1",
            "vpcnetworkingsupport": "true",
        },
        "id": "m1.medium",
        "name": "m1.medium",
        "ram": 3840,
    },
    "m1.small": {
        "bandwidth": None,
        "disk": 160,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "currentGeneration": "No",
            "ecu": "1",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m1.small",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "1.7 GiB",
            "networkPerformance": "Low",
            "normalizationSizeFactor": "1",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "32-bit or 64-bit",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 160 SSD",
            "vcpu": "1",
            "vpcnetworkingsupport": "true",
        },
        "id": "m1.small",
        "name": "m1.small",
        "ram": 1740,
    },
    "m1.xlarge": {
        "bandwidth": None,
        "disk": 1680,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "currentGeneration": "No",
            "ecu": "8",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m1.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "15 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 420 SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "m1.xlarge",
        "name": "m1.xlarge",
        "ram": 15360,
    },
    "m2.2xlarge": {
        "bandwidth": None,
        "disk": 850,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "currentGeneration": "No",
            "ecu": "13",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "m2.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "34.2 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 850 SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "m2.2xlarge",
        "name": "m2.2xlarge",
        "ram": 35020,
    },
    "m2.4xlarge": {
        "bandwidth": None,
        "disk": 1680,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "currentGeneration": "No",
            "ecu": "26",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "m2.4xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "68.4 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 840 SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "m2.4xlarge",
        "name": "m2.4xlarge",
        "ram": 70041,
    },
    "m2.xlarge": {
        "bandwidth": None,
        "disk": 420,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "currentGeneration": "No",
            "ecu": "6.5",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "m2.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "17.1 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "64-bit",
            "regionCode": "sa-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 420 SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "m2.xlarge",
        "name": "m2.xlarge",
        "ram": 17510,
    },
    "m3.2xlarge": {
        "bandwidth": None,
        "disk": 160,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "No",
            "dedicatedEbsThroughput": "1000 Mbps",
            "ecu": "26",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m3.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "30 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon E5-2670 v2 (Ivy Bridge/Sandy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 80 SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "m3.2xlarge",
        "name": "m3.2xlarge",
        "ram": 30720,
    },
    "m3.large": {
        "bandwidth": None,
        "disk": 32,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "No",
            "ecu": "6.5",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m3.large",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "7.5 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon E5-2670 v2 (Ivy Bridge/Sandy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "regionCode": "sa-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 32 SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "m3.large",
        "name": "m3.large",
        "ram": 7680,
    },
    "m3.medium": {
        "bandwidth": None,
        "disk": 4,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "No",
            "ecu": "3",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m3.medium",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "3.75 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "Intel Xeon E5-2670 v2 (Ivy Bridge/Sandy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "regionCode": "ap-southeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 4 SSD",
            "vcpu": "1",
            "vpcnetworkingsupport": "true",
        },
        "id": "m3.medium",
        "name": "m3.medium",
        "ram": 3840,
    },
    "m3.xlarge": {
        "bandwidth": None,
        "disk": 80,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "No",
            "dedicatedEbsThroughput": "500 Mbps",
            "ecu": "13",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m3.xlarge",
            "instancesku": "JCFYMXQEQ3BA57K3",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "15 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon E5-2670 v2 (Ivy Bridge/Sandy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "regionCode": "ap-southeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 40 SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "m3.xlarge",
        "name": "m3.xlarge",
        "ram": 15360,
    },
    "m4.10xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4000 Mbps",
            "ecu": "124.5",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m4.10xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "160 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "80",
            "physicalProcessor": "Intel Xeon E5-2676 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "ap-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "40",
            "vpcnetworkingsupport": "true",
        },
        "id": "m4.10xlarge",
        "name": "m4.10xlarge",
        "ram": 163840,
    },
    "m4.16xlarge": {
        "bandwidth": 20,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "188",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m4.16xlarge",
            "instancesku": "9RRW6TTJ2K77NSDQ",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "m4.16xlarge",
        "name": "m4.16xlarge",
        "ram": 262144,
    },
    "m4.2xlarge": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "1000 Mbps",
            "ecu": "26",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m4.2xlarge",
            "instancesku": "4DP9JW528SDT7HRV",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon E5-2676 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "m4.2xlarge",
        "name": "m4.2xlarge",
        "ram": 32768,
    },
    "m4.4xlarge": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "2000 Mbps",
            "ecu": "53.5",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m4.4xlarge",
            "instancesku": "B5A2KQPVUB48X3ES",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon E5-2676 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "m4.4xlarge",
        "name": "m4.4xlarge",
        "ram": 65536,
    },
    "m4.large": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "450 Mbps",
            "ecu": "6.5",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m4.large",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon E5-2676 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "sa-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "m4.large",
        "name": "m4.large",
        "ram": 8192,
    },
    "m4.xlarge": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "750 Mbps",
            "ecu": "13",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m4.xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon E5-2676 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "m4.xlarge",
        "name": "m4.xlarge",
        "ram": 16384,
    },
    "m5.12xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "168",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5.12xlarge",
            "instancesku": "5SY86D6JMSFHE3FJ",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5.12xlarge",
        "name": "m5.12xlarge",
        "ram": 196608,
    },
    "m5.16xlarge": {
        "bandwidth": 20,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "256",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5.16xlarge",
            "instancesku": "ZXZ774GKQ54RUX72",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5.16xlarge",
        "name": "m5.16xlarge",
        "ram": 262144,
    },
    "m5.24xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "12000 Mbps",
            "ecu": "337",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5.24xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "sa-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5.24xlarge",
        "name": "m5.24xlarge",
        "ram": 393216,
    },
    "m5.2xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "37",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5.2xlarge",
            "instancesku": "7DHN2N5PWU2GRR8X",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5.2xlarge",
        "name": "m5.2xlarge",
        "ram": 32768,
    },
    "m5.4xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "2120 Mbps",
            "ecu": "70",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5.4xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5.4xlarge",
        "name": "m5.4xlarge",
        "ram": 65536,
    },
    "m5.8xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "5000 Mbps",
            "ecu": "128",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5.8xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-3",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5.8xlarge",
        "name": "m5.8xlarge",
        "ram": 131072,
    },
    "m5.large": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "10",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5.large",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5.large",
        "name": "m5.large",
        "ram": 8192,
    },
    "m5.xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "16",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5.xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5.xlarge",
        "name": "m5.xlarge",
        "ram": 16384,
    },
    "m5a.12xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "5000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5a.12xlarge",
            "instancesku": "6BBZ8RXFRQR2EXD6",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "sa-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5a.12xlarge",
        "name": "m5a.12xlarge",
        "ram": 196608,
    },
    "m5a.16xlarge": {
        "bandwidth": 12,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5a.16xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "12 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "eu-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5a.16xlarge",
        "name": "m5a.16xlarge",
        "ram": 262144,
    },
    "m5a.24xlarge": {
        "bandwidth": 20,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5a.24xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5a.24xlarge",
        "name": "m5a.24xlarge",
        "ram": 393216,
    },
    "m5a.2xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5a.2xlarge",
            "instancesku": "NP557QB7VFTC48QX",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "eu-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5a.2xlarge",
        "name": "m5a.2xlarge",
        "ram": 32768,
    },
    "m5a.4xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5a.4xlarge",
            "instancesku": "YRM7GSGQEF9HKS2Q",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5a.4xlarge",
        "name": "m5a.4xlarge",
        "ram": 65536,
    },
    "m5a.8xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5a.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5a.8xlarge",
        "name": "m5a.8xlarge",
        "ram": 131072,
    },
    "m5a.large": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5a.large",
            "instancesku": "JP2HVYDRPYGCX6DK",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "eu-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5a.large",
        "name": "m5a.large",
        "ram": 8192,
    },
    "m5a.xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5a.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "eu-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5a.xlarge",
        "name": "m5a.xlarge",
        "ram": 16384,
    },
    "m5ad.12xlarge": {
        "bandwidth": 10,
        "disk": 1800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "5000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5ad.12xlarge",
            "instancesku": "QMMYQRZBSV7ANC3N",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "ap-southeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 900 NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5ad.12xlarge",
        "name": "m5ad.12xlarge",
        "ram": 196608,
    },
    "m5ad.16xlarge": {
        "bandwidth": 12,
        "disk": 2400,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5ad.16xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "12 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 600 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5ad.16xlarge",
        "name": "m5ad.16xlarge",
        "ram": 262144,
    },
    "m5ad.24xlarge": {
        "bandwidth": 20,
        "disk": 3600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5ad.24xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 900 NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5ad.24xlarge",
        "name": "m5ad.24xlarge",
        "ram": 393216,
    },
    "m5ad.2xlarge": {
        "bandwidth": 10,
        "disk": 300,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5ad.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 300 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5ad.2xlarge",
        "name": "m5ad.2xlarge",
        "ram": 32768,
    },
    "m5ad.4xlarge": {
        "bandwidth": 10,
        "disk": 600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5ad.4xlarge",
            "instancesku": "BQ3N5HBDC5RD4EQK",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 300 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5ad.4xlarge",
        "name": "m5ad.4xlarge",
        "ram": 65536,
    },
    "m5ad.8xlarge": {
        "bandwidth": 10,
        "disk": 1200,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5ad.8xlarge",
            "instancesku": "ZSG6WH8AYEW9Y5M2",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 600 NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5ad.8xlarge",
        "name": "m5ad.8xlarge",
        "ram": 131072,
    },
    "m5ad.large": {
        "bandwidth": 10,
        "disk": 75,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5ad.large",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 75 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5ad.large",
        "name": "m5ad.large",
        "ram": 8192,
    },
    "m5ad.xlarge": {
        "bandwidth": 10,
        "disk": 150,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5ad.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 150 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5ad.xlarge",
        "name": "m5ad.xlarge",
        "ram": 16384,
    },
    "m5d.12xlarge": {
        "bandwidth": 10,
        "disk": 1800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "6000 Mbps",
            "ecu": "168",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5d.12xlarge",
            "instancesku": "UBWKCRY37QGF892F",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 900 NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5d.12xlarge",
        "name": "m5d.12xlarge",
        "ram": 196608,
    },
    "m5d.16xlarge": {
        "bandwidth": 20,
        "disk": 2400,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "256",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5d.16xlarge",
            "instancesku": "Z9BCFKSJEVHVJTTP",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 600 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5d.16xlarge",
        "name": "m5d.16xlarge",
        "ram": 262144,
    },
    "m5d.24xlarge": {
        "bandwidth": 25,
        "disk": 3600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "12000 Mbps",
            "ecu": "337",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5d.24xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 900 NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5d.24xlarge",
        "name": "m5d.24xlarge",
        "ram": 393216,
    },
    "m5d.2xlarge": {
        "bandwidth": 10,
        "disk": 300,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "37",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5d.2xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 300 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5d.2xlarge",
        "name": "m5d.2xlarge",
        "ram": 32768,
    },
    "m5d.4xlarge": {
        "bandwidth": 10,
        "disk": 600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "2120 Mbps",
            "ecu": "70",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5d.4xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 300 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5d.4xlarge",
        "name": "m5d.4xlarge",
        "ram": 65536,
    },
    "m5d.8xlarge": {
        "bandwidth": 10,
        "disk": 1200,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "5000 Mbps",
            "ecu": "128",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5d.8xlarge",
            "instancesku": "EQ35NJ6KDC4A3TS8",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 600 NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5d.8xlarge",
        "name": "m5d.8xlarge",
        "ram": 131072,
    },
    "m5d.large": {
        "bandwidth": 10,
        "disk": 75,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "10",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5d.large",
            "instancesku": "PE8JXAQ4QGZHB9GT",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 75 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5d.large",
        "name": "m5d.large",
        "ram": 8192,
    },
    "m5d.xlarge": {
        "bandwidth": 10,
        "disk": 150,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "16",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5d.xlarge",
            "instancesku": "GKEVTCW4GXTX52AW",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 150 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5d.xlarge",
        "name": "m5d.xlarge",
        "ram": 16384,
    },
    "m5dn.12xlarge": {
        "bandwidth": 50,
        "disk": 1800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5dn.12xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 900 GB NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5dn.12xlarge",
        "name": "m5dn.12xlarge",
        "ram": 196608,
    },
    "m5dn.16xlarge": {
        "bandwidth": 75,
        "disk": 2400,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5dn.16xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "75 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 600 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5dn.16xlarge",
        "name": "m5dn.16xlarge",
        "ram": 262144,
    },
    "m5dn.24xlarge": {
        "bandwidth": 100,
        "disk": 3600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5dn.24xlarge",
            "instancesku": "G9H536RMY737GEM4",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 900 NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5dn.24xlarge",
        "name": "m5dn.24xlarge",
        "ram": 393216,
    },
    "m5dn.2xlarge": {
        "bandwidth": 25,
        "disk": 300,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5dn.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 300 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5dn.2xlarge",
        "name": "m5dn.2xlarge",
        "ram": 32768,
    },
    "m5dn.4xlarge": {
        "bandwidth": 25,
        "disk": 600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5dn.4xlarge",
            "instancesku": "3Y5Q72YU4JNPU6S5",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 300 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5dn.4xlarge",
        "name": "m5dn.4xlarge",
        "ram": 65536,
    },
    "m5dn.8xlarge": {
        "bandwidth": 25,
        "disk": 1200,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "5000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5dn.8xlarge",
            "instancesku": "622SFRZSP5NFDFPF",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 600 NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5dn.8xlarge",
        "name": "m5dn.8xlarge",
        "ram": 131072,
    },
    "m5dn.large": {
        "bandwidth": 25,
        "disk": 75,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5dn.large",
            "instancesku": "SVUJNCR3NN2359RD",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 75 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5dn.large",
        "name": "m5dn.large",
        "ram": 8192,
    },
    "m5dn.xlarge": {
        "bandwidth": 25,
        "disk": 150,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5dn.xlarge",
            "instancesku": "ZKG2QXGKGMNM2V2Q",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 150 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5dn.xlarge",
        "name": "m5dn.xlarge",
        "ram": 16384,
    },
    "m5n.12xlarge": {
        "bandwidth": 50,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5n.12xlarge",
            "instancesku": "BEXX757YG98NCW3T",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5n.12xlarge",
        "name": "m5n.12xlarge",
        "ram": 196608,
    },
    "m5n.16xlarge": {
        "bandwidth": 75,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5n.16xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "75 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5n.16xlarge",
        "name": "m5n.16xlarge",
        "ram": 262144,
    },
    "m5n.24xlarge": {
        "bandwidth": 100,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5n.24xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5n.24xlarge",
        "name": "m5n.24xlarge",
        "ram": 393216,
    },
    "m5n.2xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5n.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5n.2xlarge",
        "name": "m5n.2xlarge",
        "ram": 32768,
    },
    "m5n.4xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5n.4xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5n.4xlarge",
        "name": "m5n.4xlarge",
        "ram": 65536,
    },
    "m5n.8xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "5000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5n.8xlarge",
            "instancesku": "B9H9FBY9ZKWXQT4E",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5n.8xlarge",
        "name": "m5n.8xlarge",
        "ram": 131072,
    },
    "m5n.large": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5n.large",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5n.large",
        "name": "m5n.large",
        "ram": 8192,
    },
    "m5n.xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5n.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5n.xlarge",
        "name": "m5n.xlarge",
        "ram": 16384,
    },
    "m5zn.12xlarge": {
        "bandwidth": 100,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "19000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5zn.12xlarge",
            "instancesku": "USMQJ4AAYGRVZ5AP",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Platinum 8252",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5zn.12xlarge",
        "name": "m5zn.12xlarge",
        "ram": 196608,
    },
    "m5zn.2xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3170 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5zn.2xlarge",
            "instancesku": "39TEUH9AS6NUZN5P",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8252",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5zn.2xlarge",
        "name": "m5zn.2xlarge",
        "ram": 32768,
    },
    "m5zn.3xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4750 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5zn.3xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "48 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "24",
            "physicalProcessor": "Intel Xeon Platinum 8252",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "12",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5zn.3xlarge",
        "name": "m5zn.3xlarge",
        "ram": 49152,
    },
    "m5zn.6xlarge": {
        "bandwidth": 50,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "9500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5zn.6xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "48",
            "physicalProcessor": "Intel Xeon Platinum 8252",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "24",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5zn.6xlarge",
        "name": "m5zn.6xlarge",
        "ram": 98304,
    },
    "m5zn.large": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3170 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5zn.large",
            "instancesku": "ZBF4HFE69XQJW6D8",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Platinum 8252",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5zn.large",
        "name": "m5zn.large",
        "ram": 8192,
    },
    "m5zn.xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 3170 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m5zn.xlarge",
            "instancesku": "Z5KETQXBVZT4M8PQ",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8252",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "m5zn.xlarge",
        "name": "m5zn.xlarge",
        "ram": 16384,
    },
    "m6a.12xlarge": {
        "bandwidth": 18750,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6a.12xlarge",
            "instancesku": "KFZ2VQU4FWHRV75J",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "18750 Megabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6a.12xlarge",
        "name": "m6a.12xlarge",
        "ram": 196608,
    },
    "m6a.16xlarge": {
        "bandwidth": 25000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "13333 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6a.16xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "25000 Megabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "ap-southeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6a.16xlarge",
        "name": "m6a.16xlarge",
        "ram": 262144,
    },
    "m6a.24xlarge": {
        "bandwidth": 37500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6a.24xlarge",
            "instancesku": "6XNGM5G79WDEEFRP",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "37500 Megabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "sa-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6a.24xlarge",
        "name": "m6a.24xlarge",
        "ram": 393216,
    },
    "m6a.2xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 6667 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6a.2xlarge",
            "instancesku": "FSC3WZK998TWNN9R",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "us-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6a.2xlarge",
        "name": "m6a.2xlarge",
        "ram": 32768,
    },
    "m6a.32xlarge": {
        "bandwidth": 50000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "26667 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6a.32xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "50000 Megabit",
            "normalizationSizeFactor": "256",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "us-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "128",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6a.32xlarge",
        "name": "m6a.32xlarge",
        "ram": 524288,
    },
    "m6a.48xlarge": {
        "bandwidth": 50000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "40000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6a.48xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "50000 Megabit",
            "normalizationSizeFactor": "384",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "192",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6a.48xlarge",
        "name": "m6a.48xlarge",
        "ram": 786432,
    },
    "m6a.4xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 6667 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6a.4xlarge",
            "instancesku": "VE3T4952S7TVX7NF",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6a.4xlarge",
        "name": "m6a.4xlarge",
        "ram": 65536,
    },
    "m6a.8xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "6667 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6a.8xlarge",
            "instancesku": "TP4UNYCCECX5SY5U",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "12500 Megabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6a.8xlarge",
        "name": "m6a.8xlarge",
        "ram": 131072,
    },
    "m6a.large": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 6667 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6a.large",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "eu-west-3",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6a.large",
        "name": "m6a.large",
        "ram": 8192,
    },
    "m6a.xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 6667 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6a.xlarge",
            "instancesku": "ZW78VK6K5BBFETQM",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "ap-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6a.xlarge",
        "name": "m6a.xlarge",
        "ram": 16384,
    },
    "m6g.12xlarge": {
        "bandwidth": 20,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "13500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6g.12xlarge",
            "instancesku": "PZEND6REAYMUS8YC",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6g.12xlarge",
        "name": "m6g.12xlarge",
        "ram": 196608,
    },
    "m6g.16xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "18000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6g.16xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6g.16xlarge",
        "name": "m6g.16xlarge",
        "ram": 262144,
    },
    "m6g.2xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "2250 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6g.2xlarge",
            "instancesku": "PM34ANGUVKS67VQH",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6g.2xlarge",
        "name": "m6g.2xlarge",
        "ram": 32768,
    },
    "m6g.4xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6g.4xlarge",
            "instancesku": "KFJKZ424RWQ9548M",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6g.4xlarge",
        "name": "m6g.4xlarge",
        "ram": 65536,
    },
    "m6g.8xlarge": {
        "bandwidth": 12,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "9000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6g.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "12 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6g.8xlarge",
        "name": "m6g.8xlarge",
        "ram": 131072,
    },
    "m6g.large": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "600 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6g.large",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "sa-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6g.large",
        "name": "m6g.large",
        "ram": 8192,
    },
    "m6g.medium": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "300 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6g.medium",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "1",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6g.medium",
        "name": "m6g.medium",
        "ram": 4096,
    },
    "m6g.xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "1125 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6g.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6g.xlarge",
        "name": "m6g.xlarge",
        "ram": 16384,
    },
    "m6gd.12xlarge": {
        "bandwidth": 20,
        "disk": 2850,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "13500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6gd.12xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-southeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1425 NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6gd.12xlarge",
        "name": "m6gd.12xlarge",
        "ram": 196608,
    },
    "m6gd.16xlarge": {
        "bandwidth": 25,
        "disk": 3800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "18000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6gd.16xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1900 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6gd.16xlarge",
        "name": "m6gd.16xlarge",
        "ram": 262144,
    },
    "m6gd.2xlarge": {
        "bandwidth": 10,
        "disk": 475,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6gd.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-3",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 475 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6gd.2xlarge",
        "name": "m6gd.2xlarge",
        "ram": 32768,
    },
    "m6gd.4xlarge": {
        "bandwidth": 10,
        "disk": 950,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6gd.4xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 950 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6gd.4xlarge",
        "name": "m6gd.4xlarge",
        "ram": 65536,
    },
    "m6gd.8xlarge": {
        "bandwidth": 12,
        "disk": 1900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "9000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6gd.8xlarge",
            "instancesku": "HKW2CNJU63NTP3VD",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "12 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1900 NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6gd.8xlarge",
        "name": "m6gd.8xlarge",
        "ram": 131072,
    },
    "m6gd.large": {
        "bandwidth": 10,
        "disk": 118,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6gd.large",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 118 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6gd.large",
        "name": "m6gd.large",
        "ram": 8192,
    },
    "m6gd.medium": {
        "bandwidth": 10,
        "disk": 59,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6gd.medium",
            "instancesku": "6H7QZUZ9TJ7BPSXV",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-3",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 59 NVMe SSD",
            "vcpu": "1",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6gd.medium",
        "name": "m6gd.medium",
        "ram": 4096,
    },
    "m6gd.xlarge": {
        "bandwidth": 10,
        "disk": 237,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6gd.xlarge",
            "instancesku": "TR5SD9T2SF7Q89SF",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-3",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 237 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6gd.xlarge",
        "name": "m6gd.xlarge",
        "ram": 16384,
    },
    "m6i.12xlarge": {
        "bandwidth": 18750,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "15000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6i.12xlarge",
            "instancesku": "XWT95UWVSUQJ3HFY",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "18750 Megabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6i.12xlarge",
        "name": "m6i.12xlarge",
        "ram": 196608,
    },
    "m6i.16xlarge": {
        "bandwidth": 25000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6i.16xlarge",
            "instancesku": "QGFHA8Q4X2J9XPN7",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "25000 Megabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6i.16xlarge",
        "name": "m6i.16xlarge",
        "ram": 262144,
    },
    "m6i.24xlarge": {
        "bandwidth": 37500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "30000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6i.24xlarge",
            "instancesku": "FMKFHSXPVWDFY6YQ",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "37500 Megabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6i.24xlarge",
        "name": "m6i.24xlarge",
        "ram": 393216,
    },
    "m6i.2xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6i.2xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6i.2xlarge",
        "name": "m6i.2xlarge",
        "ram": 32768,
    },
    "m6i.32xlarge": {
        "bandwidth": 50000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "40000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6i.32xlarge",
            "instancesku": "J2589GXE2X4QGE87",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "50000 Megabit",
            "normalizationSizeFactor": "256",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-southeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "128",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6i.32xlarge",
        "name": "m6i.32xlarge",
        "ram": 524288,
    },
    "m6i.4xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6i.4xlarge",
            "instancesku": "T9KPTQ9DG8SF4SU7",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6i.4xlarge",
        "name": "m6i.4xlarge",
        "ram": 65536,
    },
    "m6i.8xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6i.8xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "12500 Megabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6i.8xlarge",
        "name": "m6i.8xlarge",
        "ram": 131072,
    },
    "m6i.large": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6i.large",
            "instancesku": "3QKHM8YZH37EMM3S",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6i.large",
        "name": "m6i.large",
        "ram": 8192,
    },
    "m6i.xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6i.xlarge",
            "instancesku": "XA69DJ9RU98WA2XQ",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6i.xlarge",
        "name": "m6i.xlarge",
        "ram": 16384,
    },
    "m6id.12xlarge": {
        "bandwidth": 18750,
        "disk": 2850,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "15000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6id.12xlarge",
            "instancesku": "SJT6DHS3Q4CSDS65",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "18750 Megabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1425 NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6id.12xlarge",
        "name": "m6id.12xlarge",
        "ram": 196608,
    },
    "m6id.16xlarge": {
        "bandwidth": 25000,
        "disk": 3800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6id.16xlarge",
            "instancesku": "V8NNMRT3FKM7YMWP",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "25000 Megabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-southeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1900 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6id.16xlarge",
        "name": "m6id.16xlarge",
        "ram": 262144,
    },
    "m6id.24xlarge": {
        "bandwidth": 37500,
        "disk": 5700,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "30000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6id.24xlarge",
            "instancesku": "EQ5KP8HDGVJZRZ2Z",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "37500 Megabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 1425 NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6id.24xlarge",
        "name": "m6id.24xlarge",
        "ram": 393216,
    },
    "m6id.2xlarge": {
        "bandwidth": 12500,
        "disk": 474,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6id.2xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 474 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6id.2xlarge",
        "name": "m6id.2xlarge",
        "ram": 32768,
    },
    "m6id.32xlarge": {
        "bandwidth": 50000,
        "disk": 7600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "40000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6id.32xlarge",
            "instancesku": "9WB3HXHUUECTE545",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "50000 Megabit",
            "normalizationSizeFactor": "256",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 1900 NVMe SSD",
            "vcpu": "128",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6id.32xlarge",
        "name": "m6id.32xlarge",
        "ram": 524288,
    },
    "m6id.4xlarge": {
        "bandwidth": 12500,
        "disk": 950,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6id.4xlarge",
            "instancesku": "HJ9BZE6MX5MKXYEQ",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 950 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6id.4xlarge",
        "name": "m6id.4xlarge",
        "ram": 65536,
    },
    "m6id.8xlarge": {
        "bandwidth": 12500,
        "disk": 1900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6id.8xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "12500 Megabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1900 NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6id.8xlarge",
        "name": "m6id.8xlarge",
        "ram": 131072,
    },
    "m6id.large": {
        "bandwidth": 12500,
        "disk": 118,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6id.large",
            "instancesku": "RXJZCED6FJWF8QJ2",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 118 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6id.large",
        "name": "m6id.large",
        "ram": 8192,
    },
    "m6id.xlarge": {
        "bandwidth": 12500,
        "disk": 237,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6id.xlarge",
            "instancesku": "5AD6W95DBZ8CEEMB",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 237 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6id.xlarge",
        "name": "m6id.xlarge",
        "ram": 16384,
    },
    "m6idn.12xlarge": {
        "bandwidth": 75000,
        "disk": 2850,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "37500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6idn.12xlarge",
            "instancesku": "9K8VZMWCJENAA2F5",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "75000 Megabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-southeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1425 NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6idn.12xlarge",
        "name": "m6idn.12xlarge",
        "ram": 196608,
    },
    "m6idn.16xlarge": {
        "bandwidth": 100000,
        "disk": 3800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "50000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6idn.16xlarge",
            "instancesku": "JB3UMKUBCXFM8QEP",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "100000 Megabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1900 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6idn.16xlarge",
        "name": "m6idn.16xlarge",
        "ram": 262144,
    },
    "m6idn.24xlarge": {
        "bandwidth": 150000,
        "disk": 5700,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "75000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6idn.24xlarge",
            "instancesku": "QW53MBRRTP449U9U",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "150000 Megabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 1425 NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6idn.24xlarge",
        "name": "m6idn.24xlarge",
        "ram": 393216,
    },
    "m6idn.2xlarge": {
        "bandwidth": 40000,
        "disk": 474,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 25000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6idn.2xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 40000 Megabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 474 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6idn.2xlarge",
        "name": "m6idn.2xlarge",
        "ram": 32768,
    },
    "m6idn.32xlarge": {
        "bandwidth": 200000,
        "disk": 7600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "100000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6idn.32xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "200000 Megabit",
            "normalizationSizeFactor": "256",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 1900 NVMe SSD",
            "vcpu": "128",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6idn.32xlarge",
        "name": "m6idn.32xlarge",
        "ram": 524288,
    },
    "m6idn.4xlarge": {
        "bandwidth": 50000,
        "disk": 950,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 25000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6idn.4xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 50000 Megabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-southeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 950 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6idn.4xlarge",
        "name": "m6idn.4xlarge",
        "ram": 65536,
    },
    "m6idn.8xlarge": {
        "bandwidth": 50000,
        "disk": 1900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "25000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6idn.8xlarge",
            "instancesku": "TBYR7R7YDPZEXZHZ",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "50000 Megabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-southeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1900 NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6idn.8xlarge",
        "name": "m6idn.8xlarge",
        "ram": 131072,
    },
    "m6idn.large": {
        "bandwidth": 25000,
        "disk": 118,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 25000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6idn.large",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 25000 Megabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 118 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6idn.large",
        "name": "m6idn.large",
        "ram": 8192,
    },
    "m6idn.xlarge": {
        "bandwidth": 30000,
        "disk": 237,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 25000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6idn.xlarge",
            "instancesku": "5U57Y2C6X36HMCDU",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 30000 Megabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 237 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6idn.xlarge",
        "name": "m6idn.xlarge",
        "ram": 16384,
    },
    "m6in.12xlarge": {
        "bandwidth": 75000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "37500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6in.12xlarge",
            "instancesku": "3G8D9HES8QGQK6RD",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "75000 Megabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6in.12xlarge",
        "name": "m6in.12xlarge",
        "ram": 196608,
    },
    "m6in.16xlarge": {
        "bandwidth": 100000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "50000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6in.16xlarge",
            "instancesku": "E9REQNG3Y4NJ4J96",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "100000 Megabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-southeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6in.16xlarge",
        "name": "m6in.16xlarge",
        "ram": 262144,
    },
    "m6in.24xlarge": {
        "bandwidth": 150000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "75000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6in.24xlarge",
            "instancesku": "H24WAFC9E87MTRGD",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "150000 Megabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6in.24xlarge",
        "name": "m6in.24xlarge",
        "ram": 393216,
    },
    "m6in.2xlarge": {
        "bandwidth": 40000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 25000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6in.2xlarge",
            "instancesku": "KC365TBZWZYP44X4",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 40000 Megabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6in.2xlarge",
        "name": "m6in.2xlarge",
        "ram": 32768,
    },
    "m6in.32xlarge": {
        "bandwidth": 200000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "100000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6in.32xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "200000 Megabit",
            "normalizationSizeFactor": "256",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "128",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6in.32xlarge",
        "name": "m6in.32xlarge",
        "ram": 524288,
    },
    "m6in.4xlarge": {
        "bandwidth": 50000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 25000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6in.4xlarge",
            "instancesku": "TQ9XX6PAN3H74WPV",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 50000 Megabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6in.4xlarge",
        "name": "m6in.4xlarge",
        "ram": 65536,
    },
    "m6in.8xlarge": {
        "bandwidth": 50000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "25000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6in.8xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "50000 Megabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6in.8xlarge",
        "name": "m6in.8xlarge",
        "ram": 131072,
    },
    "m6in.large": {
        "bandwidth": 25000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 25000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6in.large",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 25000 Megabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6in.large",
        "name": "m6in.large",
        "ram": 8192,
    },
    "m6in.xlarge": {
        "bandwidth": 30000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 25000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m6in.xlarge",
            "instancesku": "KU4YBKEN6838UZCE",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 30000 Megabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "m6in.xlarge",
        "name": "m6in.xlarge",
        "ram": 16384,
    },
    "m7a.12xlarge": {
        "bandwidth": 18750,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "15000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7a.12xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "18750 Megabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7a.12xlarge",
        "name": "m7a.12xlarge",
        "ram": 196608,
    },
    "m7a.16xlarge": {
        "bandwidth": 25000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7a.16xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "25000 Megabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7a.16xlarge",
        "name": "m7a.16xlarge",
        "ram": 262144,
    },
    "m7a.24xlarge": {
        "bandwidth": 37500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "30000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7a.24xlarge",
            "instancesku": "WD4GMXW9CSWQQ676",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "37500 Megabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7a.24xlarge",
        "name": "m7a.24xlarge",
        "ram": 393216,
    },
    "m7a.2xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7a.2xlarge",
            "instancesku": "TTV4V7G52BN2MHHJ",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7a.2xlarge",
        "name": "m7a.2xlarge",
        "ram": 32768,
    },
    "m7a.32xlarge": {
        "bandwidth": 50000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "40000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7a.32xlarge",
            "instancesku": "MPXAFP6AXX98RT7Z",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "50000 Megabit",
            "normalizationSizeFactor": "256",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "128",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7a.32xlarge",
        "name": "m7a.32xlarge",
        "ram": 524288,
    },
    "m7a.48xlarge": {
        "bandwidth": 50000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "40000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7a.48xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "50000 Megabit",
            "normalizationSizeFactor": "384",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "192",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7a.48xlarge",
        "name": "m7a.48xlarge",
        "ram": 786432,
    },
    "m7a.4xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7a.4xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7a.4xlarge",
        "name": "m7a.4xlarge",
        "ram": 65536,
    },
    "m7a.8xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7a.8xlarge",
            "instancesku": "AJ4DT835ZFN2SHZK",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "12500 Megabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7a.8xlarge",
        "name": "m7a.8xlarge",
        "ram": 131072,
    },
    "m7a.large": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7a.large",
            "instancesku": "4J24EAVW8MK2ZBUZ",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7a.large",
        "name": "m7a.large",
        "ram": 8192,
    },
    "m7a.medium": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7a.medium",
            "instancesku": "CHCMF8EXG35ZCG8C",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "1",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7a.medium",
        "name": "m7a.medium",
        "ram": 4096,
    },
    "m7a.xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7a.xlarge",
            "instancesku": "2387YVTKZPXRK3D8",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7a.xlarge",
        "name": "m7a.xlarge",
        "ram": 16384,
    },
    "m7g.12xlarge": {
        "bandwidth": 22500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "15000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7g.12xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "22500 Megabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "sa-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7g.12xlarge",
        "name": "m7g.12xlarge",
        "ram": 196608,
    },
    "m7g.16xlarge": {
        "bandwidth": 30,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7g.16xlarge",
            "instancesku": "GGHDPF534R2KZS4J",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "30 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-southeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7g.16xlarge",
        "name": "m7g.16xlarge",
        "ram": 262144,
    },
    "m7g.2xlarge": {
        "bandwidth": 15,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7g.2xlarge",
            "instancesku": "SAV6N5BX7BW94B2X",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 15 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7g.2xlarge",
        "name": "m7g.2xlarge",
        "ram": 32768,
    },
    "m7g.4xlarge": {
        "bandwidth": 15,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7g.4xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 15 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7g.4xlarge",
        "name": "m7g.4xlarge",
        "ram": 65536,
    },
    "m7g.8xlarge": {
        "bandwidth": 15,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7g.8xlarge",
            "instancesku": "J7ZYG6NVGP3KXF6T",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "15 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7g.8xlarge",
        "name": "m7g.8xlarge",
        "ram": 131072,
    },
    "m7g.large": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7g.large",
            "instancesku": "QV8RCNQE2XGXV9J7",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7g.large",
        "name": "m7g.large",
        "ram": 8192,
    },
    "m7g.medium": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7g.medium",
            "instancesku": "PYT8U3X9MJ82GKR9",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "1",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7g.medium",
        "name": "m7g.medium",
        "ram": 4096,
    },
    "m7g.xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7g.xlarge",
            "instancesku": "A5ZDYYD96FGXHU22",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7g.xlarge",
        "name": "m7g.xlarge",
        "ram": 16384,
    },
    "m7gd.12xlarge": {
        "bandwidth": 22500,
        "disk": 2850,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "15000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7gd.12xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "22500 Megabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1425 NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7gd.12xlarge",
        "name": "m7gd.12xlarge",
        "ram": 196608,
    },
    "m7gd.16xlarge": {
        "bandwidth": 30,
        "disk": 3800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7gd.16xlarge",
            "instancesku": "3SRUAFSQTGMDY2B8",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "30 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1900 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7gd.16xlarge",
        "name": "m7gd.16xlarge",
        "ram": 262144,
    },
    "m7gd.2xlarge": {
        "bandwidth": 15,
        "disk": 475,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7gd.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 15 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 475 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7gd.2xlarge",
        "name": "m7gd.2xlarge",
        "ram": 32768,
    },
    "m7gd.4xlarge": {
        "bandwidth": 15,
        "disk": 950,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7gd.4xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 15 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-3",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 950 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7gd.4xlarge",
        "name": "m7gd.4xlarge",
        "ram": 65536,
    },
    "m7gd.8xlarge": {
        "bandwidth": 15,
        "disk": 1900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7gd.8xlarge",
            "instancesku": "SVZECXW4DYGVAZAA",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "15 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1900 NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7gd.8xlarge",
        "name": "m7gd.8xlarge",
        "ram": 131072,
    },
    "m7gd.large": {
        "bandwidth": 12500,
        "disk": 118,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7gd.large",
            "instancesku": "KY3EMBYV494XVNZC",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 118 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7gd.large",
        "name": "m7gd.large",
        "ram": 8192,
    },
    "m7gd.medium": {
        "bandwidth": 12500,
        "disk": 59,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7gd.medium",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 59 NVMe SSD",
            "vcpu": "1",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7gd.medium",
        "name": "m7gd.medium",
        "ram": 4096,
    },
    "m7gd.xlarge": {
        "bandwidth": 12500,
        "disk": 237,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7gd.xlarge",
            "instancesku": "DX78QR4R4G9TEN68",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 237 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7gd.xlarge",
        "name": "m7gd.xlarge",
        "ram": 16384,
    },
    "m7i-flex.12xlarge": {
        "bandwidth": 18750,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 15000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7i-flex.12xlarge",
            "instancesku": "T434T6R9XBK6XNCH",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "Up to 18750 Megabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7i-flex.12xlarge",
        "name": "m7i-flex.12xlarge",
        "ram": 196608,
    },
    "m7i-flex.16xlarge": {
        "bandwidth": 25000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7i-flex.16xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "Up to 25000 Megabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "ap-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7i-flex.16xlarge",
        "name": "m7i-flex.16xlarge",
        "ram": 262144,
    },
    "m7i-flex.2xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7i-flex.2xlarge",
            "instancesku": "WQZ9EGYW4VCTDNC6",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7i-flex.2xlarge",
        "name": "m7i-flex.2xlarge",
        "ram": 32768,
    },
    "m7i-flex.4xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7i-flex.4xlarge",
            "instancesku": "3CECDFXBGJUVXA2K",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7i-flex.4xlarge",
        "name": "m7i-flex.4xlarge",
        "ram": 65536,
    },
    "m7i-flex.8xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7i-flex.8xlarge",
            "instancesku": "PR36FCC4XVPUX2RA",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7i-flex.8xlarge",
        "name": "m7i-flex.8xlarge",
        "ram": 131072,
    },
    "m7i-flex.large": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7i-flex.large",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "eu-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7i-flex.large",
        "name": "m7i-flex.large",
        "ram": 8192,
    },
    "m7i-flex.xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7i-flex.xlarge",
            "instancesku": "G48VQNBX7U5V442V",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7i-flex.xlarge",
        "name": "m7i-flex.xlarge",
        "ram": 16384,
    },
    "m7i.12xlarge": {
        "bandwidth": 18750,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "15000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7i.12xlarge",
            "instancesku": "86SAJ2BS2H9N34K8",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "18750 Megabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "us-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7i.12xlarge",
        "name": "m7i.12xlarge",
        "ram": 196608,
    },
    "m7i.16xlarge": {
        "bandwidth": 25000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7i.16xlarge",
            "instancesku": "KFXDGYA5B6TGN7C3",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "25000 Megabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7i.16xlarge",
        "name": "m7i.16xlarge",
        "ram": 262144,
    },
    "m7i.24xlarge": {
        "bandwidth": 37500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "30000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7i.24xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "37500 Megabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7i.24xlarge",
        "name": "m7i.24xlarge",
        "ram": 393216,
    },
    "m7i.2xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7i.2xlarge",
            "instancesku": "2K3M2YB5KMHAWPSU",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7i.2xlarge",
        "name": "m7i.2xlarge",
        "ram": 32768,
    },
    "m7i.48xlarge": {
        "bandwidth": 50000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "40000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7i.48xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "50000 Megabit",
            "normalizationSizeFactor": "384",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "192",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7i.48xlarge",
        "name": "m7i.48xlarge",
        "ram": 786432,
    },
    "m7i.4xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7i.4xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7i.4xlarge",
        "name": "m7i.4xlarge",
        "ram": 65536,
    },
    "m7i.8xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7i.8xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "12500 Megabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7i.8xlarge",
        "name": "m7i.8xlarge",
        "ram": 131072,
    },
    "m7i.large": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7i.large",
            "instancesku": "RFCGCPF2C7PHAQ6U",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "us-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7i.large",
        "name": "m7i.large",
        "ram": 8192,
    },
    "m7i.xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m7i.xlarge",
            "instancesku": "2AJT6K2EBSS7QZFR",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "m7i.xlarge",
        "name": "m7i.xlarge",
        "ram": 16384,
    },
    "m8g.12xlarge": {
        "bandwidth": 22,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "15000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m8g.12xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "22.5 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "m8g.12xlarge",
        "name": "m8g.12xlarge",
        "ram": 196608,
    },
    "m8g.16xlarge": {
        "bandwidth": 30,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m8g.16xlarge",
            "instancesku": "XBV98BGWRJ7QUGPC",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "30 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "m8g.16xlarge",
        "name": "m8g.16xlarge",
        "ram": 262144,
    },
    "m8g.24xlarge": {
        "bandwidth": 40,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "30000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m8g.24xlarge",
            "instancesku": "MYF3RF37SMVHCN94",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "40 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "m8g.24xlarge",
        "name": "m8g.24xlarge",
        "ram": 393216,
    },
    "m8g.2xlarge": {
        "bandwidth": 15,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m8g.2xlarge",
            "instancesku": "VHR6K5EGBNXFEVFY",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 15 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "m8g.2xlarge",
        "name": "m8g.2xlarge",
        "ram": 32768,
    },
    "m8g.48xlarge": {
        "bandwidth": 50,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "40000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m8g.48xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "384",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "192",
            "vpcnetworkingsupport": "true",
        },
        "id": "m8g.48xlarge",
        "name": "m8g.48xlarge",
        "ram": 786432,
    },
    "m8g.4xlarge": {
        "bandwidth": 15,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m8g.4xlarge",
            "instancesku": "Y4GWK766P3Q4AD72",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 15 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "m8g.4xlarge",
        "name": "m8g.4xlarge",
        "ram": 65536,
    },
    "m8g.8xlarge": {
        "bandwidth": 15,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m8g.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "15 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "m8g.8xlarge",
        "name": "m8g.8xlarge",
        "ram": 131072,
    },
    "m8g.large": {
        "bandwidth": 12,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m8g.large",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 12.5 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "m8g.large",
        "name": "m8g.large",
        "ram": 8192,
    },
    "m8g.medium": {
        "bandwidth": 12,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m8g.medium",
            "instancesku": "HT8KKMMTD48AR8U8",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 12.5 Gigabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "1",
            "vpcnetworkingsupport": "true",
        },
        "id": "m8g.medium",
        "name": "m8g.medium",
        "ram": 4096,
    },
    "m8g.xlarge": {
        "bandwidth": 12,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "m8g.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 12.5 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "m8g.xlarge",
        "name": "m8g.xlarge",
        "ram": 16384,
    },
    "p2.16xlarge": {
        "bandwidth": 20,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "201",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "16",
            "gpuMemory": "NA",
            "instanceFamily": "GPU instance",
            "instanceType": "p2.16xlarge",
            "instancesku": "8NZJVJPUUPFQD5XT",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "732 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "p2.16xlarge",
        "name": "p2.16xlarge",
        "ram": 749568,
    },
    "p2.8xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "5000 Mbps",
            "ecu": "97",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "8",
            "gpuMemory": "NA",
            "instanceFamily": "GPU instance",
            "instanceType": "p2.8xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "488 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "p2.8xlarge",
        "name": "p2.8xlarge",
        "ram": 499712,
    },
    "p2.xlarge": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "750 Mbps",
            "ecu": "16",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "gpuMemory": "NA",
            "instanceFamily": "GPU instance",
            "instanceType": "p2.xlarge",
            "instancesku": "BKKRW325YG73QXYH",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "61 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "ap-southeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "p2.xlarge",
        "name": "p2.xlarge",
        "ram": 62464,
    },
    "p3.16xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14000 Mbps",
            "ecu": "201",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "8",
            "gpuMemory": "128 GiB",
            "instanceFamily": "GPU instance",
            "instanceType": "p3.16xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "488 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "p3.16xlarge",
        "name": "p3.16xlarge",
        "ram": 499712,
    },
    "p3.2xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "1750 Mbps",
            "ecu": "31",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "1",
            "gpuMemory": "16 GiB",
            "instanceFamily": "GPU instance",
            "instanceType": "p3.2xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "61 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "p3.2xlarge",
        "name": "p3.2xlarge",
        "ram": 62464,
    },
    "p3.8xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "97",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "4",
            "gpuMemory": "64 GiB",
            "instanceFamily": "GPU instance",
            "instanceType": "p3.8xlarge",
            "instancesku": "HCAHXPK7A9MBBZGN",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "244 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "p3.8xlarge",
        "name": "p3.8xlarge",
        "ram": 249856,
    },
    "p3dn.24xlarge": {
        "bandwidth": 100,
        "disk": 1800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14000 Mbps",
            "ecu": "337",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "8",
            "gpuMemory": "256 GiB",
            "instanceFamily": "GPU instance",
            "instanceType": "p3dn.24xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8175 (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 900 NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "p3dn.24xlarge",
        "name": "p3dn.24xlarge",
        "ram": 786432,
    },
    "p4d.24xlarge": {
        "bandwidth": 400,
        "disk": 8000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "19000 Mbps",
            "ecu": "345",
            "enhancedNetworkingSupported": "No",
            "gpu": "8",
            "gpuMemory": "320 GB HBM2",
            "instanceFamily": "GPU instance",
            "instanceType": "p4d.24xlarge",
            "instancesku": "2Z7ARVXU7PT7RXSJ",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "1152 GiB",
            "networkPerformance": "400 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8275L",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "8 x 1000 SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "p4d.24xlarge",
        "name": "p4d.24xlarge",
        "ram": 1179648,
    },
    "p4de.24xlarge": {
        "bandwidth": 400,
        "disk": 8000,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "19000 Mbps",
            "ecu": "345",
            "enhancedNetworkingSupported": "No",
            "gpu": "8",
            "gpuMemory": "640 GB HBM2e",
            "instanceFamily": "GPU instance",
            "instanceType": "p4de.24xlarge",
            "instancesku": "DSNTTSZSNTPZW8NS",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "1152 GiB",
            "networkPerformance": "400 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8275L",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "8 x 1000 SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "p4de.24xlarge",
        "name": "p4de.24xlarge",
        "ram": 1179648,
    },
    "p5.48xlarge": {
        "bandwidth": 3200,
        "disk": 30720,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.95 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "80 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "8",
            "gpuMemory": "640 GB HBM3",
            "instanceFamily": "GPU instance",
            "instanceType": "p5.48xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "2048 GiB",
            "networkPerformance": "3200 Gigabit",
            "normalizationSizeFactor": "384",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "8 x 3840 GB SSD",
            "vcpu": "192",
            "vpcnetworkingsupport": "true",
        },
        "id": "p5.48xlarge",
        "name": "p5.48xlarge",
        "ram": 2097152,
    },
    "p5e.48xlarge": {
        "bandwidth": 3200,
        "disk": 30720,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.95 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "80 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "8",
            "gpuMemory": "1128 GB HBM3",
            "instanceFamily": "GPU instance",
            "instanceType": "p5e.48xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "CapacityBlock",
            "memory": "2048 GiB",
            "networkPerformance": "3200 Gigabit",
            "normalizationSizeFactor": "384",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "8 x 3840 GB SSD",
            "vcpu": "192",
            "vpcnetworkingsupport": "true",
        },
        "id": "p5e.48xlarge",
        "name": "p5e.48xlarge",
        "ram": 2097152,
    },
    "p5en.48xlarge": {
        "bandwidth": 3200,
        "disk": 30720,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "100 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpu": "8",
            "gpuMemory": "141 GB HBM3e",
            "instanceFamily": "GPU instance",
            "instanceType": "p5en.48xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "CapacityBlock",
            "memory": "2048 GiB",
            "networkPerformance": "3200 Gigabit",
            "normalizationSizeFactor": "384",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AMX",
            "regionCode": "ap-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "8 x 3840 NVMe SSD",
            "vcpu": "192",
            "vpcnetworkingsupport": "true",
        },
        "id": "p5en.48xlarge",
        "name": "p5en.48xlarge",
        "ram": 2097152,
    },
    "r3.2xlarge": {
        "bandwidth": None,
        "disk": 160,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "No",
            "ecu": "26",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r3.2xlarge",
            "instancesku": "MYE2APSUBDYT5XGN",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "61 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon E5-2670 v2 (Ivy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 160 SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "r3.2xlarge",
        "name": "r3.2xlarge",
        "ram": 62464,
    },
    "r3.4xlarge": {
        "bandwidth": None,
        "disk": 320,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "No",
            "ecu": "52",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r3.4xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "122 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon E5-2670 v2 (Ivy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "regionCode": "sa-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 320 SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "r3.4xlarge",
        "name": "r3.4xlarge",
        "ram": 124928,
    },
    "r3.8xlarge": {
        "bandwidth": 10,
        "disk": 640,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "No",
            "ecu": "104",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r3.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "244 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon E5-2670 v2 (Ivy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 320 SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "r3.8xlarge",
        "name": "r3.8xlarge",
        "ram": 249856,
    },
    "r3.large": {
        "bandwidth": None,
        "disk": 32,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "No",
            "ecu": "6.5",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r3.large",
            "instancesku": "JXRNUMXAZNYRPC72",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "15.25 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon E5-2670 v2 (Ivy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 32 SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "r3.large",
        "name": "r3.large",
        "ram": 15616,
    },
    "r3.xlarge": {
        "bandwidth": None,
        "disk": 80,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "true",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "No",
            "ecu": "13",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r3.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "30.5 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon E5-2670 v2 (Ivy Bridge)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 80 SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "r3.xlarge",
        "name": "r3.xlarge",
        "ram": 31232,
    },
    "r4.16xlarge": {
        "bandwidth": 20,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "12000 Mbps",
            "ecu": "201",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r4.16xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "488 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "r4.16xlarge",
        "name": "r4.16xlarge",
        "ram": 499712,
    },
    "r4.2xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "1600 Mbps",
            "ecu": "31",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r4.2xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "61 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "r4.2xlarge",
        "name": "r4.2xlarge",
        "ram": 62464,
    },
    "r4.4xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3000 Mbps",
            "ecu": "58",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r4.4xlarge",
            "instancesku": "C9S5QU5S5T9EP9E3",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "122 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "r4.4xlarge",
        "name": "r4.4xlarge",
        "ram": 124928,
    },
    "r4.8xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "6000 Mbps",
            "ecu": "97",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r4.8xlarge",
            "instancesku": "6QBU7TR6TCVBPJC3",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "244 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "r4.8xlarge",
        "name": "r4.8xlarge",
        "ram": 249856,
    },
    "r4.large": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "400 Mbps",
            "ecu": "8",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r4.large",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "15.25 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "sa-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "r4.large",
        "name": "r4.large",
        "ram": 15616,
    },
    "r4.xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "800 Mbps",
            "ecu": "16",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r4.xlarge",
            "instancesku": "MSR496BA8MMSSHES",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "30.5 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon E5-2686 v4 (Broadwell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "us-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "r4.xlarge",
        "name": "r4.xlarge",
        "ram": 31232,
    },
    "r5.12xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "9500 Mbps",
            "ecu": "168",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5.12xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5.12xlarge",
        "name": "r5.12xlarge",
        "ram": 393216,
    },
    "r5.16xlarge": {
        "bandwidth": 20,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "13600 Mbps",
            "ecu": "256",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5.16xlarge",
            "instancesku": "NAT7A478TMR4QGPN",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5.16xlarge",
        "name": "r5.16xlarge",
        "ram": 524288,
    },
    "r5.24xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "19000 Mbps",
            "ecu": "337",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5.24xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5.24xlarge",
        "name": "r5.24xlarge",
        "ram": 786432,
    },
    "r5.2xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4750 Mbps",
            "ecu": "37",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5.2xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5.2xlarge",
        "name": "r5.2xlarge",
        "ram": 65536,
    },
    "r5.4xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4750 Mbps",
            "ecu": "70",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5.4xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5.4xlarge",
        "name": "r5.4xlarge",
        "ram": 131072,
    },
    "r5.8xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "6800 Mbps",
            "ecu": "128",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5.8xlarge",
            "instancesku": "5Q2X664FUPK733XT",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "sa-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5.8xlarge",
        "name": "r5.8xlarge",
        "ram": 262144,
    },
    "r5.large": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4750 Mbps",
            "ecu": "10",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5.large",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5.large",
        "name": "r5.large",
        "ram": 16384,
    },
    "r5.xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4750 Mbps",
            "ecu": "19",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5.xlarge",
            "instancesku": "8UGGWQHV44F4KH8W",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5.xlarge",
        "name": "r5.xlarge",
        "ram": 32768,
    },
    "r5a.12xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "5000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5a.12xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5a.12xlarge",
        "name": "r5a.12xlarge",
        "ram": 393216,
    },
    "r5a.16xlarge": {
        "bandwidth": 12,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5a.16xlarge",
            "instancesku": "2B3VAYDSHV6M39S6",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "12 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5a.16xlarge",
        "name": "r5a.16xlarge",
        "ram": 524288,
    },
    "r5a.24xlarge": {
        "bandwidth": 20,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5a.24xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5a.24xlarge",
        "name": "r5a.24xlarge",
        "ram": 786432,
    },
    "r5a.2xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5a.2xlarge",
            "instancesku": "4YMCRU5MXH67KH3C",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5a.2xlarge",
        "name": "r5a.2xlarge",
        "ram": 65536,
    },
    "r5a.4xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5a.4xlarge",
            "instancesku": "BCXRZNBDWWSKS95T",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "eu-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5a.4xlarge",
        "name": "r5a.4xlarge",
        "ram": 131072,
    },
    "r5a.8xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5a.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5a.8xlarge",
        "name": "r5a.8xlarge",
        "ram": 262144,
    },
    "r5a.large": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5a.large",
            "instancesku": "HXFJ8J7KRRDCUQDF",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5a.large",
        "name": "r5a.large",
        "ram": 16384,
    },
    "r5a.xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5a.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "eu-west-3",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5a.xlarge",
        "name": "r5a.xlarge",
        "ram": 32768,
    },
    "r5ad.12xlarge": {
        "bandwidth": 10,
        "disk": 1800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "5000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5ad.12xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 900 NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5ad.12xlarge",
        "name": "r5ad.12xlarge",
        "ram": 393216,
    },
    "r5ad.16xlarge": {
        "bandwidth": 12,
        "disk": 2400,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5ad.16xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "12 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "sa-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 600 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5ad.16xlarge",
        "name": "r5ad.16xlarge",
        "ram": 524288,
    },
    "r5ad.24xlarge": {
        "bandwidth": 20,
        "disk": 3600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5ad.24xlarge",
            "instancesku": "AYARQMMWY3KRT3Z8",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "ap-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 900 NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5ad.24xlarge",
        "name": "r5ad.24xlarge",
        "ram": 786432,
    },
    "r5ad.2xlarge": {
        "bandwidth": 10,
        "disk": 300,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5ad.2xlarge",
            "instancesku": "UQNYM2QVMBNC9SBD",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 300 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5ad.2xlarge",
        "name": "r5ad.2xlarge",
        "ram": 65536,
    },
    "r5ad.4xlarge": {
        "bandwidth": 10,
        "disk": 600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5ad.4xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 300 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5ad.4xlarge",
        "name": "r5ad.4xlarge",
        "ram": 131072,
    },
    "r5ad.8xlarge": {
        "bandwidth": 10,
        "disk": 1200,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5ad.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 600 NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5ad.8xlarge",
        "name": "r5ad.8xlarge",
        "ram": 262144,
    },
    "r5ad.large": {
        "bandwidth": 10,
        "disk": 75,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5ad.large",
            "instancesku": "ZT4UK6N7AH37F6Q2",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "ap-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 75 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5ad.large",
        "name": "r5ad.large",
        "ram": 16384,
    },
    "r5ad.xlarge": {
        "bandwidth": 10,
        "disk": 150,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5ad.xlarge",
            "instancesku": "HZJ8V6JGDWGUM5QF",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 150 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5ad.xlarge",
        "name": "r5ad.xlarge",
        "ram": 32768,
    },
    "r5b.12xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "30 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5b.12xlarge",
            "instancesku": "XFAAGEXXCF8D5UCC",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5b.12xlarge",
        "name": "r5b.12xlarge",
        "ram": 393216,
    },
    "r5b.16xlarge": {
        "bandwidth": 20,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "40 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5b.16xlarge",
            "instancesku": "6U7DBQWKEUPM26YD",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5b.16xlarge",
        "name": "r5b.16xlarge",
        "ram": 524288,
    },
    "r5b.24xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "60 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5b.24xlarge",
            "instancesku": "ACQGXR86YVD63UND",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5b.24xlarge",
        "name": "r5b.24xlarge",
        "ram": 786432,
    },
    "r5b.2xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 12.5 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5b.2xlarge",
            "instancesku": "UMH7BAMYN73TQGWY",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "sa-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5b.2xlarge",
        "name": "r5b.2xlarge",
        "ram": 65536,
    },
    "r5b.4xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5b.4xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5b.4xlarge",
        "name": "r5b.4xlarge",
        "ram": 131072,
    },
    "r5b.8xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "5 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5b.8xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5b.8xlarge",
        "name": "r5b.8xlarge",
        "ram": 262144,
    },
    "r5b.large": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5b.large",
            "instancesku": "P8BAP4U25KDKYSE7",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5b.large",
        "name": "r5b.large",
        "ram": 16384,
    },
    "r5b.xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5b.xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-southeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5b.xlarge",
        "name": "r5b.xlarge",
        "ram": 32768,
    },
    "r5d.12xlarge": {
        "bandwidth": 10,
        "disk": 1800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "6000 Mbps",
            "ecu": "168",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5d.12xlarge",
            "instancesku": "JD26ZJK92H6X64TZ",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 900 NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5d.12xlarge",
        "name": "r5d.12xlarge",
        "ram": 393216,
    },
    "r5d.16xlarge": {
        "bandwidth": 20,
        "disk": 2400,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "256",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5d.16xlarge",
            "instancesku": "XJ5CFKTMG2X7XGD7",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 600 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5d.16xlarge",
        "name": "r5d.16xlarge",
        "ram": 524288,
    },
    "r5d.24xlarge": {
        "bandwidth": 25,
        "disk": 3600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "12000 Mbps",
            "ecu": "337",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5d.24xlarge",
            "instancesku": "9H8W7UEREMEYWDWH",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 900 NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5d.24xlarge",
        "name": "r5d.24xlarge",
        "ram": 786432,
    },
    "r5d.2xlarge": {
        "bandwidth": 10,
        "disk": 300,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "37",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5d.2xlarge",
            "instancesku": "W5UKUYPDGRSFBTEH",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 300 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5d.2xlarge",
        "name": "r5d.2xlarge",
        "ram": 65536,
    },
    "r5d.4xlarge": {
        "bandwidth": 10,
        "disk": 600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "2120 Mbps",
            "ecu": "70",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5d.4xlarge",
            "instancesku": "XV8MVPMGEYQC3PX7",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 300 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5d.4xlarge",
        "name": "r5d.4xlarge",
        "ram": 131072,
    },
    "r5d.8xlarge": {
        "bandwidth": 10,
        "disk": 1200,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "5000 Mbps",
            "ecu": "128",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5d.8xlarge",
            "instancesku": "QCJ53WA9MJMS6BDC",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 600 NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5d.8xlarge",
        "name": "r5d.8xlarge",
        "ram": 262144,
    },
    "r5d.large": {
        "bandwidth": 10,
        "disk": 75,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "10",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5d.large",
            "instancesku": "CUZ6G9ZCUKT64FJS",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 75 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5d.large",
        "name": "r5d.large",
        "ram": 16384,
    },
    "r5d.xlarge": {
        "bandwidth": 10,
        "disk": 150,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "19",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5d.xlarge",
            "instancesku": "2NUFS4WQ7FCZY3CX",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8175",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 150 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5d.xlarge",
        "name": "r5d.xlarge",
        "ram": 32768,
    },
    "r5dn.12xlarge": {
        "bandwidth": 50,
        "disk": 1800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "6000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5dn.12xlarge",
            "instancesku": "KP8X3ZRFBR5BE2QG",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 900 GB NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5dn.12xlarge",
        "name": "r5dn.12xlarge",
        "ram": 393216,
    },
    "r5dn.16xlarge": {
        "bandwidth": 75,
        "disk": 2400,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5dn.16xlarge",
            "instancesku": "HESW5QVR6YKYRWZ8",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "75 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 600 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5dn.16xlarge",
        "name": "r5dn.16xlarge",
        "ram": 524288,
    },
    "r5dn.24xlarge": {
        "bandwidth": 100,
        "disk": 3600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "12000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5dn.24xlarge",
            "instancesku": "8PAUA987J4Q8MTRH",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 900 NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5dn.24xlarge",
        "name": "r5dn.24xlarge",
        "ram": 786432,
    },
    "r5dn.2xlarge": {
        "bandwidth": 25,
        "disk": 300,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5dn.2xlarge",
            "instancesku": "5TFDK9PHN5A9RS5A",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 300 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5dn.2xlarge",
        "name": "r5dn.2xlarge",
        "ram": 65536,
    },
    "r5dn.4xlarge": {
        "bandwidth": 25,
        "disk": 600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5dn.4xlarge",
            "instancesku": "PVTMYYCT833KAG2E",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 300 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5dn.4xlarge",
        "name": "r5dn.4xlarge",
        "ram": 131072,
    },
    "r5dn.8xlarge": {
        "bandwidth": 25,
        "disk": 1200,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "5000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5dn.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 600 NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5dn.8xlarge",
        "name": "r5dn.8xlarge",
        "ram": 262144,
    },
    "r5dn.large": {
        "bandwidth": 25,
        "disk": 75,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5dn.large",
            "instancesku": "K4T5UGP4HDGJPFH7",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 75 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5dn.large",
        "name": "r5dn.large",
        "ram": 16384,
    },
    "r5dn.xlarge": {
        "bandwidth": 25,
        "disk": 150,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5dn.xlarge",
            "instancesku": "PTAK6AMJWZ33TRF6",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 150 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5dn.xlarge",
        "name": "r5dn.xlarge",
        "ram": 32768,
    },
    "r5n.12xlarge": {
        "bandwidth": 50,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "6000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5n.12xlarge",
            "instancesku": "SMH7GCC45SZ6USV6",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5n.12xlarge",
        "name": "r5n.12xlarge",
        "ram": 393216,
    },
    "r5n.16xlarge": {
        "bandwidth": 75,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5n.16xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "75 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5n.16xlarge",
        "name": "r5n.16xlarge",
        "ram": 524288,
    },
    "r5n.24xlarge": {
        "bandwidth": 100,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "12000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5n.24xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5n.24xlarge",
        "name": "r5n.24xlarge",
        "ram": 786432,
    },
    "r5n.2xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5n.2xlarge",
            "instancesku": "E73XR9X29XEVC78A",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5n.2xlarge",
        "name": "r5n.2xlarge",
        "ram": 65536,
    },
    "r5n.4xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5n.4xlarge",
            "instancesku": "XAPBMPSJVJ5EWR55",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5n.4xlarge",
        "name": "r5n.4xlarge",
        "ram": 131072,
    },
    "r5n.8xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "5000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5n.8xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-3",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5n.8xlarge",
        "name": "r5n.8xlarge",
        "ram": 262144,
    },
    "r5n.large": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5n.large",
            "instancesku": "6WD6N47RFBKDZDF8",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5n.large",
        "name": "r5n.large",
        "ram": 16384,
    },
    "r5n.xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2120 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r5n.xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8259 (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "r5n.xlarge",
        "name": "r5n.xlarge",
        "ram": 32768,
    },
    "r6a.12xlarge": {
        "bandwidth": 18750,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6a.12xlarge",
            "instancesku": "5RG27YCVCH2MSGJR",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "18750 Megabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6a.12xlarge",
        "name": "r6a.12xlarge",
        "ram": 393216,
    },
    "r6a.16xlarge": {
        "bandwidth": 25000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "13333 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6a.16xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "25000 Megabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6a.16xlarge",
        "name": "r6a.16xlarge",
        "ram": 524288,
    },
    "r6a.24xlarge": {
        "bandwidth": 37500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6a.24xlarge",
            "instancesku": "MSN6FFMJFF9MMQVA",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "37500 Megabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6a.24xlarge",
        "name": "r6a.24xlarge",
        "ram": 786432,
    },
    "r6a.2xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 6667 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6a.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6a.2xlarge",
        "name": "r6a.2xlarge",
        "ram": 65536,
    },
    "r6a.32xlarge": {
        "bandwidth": 50000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "26667 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6a.32xlarge",
            "instancesku": "7BSD878TFSBBCHD2",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "1024 GiB",
            "networkPerformance": "50000 Megabit",
            "normalizationSizeFactor": "256",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "ap-southeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "128",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6a.32xlarge",
        "name": "r6a.32xlarge",
        "ram": 1048576,
    },
    "r6a.48xlarge": {
        "bandwidth": 50000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "40000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6a.48xlarge",
            "instancesku": "TUKRAD8KDFD7PMUX",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "1536 GiB",
            "networkPerformance": "50000 Megabit",
            "normalizationSizeFactor": "384",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "ap-southeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "192",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6a.48xlarge",
        "name": "r6a.48xlarge",
        "ram": 1572864,
    },
    "r6a.4xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 6667 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6a.4xlarge",
            "instancesku": "VM2H22UZVW5658C9",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "ap-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6a.4xlarge",
        "name": "r6a.4xlarge",
        "ram": 131072,
    },
    "r6a.8xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "6667 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6a.8xlarge",
            "instancesku": "GHFUUFQUYNGYV4M3",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "12500 Megabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6a.8xlarge",
        "name": "r6a.8xlarge",
        "ram": 262144,
    },
    "r6a.large": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 6667 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6a.large",
            "instancesku": "RTBTCPPPNNMZZZRE",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6a.large",
        "name": "r6a.large",
        "ram": 16384,
    },
    "r6a.xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.6 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 6667 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6a.xlarge",
            "instancesku": "M6ZJ4JJYJXW3FH72",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AMD EPYC 7R13 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6a.xlarge",
        "name": "r6a.xlarge",
        "ram": 32768,
    },
    "r6g.12xlarge": {
        "bandwidth": 20,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "13500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6g.12xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6g.12xlarge",
        "name": "r6g.12xlarge",
        "ram": 393216,
    },
    "r6g.16xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "18000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6g.16xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6g.16xlarge",
        "name": "r6g.16xlarge",
        "ram": 524288,
    },
    "r6g.2xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6g.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6g.2xlarge",
        "name": "r6g.2xlarge",
        "ram": 65536,
    },
    "r6g.4xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6g.4xlarge",
            "instancesku": "G9CVKPRNCE3JEHNA",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6g.4xlarge",
        "name": "r6g.4xlarge",
        "ram": 131072,
    },
    "r6g.8xlarge": {
        "bandwidth": 12,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "9000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6g.8xlarge",
            "instancesku": "JU3XS3CSJVG9PG6T",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "12 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6g.8xlarge",
        "name": "r6g.8xlarge",
        "ram": 262144,
    },
    "r6g.large": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6g.large",
            "instancesku": "3F7S3UNTHN3MP78B",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-3",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6g.large",
        "name": "r6g.large",
        "ram": 16384,
    },
    "r6g.medium": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6g.medium",
            "instancesku": "94RYVQHX4TWDPAUG",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "1",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6g.medium",
        "name": "r6g.medium",
        "ram": 8192,
    },
    "r6g.xlarge": {
        "bandwidth": 10,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6g.xlarge",
            "instancesku": "QCPDTQPKB7E87SYY",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6g.xlarge",
        "name": "r6g.xlarge",
        "ram": 32768,
    },
    "r6gd.12xlarge": {
        "bandwidth": 20,
        "disk": 2850,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "13500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6gd.12xlarge",
            "instancesku": "DQHFFY2N32GJ4KHG",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1425 NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6gd.12xlarge",
        "name": "r6gd.12xlarge",
        "ram": 393216,
    },
    "r6gd.16xlarge": {
        "bandwidth": 25,
        "disk": 3800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "18000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6gd.16xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-3",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1900 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6gd.16xlarge",
        "name": "r6gd.16xlarge",
        "ram": 524288,
    },
    "r6gd.2xlarge": {
        "bandwidth": 10,
        "disk": 475,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6gd.2xlarge",
            "instancesku": "NKKKMUC8VTHJGNJ3",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 475 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6gd.2xlarge",
        "name": "r6gd.2xlarge",
        "ram": 65536,
    },
    "r6gd.4xlarge": {
        "bandwidth": 10,
        "disk": 950,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6gd.4xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 950 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6gd.4xlarge",
        "name": "r6gd.4xlarge",
        "ram": 131072,
    },
    "r6gd.8xlarge": {
        "bandwidth": 12,
        "disk": 1900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "9000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6gd.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "12 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1900 NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6gd.8xlarge",
        "name": "r6gd.8xlarge",
        "ram": 262144,
    },
    "r6gd.large": {
        "bandwidth": 10,
        "disk": 118,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6gd.large",
            "instancesku": "E4YG7JC73SCP7EKG",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 118 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6gd.large",
        "name": "r6gd.large",
        "ram": 16384,
    },
    "r6gd.medium": {
        "bandwidth": 10,
        "disk": 59,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6gd.medium",
            "instancesku": "6UT3UMQBXYMNWFKU",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 59 NVMe SSD",
            "vcpu": "1",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6gd.medium",
        "name": "r6gd.medium",
        "ram": 8192,
    },
    "r6gd.xlarge": {
        "bandwidth": 10,
        "disk": 237,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6gd.xlarge",
            "instancesku": "T8TGZ93MC3QPTCKS",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 237 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6gd.xlarge",
        "name": "r6gd.xlarge",
        "ram": 32768,
    },
    "r6i.12xlarge": {
        "bandwidth": 18750,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "15000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6i.12xlarge",
            "instancesku": "KBT26J6CP63QFBHJ",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "18750 Megabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6i.12xlarge",
        "name": "r6i.12xlarge",
        "ram": 393216,
    },
    "r6i.16xlarge": {
        "bandwidth": 25000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6i.16xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "25000 Megabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "sa-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6i.16xlarge",
        "name": "r6i.16xlarge",
        "ram": 524288,
    },
    "r6i.24xlarge": {
        "bandwidth": 37500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "30000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6i.24xlarge",
            "instancesku": "KAZ35PTYHRGQXF48",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "37500 Megabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6i.24xlarge",
        "name": "r6i.24xlarge",
        "ram": 786432,
    },
    "r6i.2xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6i.2xlarge",
            "instancesku": "GCMURA7T4XN4SCDF",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6i.2xlarge",
        "name": "r6i.2xlarge",
        "ram": 65536,
    },
    "r6i.32xlarge": {
        "bandwidth": 50000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "40000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6i.32xlarge",
            "instancesku": "NAXJMGJV4T3ATM9C",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "1024 GiB",
            "networkPerformance": "50000 Megabit",
            "normalizationSizeFactor": "256",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-southeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "128",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6i.32xlarge",
        "name": "r6i.32xlarge",
        "ram": 1048576,
    },
    "r6i.4xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6i.4xlarge",
            "instancesku": "VT8HNM5SDM24MGDZ",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6i.4xlarge",
        "name": "r6i.4xlarge",
        "ram": 131072,
    },
    "r6i.8xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6i.8xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "12500 Megabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6i.8xlarge",
        "name": "r6i.8xlarge",
        "ram": 262144,
    },
    "r6i.large": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6i.large",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6i.large",
        "name": "r6i.large",
        "ram": 16384,
    },
    "r6i.xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6i.xlarge",
            "instancesku": "29ZSGZP5P8R5RKJF",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6i.xlarge",
        "name": "r6i.xlarge",
        "ram": 32768,
    },
    "r6id.12xlarge": {
        "bandwidth": 18750,
        "disk": 2850,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "15000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6id.12xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "18750 Megabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1425 NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6id.12xlarge",
        "name": "r6id.12xlarge",
        "ram": 393216,
    },
    "r6id.16xlarge": {
        "bandwidth": 25000,
        "disk": 3800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6id.16xlarge",
            "instancesku": "BBCJTJTX4TBRMRTH",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "25000 Megabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1900 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6id.16xlarge",
        "name": "r6id.16xlarge",
        "ram": 524288,
    },
    "r6id.24xlarge": {
        "bandwidth": 37500,
        "disk": 5700,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "30000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6id.24xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "37500 Megabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 1425 NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6id.24xlarge",
        "name": "r6id.24xlarge",
        "ram": 786432,
    },
    "r6id.2xlarge": {
        "bandwidth": 12500,
        "disk": 474,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6id.2xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 474 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6id.2xlarge",
        "name": "r6id.2xlarge",
        "ram": 65536,
    },
    "r6id.32xlarge": {
        "bandwidth": 50000,
        "disk": 7600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "40000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6id.32xlarge",
            "instancesku": "XYMMETJ9Z87XA8TR",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "1024 GiB",
            "networkPerformance": "50000 Megabit",
            "normalizationSizeFactor": "256",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 1900 NVMe SSD",
            "vcpu": "128",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6id.32xlarge",
        "name": "r6id.32xlarge",
        "ram": 1048576,
    },
    "r6id.4xlarge": {
        "bandwidth": 12500,
        "disk": 950,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6id.4xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 950 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6id.4xlarge",
        "name": "r6id.4xlarge",
        "ram": 131072,
    },
    "r6id.8xlarge": {
        "bandwidth": 12500,
        "disk": 1900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6id.8xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "12500 Megabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1900 NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6id.8xlarge",
        "name": "r6id.8xlarge",
        "ram": 262144,
    },
    "r6id.large": {
        "bandwidth": 12500,
        "disk": 118,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6id.large",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 118 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6id.large",
        "name": "r6id.large",
        "ram": 16384,
    },
    "r6id.xlarge": {
        "bandwidth": 12500,
        "disk": 237,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6id.xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 237 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6id.xlarge",
        "name": "r6id.xlarge",
        "ram": 32768,
    },
    "r6idn.12xlarge": {
        "bandwidth": 75000,
        "disk": 2850,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "37500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6idn.12xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "75000 Megabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1425 NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6idn.12xlarge",
        "name": "r6idn.12xlarge",
        "ram": 393216,
    },
    "r6idn.16xlarge": {
        "bandwidth": 100000,
        "disk": 3800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "50000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6idn.16xlarge",
            "instancesku": "6BH5ZTEN6R6YJ3YG",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "100000 Megabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1900 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6idn.16xlarge",
        "name": "r6idn.16xlarge",
        "ram": 524288,
    },
    "r6idn.24xlarge": {
        "bandwidth": 150000,
        "disk": 5700,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "75000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6idn.24xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "150000 Megabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 1425 NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6idn.24xlarge",
        "name": "r6idn.24xlarge",
        "ram": 786432,
    },
    "r6idn.2xlarge": {
        "bandwidth": 40000,
        "disk": 474,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 25000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6idn.2xlarge",
            "instancesku": "2CCXZVH7FXHA6NM3",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 40000 Megabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 474 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6idn.2xlarge",
        "name": "r6idn.2xlarge",
        "ram": 65536,
    },
    "r6idn.32xlarge": {
        "bandwidth": 200000,
        "disk": 7600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "100000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6idn.32xlarge",
            "instancesku": "C6KYJ3WR8DKTBUC6",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "1024 GiB",
            "networkPerformance": "200000 Megabit",
            "normalizationSizeFactor": "256",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 1900 NVMe SSD",
            "vcpu": "128",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6idn.32xlarge",
        "name": "r6idn.32xlarge",
        "ram": 1048576,
    },
    "r6idn.4xlarge": {
        "bandwidth": 50000,
        "disk": 950,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 25000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6idn.4xlarge",
            "instancesku": "5CN8UXJSHXERVCZU",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 50000 Megabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 950 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6idn.4xlarge",
        "name": "r6idn.4xlarge",
        "ram": 131072,
    },
    "r6idn.8xlarge": {
        "bandwidth": 50000,
        "disk": 1900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "25000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6idn.8xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "50000 Megabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1900 NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6idn.8xlarge",
        "name": "r6idn.8xlarge",
        "ram": 262144,
    },
    "r6idn.large": {
        "bandwidth": 25000,
        "disk": 118,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 25000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6idn.large",
            "instancesku": "5CDH8SVCHUP5NGE8",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 25000 Megabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 118 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6idn.large",
        "name": "r6idn.large",
        "ram": 16384,
    },
    "r6idn.xlarge": {
        "bandwidth": 30000,
        "disk": 237,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 25000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6idn.xlarge",
            "instancesku": "VYK62GH26Z3UYC53",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 30000 Megabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 237 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6idn.xlarge",
        "name": "r6idn.xlarge",
        "ram": 32768,
    },
    "r6in.12xlarge": {
        "bandwidth": 75000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "37500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6in.12xlarge",
            "instancesku": "PMW3R75PXPMWGBG2",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "75000 Megabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-southeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6in.12xlarge",
        "name": "r6in.12xlarge",
        "ram": 393216,
    },
    "r6in.16xlarge": {
        "bandwidth": 100000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "50000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6in.16xlarge",
            "instancesku": "EDNTTCS2UZXSQU8Q",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "100000 Megabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6in.16xlarge",
        "name": "r6in.16xlarge",
        "ram": 524288,
    },
    "r6in.24xlarge": {
        "bandwidth": 150000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "75000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6in.24xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "150000 Megabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6in.24xlarge",
        "name": "r6in.24xlarge",
        "ram": 786432,
    },
    "r6in.2xlarge": {
        "bandwidth": 40000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 25000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6in.2xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 40000 Megabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6in.2xlarge",
        "name": "r6in.2xlarge",
        "ram": 65536,
    },
    "r6in.32xlarge": {
        "bandwidth": 200000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "100000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6in.32xlarge",
            "instancesku": "HK566QER6ER83H8R",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "1024 GiB",
            "networkPerformance": "200000 Megabit",
            "normalizationSizeFactor": "256",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "128",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6in.32xlarge",
        "name": "r6in.32xlarge",
        "ram": 1048576,
    },
    "r6in.4xlarge": {
        "bandwidth": 50000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 25000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6in.4xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 50000 Megabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6in.4xlarge",
        "name": "r6in.4xlarge",
        "ram": 131072,
    },
    "r6in.8xlarge": {
        "bandwidth": 50000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "25000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6in.8xlarge",
            "instancesku": "BMWXPVMCNAC4V8NB",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "50000 Megabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6in.8xlarge",
        "name": "r6in.8xlarge",
        "ram": 262144,
    },
    "r6in.large": {
        "bandwidth": 25000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 25000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6in.large",
            "instancesku": "W8E4N62D3S8UQRGJ",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 25000 Megabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6in.large",
        "name": "r6in.large",
        "ram": 16384,
    },
    "r6in.xlarge": {
        "bandwidth": 30000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 25000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r6in.xlarge",
            "instancesku": "F9ARFW2XYSF59Y7G",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 30000 Megabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon 8375C (Ice Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "r6in.xlarge",
        "name": "r6in.xlarge",
        "ram": 32768,
    },
    "r7a.12xlarge": {
        "bandwidth": 18750,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "15000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7a.12xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "18750 Megabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7a.12xlarge",
        "name": "r7a.12xlarge",
        "ram": 393216,
    },
    "r7a.16xlarge": {
        "bandwidth": 25000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7a.16xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "25000 Megabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7a.16xlarge",
        "name": "r7a.16xlarge",
        "ram": 524288,
    },
    "r7a.24xlarge": {
        "bandwidth": 37500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "30000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7a.24xlarge",
            "instancesku": "RDZ28YFNQ8W3XZ48",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "37500 Megabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7a.24xlarge",
        "name": "r7a.24xlarge",
        "ram": 786432,
    },
    "r7a.2xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7a.2xlarge",
            "instancesku": "7B3NKAHPX6WJAATD",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7a.2xlarge",
        "name": "r7a.2xlarge",
        "ram": 65536,
    },
    "r7a.32xlarge": {
        "bandwidth": 50000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "40000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7a.32xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "1024 GiB",
            "networkPerformance": "50000 Megabit",
            "normalizationSizeFactor": "256",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "128",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7a.32xlarge",
        "name": "r7a.32xlarge",
        "ram": 1048576,
    },
    "r7a.48xlarge": {
        "bandwidth": 50000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "40000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7a.48xlarge",
            "instancesku": "3SR9U89YWT5BCY6S",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "1536 GiB",
            "networkPerformance": "50000 Megabit",
            "normalizationSizeFactor": "384",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "192",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7a.48xlarge",
        "name": "r7a.48xlarge",
        "ram": 1572864,
    },
    "r7a.4xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7a.4xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7a.4xlarge",
        "name": "r7a.4xlarge",
        "ram": 131072,
    },
    "r7a.8xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7a.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "12500 Megabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7a.8xlarge",
        "name": "r7a.8xlarge",
        "ram": 262144,
    },
    "r7a.large": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7a.large",
            "instancesku": "EGGQMFNY6FSM369C",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7a.large",
        "name": "r7a.large",
        "ram": 16384,
    },
    "r7a.medium": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7a.medium",
            "instancesku": "3UZ6JNF8DAUTDUHR",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "1",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7a.medium",
        "name": "r7a.medium",
        "ram": 8192,
    },
    "r7a.xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7a.xlarge",
            "instancesku": "MKGAUZQRFHDZD57A",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AMD EPYC 9R14 Processor",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AMD Turbo; AVX; AVX2; AVX512",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7a.xlarge",
        "name": "r7a.xlarge",
        "ram": 32768,
    },
    "r7g.12xlarge": {
        "bandwidth": 22500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "15000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7g.12xlarge",
            "instancesku": "GJFFKNW2JFBP36PV",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "22500 Megabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7g.12xlarge",
        "name": "r7g.12xlarge",
        "ram": 393216,
    },
    "r7g.16xlarge": {
        "bandwidth": 30,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7g.16xlarge",
            "instancesku": "KHYZNNC6CFHUE2YN",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "30 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7g.16xlarge",
        "name": "r7g.16xlarge",
        "ram": 524288,
    },
    "r7g.2xlarge": {
        "bandwidth": 15,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7g.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 15 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7g.2xlarge",
        "name": "r7g.2xlarge",
        "ram": 65536,
    },
    "r7g.4xlarge": {
        "bandwidth": 15,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7g.4xlarge",
            "instancesku": "GM47ZT3HEKXK2RKU",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 15 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7g.4xlarge",
        "name": "r7g.4xlarge",
        "ram": 131072,
    },
    "r7g.8xlarge": {
        "bandwidth": 15,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7g.8xlarge",
            "instancesku": "5EKNZY9FZ533ACBN",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "15 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-north-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7g.8xlarge",
        "name": "r7g.8xlarge",
        "ram": 262144,
    },
    "r7g.large": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7g.large",
            "instancesku": "ECFYSTX74KPXVVBN",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7g.large",
        "name": "r7g.large",
        "ram": 16384,
    },
    "r7g.medium": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7g.medium",
            "instancesku": "5WTHGR8KRJSXYPA8",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "1",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7g.medium",
        "name": "r7g.medium",
        "ram": 8192,
    },
    "r7g.xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7g.xlarge",
            "instancesku": "G2VJQNCYXP7J952P",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7g.xlarge",
        "name": "r7g.xlarge",
        "ram": 32768,
    },
    "r7gd.12xlarge": {
        "bandwidth": 22500,
        "disk": 2850,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "15000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7gd.12xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "22500 Megabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1425 NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7gd.12xlarge",
        "name": "r7gd.12xlarge",
        "ram": 393216,
    },
    "r7gd.16xlarge": {
        "bandwidth": 30,
        "disk": 3800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7gd.16xlarge",
            "instancesku": "TVR3W2FB44WKUYV6",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "30 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1900 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7gd.16xlarge",
        "name": "r7gd.16xlarge",
        "ram": 524288,
    },
    "r7gd.2xlarge": {
        "bandwidth": 15,
        "disk": 475,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7gd.2xlarge",
            "instancesku": "JUT7C97SGGH6FUFE",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 15 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 475 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7gd.2xlarge",
        "name": "r7gd.2xlarge",
        "ram": 65536,
    },
    "r7gd.4xlarge": {
        "bandwidth": 15,
        "disk": 950,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7gd.4xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 15 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 950 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7gd.4xlarge",
        "name": "r7gd.4xlarge",
        "ram": 131072,
    },
    "r7gd.8xlarge": {
        "bandwidth": 15,
        "disk": 1900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7gd.8xlarge",
            "instancesku": "HRPNCHQ7TZK3VWSX",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "15 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1900 NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7gd.8xlarge",
        "name": "r7gd.8xlarge",
        "ram": 262144,
    },
    "r7gd.large": {
        "bandwidth": 12500,
        "disk": 118,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7gd.large",
            "instancesku": "TQWXHKT9PHTUXMS2",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 118 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7gd.large",
        "name": "r7gd.large",
        "ram": 16384,
    },
    "r7gd.medium": {
        "bandwidth": 12500,
        "disk": 59,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7gd.medium",
            "instancesku": "P88JQHEPPAWA6HWH",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 59 NVMe SSD",
            "vcpu": "1",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7gd.medium",
        "name": "r7gd.medium",
        "ram": 8192,
    },
    "r7gd.xlarge": {
        "bandwidth": 12500,
        "disk": 237,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7gd.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton3 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 237 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7gd.xlarge",
        "name": "r7gd.xlarge",
        "ram": 32768,
    },
    "r7i.12xlarge": {
        "bandwidth": 18750,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "15000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7i.12xlarge",
            "instancesku": "PUAUGZRCARPYB6F2",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "18750 Megabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7i.12xlarge",
        "name": "r7i.12xlarge",
        "ram": 393216,
    },
    "r7i.16xlarge": {
        "bandwidth": 25000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7i.16xlarge",
            "instancesku": "K8DRJU2WDAMWWK5U",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "25000 Megabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7i.16xlarge",
        "name": "r7i.16xlarge",
        "ram": 524288,
    },
    "r7i.24xlarge": {
        "bandwidth": 37500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "30000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7i.24xlarge",
            "instancesku": "HCZYPAMGW7QGVAQ3",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "37500 Megabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7i.24xlarge",
        "name": "r7i.24xlarge",
        "ram": 786432,
    },
    "r7i.2xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7i.2xlarge",
            "instancesku": "SZBRE2RJ4H5KATWH",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7i.2xlarge",
        "name": "r7i.2xlarge",
        "ram": 65536,
    },
    "r7i.48xlarge": {
        "bandwidth": 50000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "40000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7i.48xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "1536 GiB",
            "networkPerformance": "50000 Megabit",
            "normalizationSizeFactor": "384",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "ap-southeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "192",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7i.48xlarge",
        "name": "r7i.48xlarge",
        "ram": 1572864,
    },
    "r7i.4xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7i.4xlarge",
            "instancesku": "387ET9W6QRE2GUMT",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7i.4xlarge",
        "name": "r7i.4xlarge",
        "ram": 131072,
    },
    "r7i.8xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7i.8xlarge",
            "instancesku": "YUBX7VNGA8N2NN3H",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "12500 Megabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "eu-west-3",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7i.8xlarge",
        "name": "r7i.8xlarge",
        "ram": 262144,
    },
    "r7i.large": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7i.large",
            "instancesku": "GYCBG8CFZPAFMUVW",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7i.large",
        "name": "r7i.large",
        "ram": 16384,
    },
    "r7i.xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.2 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7i.xlarge",
            "instancesku": "J68SFC7TSST5CN33",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo; Intel AMX",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7i.xlarge",
        "name": "r7i.xlarge",
        "ram": 32768,
    },
    "r7iz.12xlarge": {
        "bandwidth": 25000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.9 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "19000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7iz.12xlarge",
            "instancesku": "23CZCQDT39K4N3KW",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "25000 Megabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AMX; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7iz.12xlarge",
        "name": "r7iz.12xlarge",
        "ram": 393216,
    },
    "r7iz.16xlarge": {
        "bandwidth": 25000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.9 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7iz.16xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "25000 Megabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AMX; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7iz.16xlarge",
        "name": "r7iz.16xlarge",
        "ram": 524288,
    },
    "r7iz.2xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.9 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7iz.2xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AMX; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7iz.2xlarge",
        "name": "r7iz.2xlarge",
        "ram": 65536,
    },
    "r7iz.32xlarge": {
        "bandwidth": 50000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.9 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "40000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7iz.32xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "1024 GiB",
            "networkPerformance": "50000 Megabit",
            "normalizationSizeFactor": "256",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AMX; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "128",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7iz.32xlarge",
        "name": "r7iz.32xlarge",
        "ram": 1048576,
    },
    "r7iz.4xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.9 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7iz.4xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AMX; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7iz.4xlarge",
        "name": "r7iz.4xlarge",
        "ram": 131072,
    },
    "r7iz.8xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.9 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7iz.8xlarge",
            "instancesku": "W75AYZKNFC94DB6X",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "12500 Megabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AMX; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7iz.8xlarge",
        "name": "r7iz.8xlarge",
        "ram": 262144,
    },
    "r7iz.large": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.9 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7iz.large",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AMX; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7iz.large",
        "name": "r7iz.large",
        "ram": 16384,
    },
    "r7iz.xlarge": {
        "bandwidth": 12500,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.9 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r7iz.xlarge",
            "instancesku": "7UQ2FKMSY3B4NNSV",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 12500 Megabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AMX; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "r7iz.xlarge",
        "name": "r7iz.xlarge",
        "ram": 32768,
    },
    "r8g.12xlarge": {
        "bandwidth": 22,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "15000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r8g.12xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "22.5 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "r8g.12xlarge",
        "name": "r8g.12xlarge",
        "ram": 393216,
    },
    "r8g.16xlarge": {
        "bandwidth": 30,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r8g.16xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "30 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "r8g.16xlarge",
        "name": "r8g.16xlarge",
        "ram": 524288,
    },
    "r8g.24xlarge": {
        "bandwidth": 40,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "30000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r8g.24xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "40 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "r8g.24xlarge",
        "name": "r8g.24xlarge",
        "ram": 786432,
    },
    "r8g.2xlarge": {
        "bandwidth": 15,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r8g.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 15 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "r8g.2xlarge",
        "name": "r8g.2xlarge",
        "ram": 65536,
    },
    "r8g.48xlarge": {
        "bandwidth": 50,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "40000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r8g.48xlarge",
            "instancesku": "SKK3R9PTMH7JC22C",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "1536 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "384",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "192",
            "vpcnetworkingsupport": "true",
        },
        "id": "r8g.48xlarge",
        "name": "r8g.48xlarge",
        "ram": 1572864,
    },
    "r8g.4xlarge": {
        "bandwidth": 15,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r8g.4xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 15 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "r8g.4xlarge",
        "name": "r8g.4xlarge",
        "ram": 131072,
    },
    "r8g.8xlarge": {
        "bandwidth": 15,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r8g.8xlarge",
            "instancesku": "HSKTJRMDSCHDSSPK",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "15 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "r8g.8xlarge",
        "name": "r8g.8xlarge",
        "ram": 262144,
    },
    "r8g.large": {
        "bandwidth": 12,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r8g.large",
            "instancesku": "5794PT5PDVXVMAWQ",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 12.5 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "r8g.large",
        "name": "r8g.large",
        "ram": 16384,
    },
    "r8g.medium": {
        "bandwidth": 12,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r8g.medium",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 12.5 Gigabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "1",
            "vpcnetworkingsupport": "true",
        },
        "id": "r8g.medium",
        "name": "r8g.medium",
        "ram": 8192,
    },
    "r8g.xlarge": {
        "bandwidth": 12,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "r8g.xlarge",
            "instancesku": "9Y46XFXK69WBSHFC",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 12.5 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "r8g.xlarge",
        "name": "r8g.xlarge",
        "ram": 32768,
    },
    "t1.micro": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "No",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Micro instances",
            "instanceType": "t1.micro",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "0.613 GiB",
            "networkPerformance": "Very Low",
            "normalizationSizeFactor": "0.5",
            "physicalProcessor": "Variable",
            "processorArchitecture": "32-bit or 64-bit",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "1",
            "vpcnetworkingsupport": "true",
        },
        "id": "t1.micro",
        "name": "t1.micro",
        "ram": 627,
    },
    "t2.2xlarge": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.3 GHz",
            "currentGeneration": "Yes",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "t2.2xlarge",
            "instancesku": "WTCBDF8NTYFUN3ES",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "t2.2xlarge",
        "name": "t2.2xlarge",
        "ram": 32768,
    },
    "t2.large": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.3 GHz",
            "currentGeneration": "Yes",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "t2.large",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Low to Moderate",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "regionCode": "sa-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "t2.large",
        "name": "t2.large",
        "ram": 8192,
    },
    "t2.medium": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.3 GHz",
            "currentGeneration": "Yes",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "t2.medium",
            "instancesku": "DY99AWFYB6SX74KD",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Low to Moderate",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "32-bit or 64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "regionCode": "us-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "t2.medium",
        "name": "t2.medium",
        "ram": 4096,
    },
    "t2.micro": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.3 GHz",
            "currentGeneration": "Yes",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "t2.micro",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "1 GiB",
            "networkPerformance": "Low to Moderate",
            "normalizationSizeFactor": "0.5",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "32-bit or 64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "regionCode": "eu-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "1",
            "vpcnetworkingsupport": "true",
        },
        "id": "t2.micro",
        "name": "t2.micro",
        "ram": 1024,
    },
    "t2.nano": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.3 GHz",
            "currentGeneration": "Yes",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "t2.nano",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "0.5 GiB",
            "networkPerformance": "Low",
            "normalizationSizeFactor": "0.25",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "32-bit or 64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "regionCode": "us-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "1",
            "vpcnetworkingsupport": "true",
        },
        "id": "t2.nano",
        "name": "t2.nano",
        "ram": 512,
    },
    "t2.small": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.3 GHz",
            "currentGeneration": "Yes",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "t2.small",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "2 GiB",
            "networkPerformance": "Low to Moderate",
            "normalizationSizeFactor": "1",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "32-bit or 64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "1",
            "vpcnetworkingsupport": "true",
        },
        "id": "t2.small",
        "name": "t2.small",
        "ram": 2048,
    },
    "t2.xlarge": {
        "bandwidth": None,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.3 GHz",
            "currentGeneration": "Yes",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "t2.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Moderate",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Family",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "t2.xlarge",
        "name": "t2.xlarge",
        "ram": 16384,
    },
    "t3.2xlarge": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2780 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "t3.2xlarge",
            "instancesku": "CAB4TUSN8KCWWMKW",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Skylake E5 2686 v5",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-3",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "t3.2xlarge",
        "name": "t3.2xlarge",
        "ram": 32768,
    },
    "t3.large": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2780 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "t3.large",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Skylake E5 2686 v5",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-3",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "t3.large",
        "name": "t3.large",
        "ram": 8192,
    },
    "t3.medium": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2085 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "t3.medium",
            "instancesku": "YZ5V7FVMR7UMXETK",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "Intel Skylake E5 2686 v5",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "t3.medium",
        "name": "t3.medium",
        "ram": 4096,
    },
    "t3.micro": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2085 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "t3.micro",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "1 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "0.5",
            "physicalProcessor": "Intel Skylake E5 2686 v5",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "sa-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "t3.micro",
        "name": "t3.micro",
        "ram": 1024,
    },
    "t3.nano": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2085 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "t3.nano",
            "instancesku": "WEW9WBCCJASKWNJU",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "0.5 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "0.25",
            "physicalProcessor": "Intel Skylake E5 2686 v5",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "t3.nano",
        "name": "t3.nano",
        "ram": 512,
    },
    "t3.small": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2085 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "t3.small",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "2 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "1",
            "physicalProcessor": "Intel Skylake E5 2686 v5",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "t3.small",
        "name": "t3.small",
        "ram": 2048,
    },
    "t3.xlarge": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2780 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "t3.xlarge",
            "instancesku": "24GPMCAUP9P4GSC6",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Skylake E5 2686 v5",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "t3.xlarge",
        "name": "t3.xlarge",
        "ram": 16384,
    },
    "t3a.2xlarge": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2780 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "t3a.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "eu-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "t3a.2xlarge",
        "name": "t3a.2xlarge",
        "ram": 32768,
    },
    "t3a.large": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2780 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "t3a.large",
            "instancesku": "9W9WUN8874RQJPYR",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "t3a.large",
        "name": "t3a.large",
        "ram": 8192,
    },
    "t3a.medium": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2085 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "t3a.medium",
            "instancesku": "UMAHZUQPRB2BAWBS",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "t3a.medium",
        "name": "t3a.medium",
        "ram": 4096,
    },
    "t3a.micro": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2085 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "t3a.micro",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "1 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "0.5",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "t3a.micro",
        "name": "t3a.micro",
        "ram": 1024,
    },
    "t3a.nano": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2085 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "t3a.nano",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "0.5 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "0.25",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "sa-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "t3a.nano",
        "name": "t3a.nano",
        "ram": 512,
    },
    "t3a.small": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2085 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "t3a.small",
            "instancesku": "R84E3TQPGUP973B4",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "2 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "1",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "t3a.small",
        "name": "t3a.small",
        "ram": 2048,
    },
    "t3a.xlarge": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2780 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "t3a.xlarge",
            "instancesku": "YUQH22HS49VU8AJK",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AMD EPYC 7571",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; AVX2; AMD Turbo",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "t3a.xlarge",
        "name": "t3a.xlarge",
        "ram": 16384,
    },
    "t4g.2xlarge": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2780 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "t4g.2xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "t4g.2xlarge",
        "name": "t4g.2xlarge",
        "ram": 32768,
    },
    "t4g.large": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2780 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "t4g.large",
            "instancesku": "DP29NV5PDAV9G9SM",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "8 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "t4g.large",
        "name": "t4g.large",
        "ram": 8192,
    },
    "t4g.medium": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2085 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "t4g.medium",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "4 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "t4g.medium",
        "name": "t4g.medium",
        "ram": 4096,
    },
    "t4g.micro": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2085 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "t4g.micro",
            "instancesku": "BQXJ8V9R6Y2TPCGG",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "1 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "0.5",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "t4g.micro",
        "name": "t4g.micro",
        "ram": 1024,
    },
    "t4g.nano": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2085 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "t4g.nano",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "0.5 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "0.25",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "t4g.nano",
        "name": "t4g.nano",
        "ram": 512,
    },
    "t4g.small": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2085 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "t4g.small",
            "instancesku": "Y33CHC2PYKBQ7U2C",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "2 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "1",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "t4g.small",
        "name": "t4g.small",
        "ram": 2048,
    },
    "t4g.xlarge": {
        "bandwidth": 5,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2780 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "General purpose",
            "instanceType": "t4g.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 5 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "t4g.xlarge",
        "name": "t4g.xlarge",
        "ram": 16384,
    },
    "trn1.2xlarge": {
        "bandwidth": 12500,
        "disk": 475,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 20 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "32 GB",
            "instanceFamily": "Machine Learning ASIC Instances",
            "instanceType": "trn1.2xlarge",
            "instancesku": "CSMRGVUPNG9UETH8",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "12500 Megabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Scalable (Icelake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 475 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "trn1.2xlarge",
        "name": "trn1.2xlarge",
        "ram": 32768,
    },
    "trn1.32xlarge": {
        "bandwidth": 800,
        "disk": 7600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 80 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "512 GB",
            "instanceFamily": "Machine Learning ASIC Instances",
            "instanceType": "trn1.32xlarge",
            "instancesku": "XPNPN2N6VY2GVQ9X",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "800 Gigabit",
            "normalizationSizeFactor": "256",
            "physicalProcessor": "Intel Xeon Scalable (Icelake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 1900 NVMe SSD",
            "vcpu": "128",
            "vpcnetworkingsupport": "true",
        },
        "id": "trn1.32xlarge",
        "name": "trn1.32xlarge",
        "ram": 524288,
    },
    "trn1n.32xlarge": {
        "bandwidth": 1600,
        "disk": 7600,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "80 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "512 GB",
            "instanceFamily": "Machine Learning ASIC Instances",
            "instanceType": "trn1n.32xlarge",
            "instancesku": "49HTVZD5D4J22XE8",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "1600 Gigabit",
            "normalizationSizeFactor": "256",
            "physicalProcessor": "Intel Xeon Scalable (Icelake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "4 x 1900 NVMe SSD",
            "vcpu": "128",
            "vpcnetworkingsupport": "true",
        },
        "id": "trn1n.32xlarge",
        "name": "trn1n.32xlarge",
        "ram": 524288,
    },
    "trn2.48xlarge": {
        "bandwidth": 3200,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "Up to 3.9 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "80 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "1536 GB HBM3",
            "instanceFamily": "Machine Learning ASIC Instances",
            "instanceType": "trn2.48xlarge",
            "instancesku": "A7RS779QCEMNJV96",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "CapacityBlock",
            "memory": "2048 GiB",
            "networkPerformance": "3200 Gigabit",
            "normalizationSizeFactor": "384",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AMX; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "192",
            "vpcnetworkingsupport": "true",
        },
        "id": "trn2.48xlarge",
        "name": "trn2.48xlarge",
        "ram": 2097152,
    },
    "u-12tb1.112xlarge": {
        "bandwidth": 100,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "38000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "u-12tb1.112xlarge",
            "instancesku": "REVYCR94A28EU37G",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "12288 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "NA",
            "physicalProcessor": "Intel Xeon Scalable (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "448",
            "vpcnetworkingsupport": "true",
        },
        "id": "u-12tb1.112xlarge",
        "name": "u-12tb1.112xlarge",
        "ram": 12582912,
    },
    "u-18tb1.112xlarge": {
        "bandwidth": 100,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "38000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "u-18tb1.112xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "18432 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "NA",
            "physicalProcessor": "Intel Xeon Platinum 8280L (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "448",
            "vpcnetworkingsupport": "true",
        },
        "id": "u-18tb1.112xlarge",
        "name": "u-18tb1.112xlarge",
        "ram": 18874368,
    },
    "u-24tb1.112xlarge": {
        "bandwidth": 100,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "38000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "u-24tb1.112xlarge",
            "instancesku": "3B3GWRWTH7FF4A6F",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "24576 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "NA",
            "physicalProcessor": "Intel Xeon Platinum 8280L (Cascade Lake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "448",
            "vpcnetworkingsupport": "true",
        },
        "id": "u-24tb1.112xlarge",
        "name": "u-24tb1.112xlarge",
        "ram": 25165824,
    },
    "u-3tb1.56xlarge": {
        "bandwidth": 50,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.1 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "19000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "u-3tb1.56xlarge",
            "instancesku": "SAPP6Y5NAHGYJNTE",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "3072 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "448",
            "physicalProcessor": "Intel Xeon Scalable (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX2; Intel AVX; Intel AVX2",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "224",
            "vpcnetworkingsupport": "true",
        },
        "id": "u-3tb1.56xlarge",
        "name": "u-3tb1.56xlarge",
        "ram": 3145728,
    },
    "u-6tb1.112xlarge": {
        "bandwidth": 100,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "38000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "u-6tb1.112xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "6144 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "NA",
            "physicalProcessor": "Intel Xeon Scalable (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; Intel AVX; Intel AVX2",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "448",
            "vpcnetworkingsupport": "true",
        },
        "id": "u-6tb1.112xlarge",
        "name": "u-6tb1.112xlarge",
        "ram": 6291456,
    },
    "u-6tb1.56xlarge": {
        "bandwidth": 100,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "38000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "u-6tb1.56xlarge",
            "instancesku": "PVJMNB8VSNPYZ6GP",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "6144 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "NA",
            "physicalProcessor": "Intel Xeon Scalable (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; Intel AVX; Intel AVX2",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "224",
            "vpcnetworkingsupport": "true",
        },
        "id": "u-6tb1.56xlarge",
        "name": "u-6tb1.56xlarge",
        "ram": 6291456,
    },
    "u-9tb1.112xlarge": {
        "bandwidth": 100,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "38000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "u-9tb1.112xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "9216 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "NA",
            "physicalProcessor": "Intel Xeon Scalable (Skylake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "ap-southeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "448",
            "vpcnetworkingsupport": "true",
        },
        "id": "u-9tb1.112xlarge",
        "name": "u-9tb1.112xlarge",
        "ram": 9437184,
    },
    "u7i-12tb.224xlarge": {
        "bandwidth": 100,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.9 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "60 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "u7i-12tb.224xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "12288 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "1792",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "896",
            "vpcnetworkingsupport": "true",
        },
        "id": "u7i-12tb.224xlarge",
        "name": "u7i-12tb.224xlarge",
        "ram": 12582912,
    },
    "u7i-6tb.112xlarge": {
        "bandwidth": 100,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.9 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "60 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "u7i-6tb.112xlarge",
            "instancesku": "JREF953XJBJQXJNH",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "6144 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "896",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "448",
            "vpcnetworkingsupport": "true",
        },
        "id": "u7i-6tb.112xlarge",
        "name": "u7i-6tb.112xlarge",
        "ram": 6291456,
    },
    "u7i-8tb.112xlarge": {
        "bandwidth": 100,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.9 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "60 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "u7i-8tb.112xlarge",
            "instancesku": "Y8G9PN7FHBAQMPX6",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "8192 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "896",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "448",
            "vpcnetworkingsupport": "true",
        },
        "id": "u7i-8tb.112xlarge",
        "name": "u7i-8tb.112xlarge",
        "ram": 8388608,
    },
    "u7in-16tb.224xlarge": {
        "bandwidth": 200,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.9 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "100 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "u7in-16tb.224xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16384 GiB",
            "networkPerformance": "200 Gigabit",
            "normalizationSizeFactor": "1792",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "32-bit or 64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "896",
            "vpcnetworkingsupport": "true",
        },
        "id": "u7in-16tb.224xlarge",
        "name": "u7in-16tb.224xlarge",
        "ram": 16777216,
    },
    "u7in-24tb.224xlarge": {
        "bandwidth": 200,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.9 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "100 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "u7in-24tb.224xlarge",
            "instancesku": "2QFYY9495W6NNN5X",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "24576 GiB",
            "networkPerformance": "200 Gigabit",
            "normalizationSizeFactor": "1792",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "896",
            "vpcnetworkingsupport": "true",
        },
        "id": "u7in-24tb.224xlarge",
        "name": "u7in-24tb.224xlarge",
        "ram": 25165824,
    },
    "u7in-32tb.224xlarge": {
        "bandwidth": 200,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.9 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "100 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "u7in-32tb.224xlarge",
            "instancesku": "63TTCGZUVQKPK53D",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32768 GiB",
            "networkPerformance": "200 Gigabit",
            "normalizationSizeFactor": "1792",
            "physicalProcessor": "Intel Xeon Scalable (Sapphire Rapids)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "896",
            "vpcnetworkingsupport": "true",
        },
        "id": "u7in-32tb.224xlarge",
        "name": "u7in-32tb.224xlarge",
        "ram": 33554432,
    },
    "vt1.24xlarge": {
        "bandwidth": 25000,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "19000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Media Accelerator Instances",
            "instanceType": "vt1.24xlarge",
            "instancesku": "9GT74VMCJEP299FD",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "25000 Megabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Platinum 8259CL",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "vt1.24xlarge",
        "name": "vt1.24xlarge",
        "ram": 196608,
    },
    "vt1.3xlarge": {
        "bandwidth": 3125,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "2375 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Media Accelerator Instances",
            "instanceType": "vt1.3xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "24 GiB",
            "networkPerformance": "3125 Megabit",
            "normalizationSizeFactor": "24",
            "physicalProcessor": "Intel Xeon Platinum 8259CL",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "12",
            "vpcnetworkingsupport": "true",
        },
        "id": "vt1.3xlarge",
        "name": "vt1.3xlarge",
        "ram": 24576,
    },
    "vt1.6xlarge": {
        "bandwidth": 6250,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4750 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "No",
            "gpuMemory": "NA",
            "instanceFamily": "Media Accelerator Instances",
            "instanceType": "vt1.6xlarge",
            "instancesku": "PY3K83PMVBEUYB32",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "48 GiB",
            "networkPerformance": "6250 Megabit",
            "normalizationSizeFactor": "48",
            "physicalProcessor": "Intel Xeon Platinum 8259CL",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "24",
            "vpcnetworkingsupport": "true",
        },
        "id": "vt1.6xlarge",
        "name": "vt1.6xlarge",
        "ram": 49152,
    },
    "x1.16xlarge": {
        "bandwidth": None,
        "disk": 1920,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "174.5",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x1.16xlarge",
            "instancesku": "7DHPBJ9UTU6S6WK4",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "976 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "High Frequency Intel Xeon E7-8880 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1920 SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "x1.16xlarge",
        "name": "x1.16xlarge",
        "ram": 999424,
    },
    "x1.32xlarge": {
        "bandwidth": None,
        "disk": 3840,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14000 Mbps",
            "ecu": "349",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x1.32xlarge",
            "instancesku": "STYK7K6FHHBSRAHM",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "1952 GiB",
            "networkPerformance": "High",
            "normalizationSizeFactor": "256",
            "physicalProcessor": "High Frequency Intel Xeon E7-8880 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel Turbo",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1920 SSD",
            "vcpu": "128",
            "vpcnetworkingsupport": "true",
        },
        "id": "x1.32xlarge",
        "name": "x1.32xlarge",
        "ram": 1998848,
    },
    "x1e.16xlarge": {
        "bandwidth": 10,
        "disk": 1920,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "179",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x1e.16xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "1952 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "High Frequency Intel Xeon E7-8880 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1920 SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "x1e.16xlarge",
        "name": "x1e.16xlarge",
        "ram": 1998848,
    },
    "x1e.2xlarge": {
        "bandwidth": 10,
        "disk": 240,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "1000 Mbps",
            "ecu": "23",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x1e.2xlarge",
            "instancesku": "BKRMPKCEC6DFW9NP",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "244 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "High Frequency Intel Xeon E7-8880 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 240 SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "x1e.2xlarge",
        "name": "x1e.2xlarge",
        "ram": 249856,
    },
    "x1e.32xlarge": {
        "bandwidth": 25,
        "disk": 3840,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14000 Mbps",
            "ecu": "340",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x1e.32xlarge",
            "instancesku": "YAS88XT2MXM4B2ZV",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "3904 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "256",
            "physicalProcessor": "High Frequency Intel Xeon E7-8880 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2",
            "regionCode": "sa-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1920 SSD",
            "vcpu": "128",
            "vpcnetworkingsupport": "true",
        },
        "id": "x1e.32xlarge",
        "name": "x1e.32xlarge",
        "ram": 3997696,
    },
    "x1e.4xlarge": {
        "bandwidth": 10,
        "disk": 480,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "1750 Mbps",
            "ecu": "47",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x1e.4xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "488 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "High Frequency Intel Xeon E7-8880 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 480 SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "x1e.4xlarge",
        "name": "x1e.4xlarge",
        "ram": 499712,
    },
    "x1e.8xlarge": {
        "bandwidth": 10,
        "disk": 960,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "91",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x1e.8xlarge",
            "instancesku": "QDBZKZSM8J22BVE7",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "976 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "High Frequency Intel Xeon E7-8880 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 960 SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "x1e.8xlarge",
        "name": "x1e.8xlarge",
        "ram": 999424,
    },
    "x1e.xlarge": {
        "bandwidth": 10,
        "disk": 120,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.3 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "500 Mbps",
            "ecu": "12",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x1e.xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "122 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "High Frequency Intel Xeon E7-8880 v3 (Haswell)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 120 SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "x1e.xlarge",
        "name": "x1e.xlarge",
        "ram": 124928,
    },
    "x2gd.12xlarge": {
        "bandwidth": 20,
        "disk": 2850,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14250 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x2gd.12xlarge",
            "instancesku": "YHYZ8JNVTJ3DYNPQ",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "20 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1425 SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "x2gd.12xlarge",
        "name": "x2gd.12xlarge",
        "ram": 786432,
    },
    "x2gd.16xlarge": {
        "bandwidth": 25,
        "disk": 3800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "19000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x2gd.16xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "1024 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1900 SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "x2gd.16xlarge",
        "name": "x2gd.16xlarge",
        "ram": 1048576,
    },
    "x2gd.2xlarge": {
        "bandwidth": 10,
        "disk": 474,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4750 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x2gd.2xlarge",
            "instancesku": "6SJMXRTFPNNB7XBT",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 474 SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "x2gd.2xlarge",
        "name": "x2gd.2xlarge",
        "ram": 131072,
    },
    "x2gd.4xlarge": {
        "bandwidth": 10,
        "disk": 950,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4750 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x2gd.4xlarge",
            "instancesku": "KTMUE4YKBZGAERNT",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 950 SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "x2gd.4xlarge",
        "name": "x2gd.4xlarge",
        "ram": 262144,
    },
    "x2gd.8xlarge": {
        "bandwidth": 12,
        "disk": 1900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "9500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x2gd.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "12 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1900 SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "x2gd.8xlarge",
        "name": "x2gd.8xlarge",
        "ram": 524288,
    },
    "x2gd.large": {
        "bandwidth": 10,
        "disk": 118,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4750 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x2gd.large",
            "instancesku": "NGCWNHNPM4YG27T7",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 118 SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "x2gd.large",
        "name": "x2gd.large",
        "ram": 32768,
    },
    "x2gd.medium": {
        "bandwidth": 10,
        "disk": 59,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4750 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x2gd.medium",
            "instancesku": "NGA7RTWTM43FWF3H",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 59 SSD",
            "vcpu": "1",
            "vpcnetworkingsupport": "true",
        },
        "id": "x2gd.medium",
        "name": "x2gd.medium",
        "ram": 16384,
    },
    "x2gd.xlarge": {
        "bandwidth": 10,
        "disk": 237,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 4750 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x2gd.xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton2 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 237 SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "x2gd.xlarge",
        "name": "x2gd.xlarge",
        "ram": 65536,
    },
    "x2idn.16xlarge": {
        "bandwidth": 50,
        "disk": 1900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "40 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x2idn.16xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "1024 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon Scalable (Icelake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-3",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1900 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "x2idn.16xlarge",
        "name": "x2idn.16xlarge",
        "ram": 1048576,
    },
    "x2idn.24xlarge": {
        "bandwidth": 75,
        "disk": 2850,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "60 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x2idn.24xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "1536 GiB",
            "networkPerformance": "75 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Scalable (Icelake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1425 NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "x2idn.24xlarge",
        "name": "x2idn.24xlarge",
        "ram": 1572864,
    },
    "x2idn.32xlarge": {
        "bandwidth": 100,
        "disk": 3800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "80 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x2idn.32xlarge",
            "instancesku": "AWRN7KUZ5WUQ9AE5",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "2048 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "256",
            "physicalProcessor": "Intel Xeon Scalable (Icelake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "AVX; Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ca-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1900 NVMe SSD",
            "vcpu": "128",
            "vpcnetworkingsupport": "true",
        },
        "id": "x2idn.32xlarge",
        "name": "x2idn.32xlarge",
        "ram": 2097152,
    },
    "x2iedn.16xlarge": {
        "bandwidth": 50,
        "disk": 1900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "40000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x2iedn.16xlarge",
            "instancesku": "73JJFGR8GMZVC524",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "2048 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "Intel Xeon Scalable (Icelake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 1900 NVMe SSD",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "x2iedn.16xlarge",
        "name": "x2iedn.16xlarge",
        "ram": 2097152,
    },
    "x2iedn.24xlarge": {
        "bandwidth": 75,
        "disk": 2850,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "60000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x2iedn.24xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "3072 GiB",
            "networkPerformance": "75 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "Intel Xeon Scalable (Icelake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1425 NVMe SSD",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "x2iedn.24xlarge",
        "name": "x2iedn.24xlarge",
        "ram": 3145728,
    },
    "x2iedn.2xlarge": {
        "bandwidth": 25,
        "disk": 237,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x2iedn.2xlarge",
            "instancesku": "63BDVCXPJJ8Q4CK7",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Scalable (Icelake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 237 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "x2iedn.2xlarge",
        "name": "x2iedn.2xlarge",
        "ram": 262144,
    },
    "x2iedn.32xlarge": {
        "bandwidth": 100,
        "disk": 3800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "80000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x2iedn.32xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "4096 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "256",
            "physicalProcessor": "Intel Xeon Scalable (Icelake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-south-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 1900 NVMe SSD",
            "vcpu": "128",
            "vpcnetworkingsupport": "true",
        },
        "id": "x2iedn.32xlarge",
        "name": "x2iedn.32xlarge",
        "ram": 4194304,
    },
    "x2iedn.4xlarge": {
        "bandwidth": 25,
        "disk": 475,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x2iedn.4xlarge",
            "instancesku": "MB6XH84AX23GRQ36",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Scalable (Icelake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 475 NVMe SSD",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "x2iedn.4xlarge",
        "name": "x2iedn.4xlarge",
        "ram": 524288,
    },
    "x2iedn.8xlarge": {
        "bandwidth": 25,
        "disk": 950,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x2iedn.8xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "1024 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon Scalable (Icelake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-southeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 950 NVMe SSD",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "x2iedn.8xlarge",
        "name": "x2iedn.8xlarge",
        "ram": 1048576,
    },
    "x2iedn.xlarge": {
        "bandwidth": 25,
        "disk": 118,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "3.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 20000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x2iedn.xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Scalable (Icelake)",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 118 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "x2iedn.xlarge",
        "name": "x2iedn.xlarge",
        "ram": 131072,
    },
    "x2iezn.12xlarge": {
        "bandwidth": 100,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "19000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x2iezn.12xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "1536 GiB",
            "networkPerformance": "100 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Platinum 8252",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "x2iezn.12xlarge",
        "name": "x2iezn.12xlarge",
        "ram": 1572864,
    },
    "x2iezn.2xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3170 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x2iezn.2xlarge",
            "instancesku": "D9NTTVU6X5ZXW8XW",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8252",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "x2iezn.2xlarge",
        "name": "x2iezn.2xlarge",
        "ram": 262144,
    },
    "x2iezn.4xlarge": {
        "bandwidth": 25,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "4750 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x2iezn.4xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "Up to 25 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "Intel Xeon Platinum 8252",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "x2iezn.4xlarge",
        "name": "x2iezn.4xlarge",
        "ram": 524288,
    },
    "x2iezn.6xlarge": {
        "bandwidth": 50,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "9500 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x2iezn.6xlarge",
            "instancesku": "8S5BF3CQBGR6BV8B",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "48",
            "physicalProcessor": "Intel Xeon Platinum 8252",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "24",
            "vpcnetworkingsupport": "true",
        },
        "id": "x2iezn.6xlarge",
        "name": "x2iezn.6xlarge",
        "ram": 786432,
    },
    "x2iezn.8xlarge": {
        "bandwidth": 75,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4.5 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "12000 Mbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x2iezn.8xlarge",
            "instancesku": "7349Q766H4JGZ9AW",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "1024 GiB",
            "networkPerformance": "75 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "Intel Xeon Platinum 8252",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "x2iezn.8xlarge",
        "name": "x2iezn.8xlarge",
        "ram": 1048576,
    },
    "x8g.12xlarge": {
        "bandwidth": 22,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "15 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x8g.12xlarge",
            "instancesku": "XTRBSN3Y6XH6M6W5",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "768 GiB",
            "networkPerformance": "22.5 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "x8g.12xlarge",
        "name": "x8g.12xlarge",
        "ram": 786432,
    },
    "x8g.16xlarge": {
        "bandwidth": 30,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "20 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x8g.16xlarge",
            "instancesku": "W3K52Q2ANTFXD2VC",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "1024 GiB",
            "networkPerformance": "30 Gigabit",
            "normalizationSizeFactor": "128",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "64",
            "vpcnetworkingsupport": "true",
        },
        "id": "x8g.16xlarge",
        "name": "x8g.16xlarge",
        "ram": 1048576,
    },
    "x8g.24xlarge": {
        "bandwidth": 40,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "30 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x8g.24xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "1536 GiB",
            "networkPerformance": "40 Gigabit",
            "normalizationSizeFactor": "192",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "96",
            "vpcnetworkingsupport": "true",
        },
        "id": "x8g.24xlarge",
        "name": "x8g.24xlarge",
        "ram": 1572864,
    },
    "x8g.2xlarge": {
        "bandwidth": 15,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x8g.2xlarge",
            "instancesku": "DYZFKST25Q3D7D8G",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "128 GiB",
            "networkPerformance": "Up to 15 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "x8g.2xlarge",
        "name": "x8g.2xlarge",
        "ram": 131072,
    },
    "x8g.48xlarge": {
        "bandwidth": 50,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "40 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x8g.48xlarge",
            "instancesku": "CFJS22QC729XBQZR",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "3072 GiB",
            "networkPerformance": "50 Gigabit",
            "normalizationSizeFactor": "384",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "192",
            "vpcnetworkingsupport": "true",
        },
        "id": "x8g.48xlarge",
        "name": "x8g.48xlarge",
        "ram": 3145728,
    },
    "x8g.4xlarge": {
        "bandwidth": 15,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x8g.4xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "256 GiB",
            "networkPerformance": "Up to 15 Gigabit",
            "normalizationSizeFactor": "32",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "16",
            "vpcnetworkingsupport": "true",
        },
        "id": "x8g.4xlarge",
        "name": "x8g.4xlarge",
        "ram": 262144,
    },
    "x8g.8xlarge": {
        "bandwidth": 15,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x8g.8xlarge",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "512 GiB",
            "networkPerformance": "15 Gigabit",
            "normalizationSizeFactor": "64",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "32",
            "vpcnetworkingsupport": "true",
        },
        "id": "x8g.8xlarge",
        "name": "x8g.8xlarge",
        "ram": 524288,
    },
    "x8g.large": {
        "bandwidth": 12,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x8g.large",
            "instancesku": "8DZHQBW3QMV56WF8",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 12.5 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "x8g.large",
        "name": "x8g.large",
        "ram": 32768,
    },
    "x8g.medium": {
        "bandwidth": 12,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x8g.medium",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 12.5 Gigabit",
            "normalizationSizeFactor": "2",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "us-west-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "1",
            "vpcnetworkingsupport": "true",
        },
        "id": "x8g.medium",
        "name": "x8g.medium",
        "ram": 16384,
    },
    "x8g.xlarge": {
        "bandwidth": 12,
        "disk": 0,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "2.7 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 10 Gbps",
            "ecu": "NA",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "x8g.xlarge",
            "instancesku": "D45QVQ5D8P59AAQW",
            "intelAvx2Available": "No",
            "intelAvxAvailable": "No",
            "intelTurboAvailable": "No",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 12.5 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "AWS Graviton4 Processor",
            "processorArchitecture": "64-bit",
            "regionCode": "eu-central-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "EBS only",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "x8g.xlarge",
        "name": "x8g.xlarge",
        "ram": 65536,
    },
    "z1d.12xlarge": {
        "bandwidth": 25,
        "disk": 1800,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "14000 Mbps",
            "ecu": "235",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "z1d.12xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "384 GiB",
            "networkPerformance": "25 Gigabit",
            "normalizationSizeFactor": "96",
            "physicalProcessor": "Intel Xeon Platinum 8151",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-southeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "2 x 900 NVMe SSD",
            "vcpu": "48",
            "vpcnetworkingsupport": "true",
        },
        "id": "z1d.12xlarge",
        "name": "z1d.12xlarge",
        "ram": 393216,
    },
    "z1d.2xlarge": {
        "bandwidth": 10,
        "disk": 300,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "2333 Mbps",
            "ecu": "45",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "z1d.2xlarge",
            "instancesku": "TJN3TJCJX99J6MMW",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "64 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "16",
            "physicalProcessor": "Intel Xeon Platinum 8151",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 300 NVMe SSD",
            "vcpu": "8",
            "vpcnetworkingsupport": "true",
        },
        "id": "z1d.2xlarge",
        "name": "z1d.2xlarge",
        "ram": 65536,
    },
    "z1d.3xlarge": {
        "bandwidth": 10,
        "disk": 450,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "3500 Mbps",
            "ecu": "64",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "z1d.3xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "96 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "24",
            "physicalProcessor": "Intel Xeon Platinum 8151",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 450 NVMe SSD",
            "vcpu": "12",
            "vpcnetworkingsupport": "true",
        },
        "id": "z1d.3xlarge",
        "name": "z1d.3xlarge",
        "ram": 98304,
    },
    "z1d.6xlarge": {
        "bandwidth": 10,
        "disk": 900,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "7000 Mbps",
            "ecu": "116",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "z1d.6xlarge",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "192 GiB",
            "networkPerformance": "10 Gigabit",
            "normalizationSizeFactor": "48",
            "physicalProcessor": "Intel Xeon Platinum 8151",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "eu-west-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 900 NVMe SSD",
            "vcpu": "24",
            "vpcnetworkingsupport": "true",
        },
        "id": "z1d.6xlarge",
        "name": "z1d.6xlarge",
        "ram": 196608,
    },
    "z1d.large": {
        "bandwidth": 10,
        "disk": 75,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2333 Mbps",
            "ecu": "12",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "z1d.large",
            "instancesku": "TC5JJ4DDT49HWM6A",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "16 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "4",
            "physicalProcessor": "Intel Xeon Platinum 8151",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "us-east-1",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 75 NVMe SSD",
            "vcpu": "2",
            "vpcnetworkingsupport": "true",
        },
        "id": "z1d.large",
        "name": "z1d.large",
        "ram": 16384,
    },
    "z1d.xlarge": {
        "bandwidth": 10,
        "disk": 150,
        "extra": {
            "availabilityzone": "NA",
            "classicnetworkingsupport": "false",
            "clockSpeed": "4 GHz",
            "currentGeneration": "Yes",
            "dedicatedEbsThroughput": "Up to 2333 Mbps",
            "ecu": "23",
            "enhancedNetworkingSupported": "Yes",
            "gpuMemory": "NA",
            "instanceFamily": "Memory optimized",
            "instanceType": "z1d.xlarge",
            "instancesku": "85BM84CW7NC939FV",
            "intelAvx2Available": "Yes",
            "intelAvxAvailable": "Yes",
            "intelTurboAvailable": "Yes",
            "marketoption": "OnDemand",
            "memory": "32 GiB",
            "networkPerformance": "Up to 10 Gigabit",
            "normalizationSizeFactor": "8",
            "physicalProcessor": "Intel Xeon Platinum 8151",
            "processorArchitecture": "64-bit",
            "processorFeatures": "Intel AVX; Intel AVX2; Intel AVX512; Intel Turbo",
            "regionCode": "ap-northeast-2",
            "servicecode": "AmazonEC2",
            "servicename": "Amazon Elastic Compute Cloud",
            "storage": "1 x 150 NVMe SSD",
            "vcpu": "4",
            "vpcnetworkingsupport": "true",
        },
        "id": "z1d.xlarge",
        "name": "z1d.xlarge",
        "ram": 32768,
    },
}
